@@ -30,6 +30,10 @@
 (defmacro truly-the (type expr)
   `(the ,type ,expr))
 
+(defmacro the* ((type &rest args) expr)
+  (declare (ignore args))
+  `(the ,type ,expr))
+
 (defmacro named-lambda (name args &body body)
   (declare (ignore name))
   `#'(lambda ,args ,@body))
@@ -198,10 +202,16 @@
 ;;; host CL package. This works around situations where the host has *more*
 ;;; symbols exported from CL than should be.
 (defun sb-xc:symbol-package (symbol)
-  (let ((p (cl:symbol-package symbol)))
+  (let ((p (cl:symbol-package symbol))
+        (name (string symbol)))
     (if (and p
-             (or (eq (find-symbol (string symbol) "XC-STRICT-CL") symbol)
-                 (eq (find-symbol (string symbol) "SB-XC") symbol)))
+             (or (eq (find-symbol name "XC-STRICT-CL") symbol)
+                 (eq (find-symbol name "SB-XC") symbol)
+                 ;; OK if the name of a symbol in the host CL package
+                 ;; is found in XC-STRICT-CL, even if the symbols
+                 ;; differ.
+                 (and (find-symbol name "XC-STRICT-CL")
+                      (eq (find-symbol name "CL") symbol))))
         *cl-package*
         p)))
 
@@ -212,6 +222,36 @@
   (write structure :stream stream :circle t))
 
 (in-package "SB-KERNEL")
+
+;;; These functions are required to emulate SBCL kernel functions
+;;; in a vanilla ANSI Common Lisp cross-compilation host.
+;;; The emulation doesn't need to be efficient, since it's needed
+;;; only for object dumping.
+
+;; The set of structure types that we access by slot position at cross-compile
+;; time is fairly small:
+;;   - DEFINITION-SOURCE-LOCATION
+;;   - DEFSTRUCT-DESCRIPTION, DEFSTRUCT-SLOT-DESCRIPTION
+;;   - DEBUG-SOURCE, COMPILED-DEBUG-INFO, COMPILED-DEBUG-FUN-{something}
+;;   - HEAP-ALIEN-INFO and ALIEN-{something}-TYPE
+;;   - COMMA
+(defun classoid-layout (x)
+  (declare (notinline classoid-wrapper))
+  (classoid-wrapper x))
+(defun layout-friend (x) x)
+(defun wrapper-friend (x) x)
+(defmacro wrapper-%info (x) `(layout-info ,x))
+(defun %instance-layout (instance)
+  (classoid-layout (find-classoid (type-of instance))))
+(defun %instance-length (instance)
+  (declare (notinline layout-length))
+  ;; In the target, it is theoretically possible to have %INSTANCE-LENGTH
+  ;; exceeed layout length, but in the cross-compiler they're the same.
+  (layout-length (%instance-layout instance)))
+(defun %raw-instance-ref/word (instance index)
+  (declare (ignore instance index))
+  (error "No such thing as raw structure access on the host"))
+
 (defun %find-position (item seq from-end start end key test)
   (let ((position (position item seq :from-end from-end
                             :start start :end end :key key :test test)))
@@ -223,8 +263,8 @@
 
 ;;; Needed for constant-folding
 (defun system-area-pointer-p (x) x nil) ; nothing is a SAP
-;;; Needed for DEFINE-MOVE-FUN LOAD-SYSTEM-AREA-POINTER
-(defun sap-int (x) (error "can't take SAP-INT ~S" x))
+(defmacro sap-ref-word (sap offset)
+  `(#+64-bit sap-ref-64 #-64-bit sap-ref-32 ,sap ,offset))
 
 (defun logically-readonlyize (x) x)
 
@@ -336,7 +376,7 @@
             (destructuring-bind (symbol value) (matchp '((quote ?) ? . :ignore) (cdr form))
               `(progn (defconstant ,symbol ,value)
                       (sb-c::%defconstant ',symbol ,symbol nil))))
-           (sb-xc:defconstant ; we see this maro as well. The host expansion will not do,
+           (defconstant ; we see this macro as well. The host expansion will not do,
             ;; because it calls our %defconstant which does not assign the symbol a value.
             ;; It might be possible to change that now that we don't use CL: symbols.
             (destructuring-bind (symbol value) (cdr form)
