@@ -50,6 +50,11 @@
   ;; a list of DEFSTRUCT-SLOT-DESCRIPTION objects for all slots
   ;; (including included ones)
   (slots () :type list)
+  ;; bit mask containing a 1 for each word that the garbage collector must visit
+  ;; (as opposed to a raw slot). Certain slot types (notably fixnum) may have either
+  ;; a 0 or a 1 in the mask because it does not matter if it is seen by GC.
+  ;; Bit index 0 in the mask is the word just after the header, and so on.
+  (bitmap +layout-all-tagged+ :type integer)
   ;; a list of (NAME . INDEX) pairs for accessors of included structures
   (inherited-accessor-alist () :type list)
   ;; number of data words, including the layout itself if the layout
@@ -82,18 +87,31 @@
   ;; PURIFY).
   ;; This is only meaningful if DD-CLASS-P = T.
   (pure nil :type (member t nil)))
-#-sb-fluid (declaim (freeze-type defstruct-description))
+(declaim (freeze-type defstruct-description))
 (!set-load-form-method defstruct-description (:host :xc :target))
 
 ;;;; basic LAYOUT stuff
 
 ;;; Careful here: if you add more bits, then adjust the bit packing for
 ;;; 64-bit layouts which also store LENGTH + DEPTHOID in the same word.
-(defconstant +structure-layout-flag+         #b00000001)
-(defconstant +pathname-layout-flag+          #b00000010)
-(defconstant +condition-layout-flag+         #b00001000)
-(defconstant +pcl-object-layout-flag+        #b00010000)
-(defconstant sb-vm:lockfree-list-node-flag   #b01000000) ; exported for use in gc-private.h
+(defconstant +structure-layout-flag+         #b000000001)
+(defconstant +pathname-layout-flag+          #b000000010)
+(defconstant +pcl-object-layout-flag+        #b000000100)
+(defconstant +condition-layout-flag+         #b000001000)
+(defconstant +simple-stream-layout-flag+     #b000010000)
+(defconstant +file-stream-layout-flag+       #b000100000)
+(defconstant +string-stream-layout-flag+     #b001000000)
+(defconstant +stream-layout-flag+            #b010000000)
+(defconstant +sequence-layout-flag+          #b100000000)
+
+;;; the type of LAYOUT-DEPTHOID and LAYOUT-LENGTH values.
+;;; Each occupies two bytes of the %BITS slot when possible,
+;;; otherwise a slot unto itself.
+(def!type layout-depthoid () '(integer -1 #x7FFF))
+(def!type layout-length () '(integer 0 #xFFFF))
+(def!type layout-bitmap () 'integer)
+;;; ID must be an fixnum for either value of n-word-bits.
+(def!type layout-id () '(signed-byte 30))
 
 ;;; The LAYOUT structure is pointed to by the first cell of instance
 ;;; (or structure) objects. It represents what we need to know for
@@ -130,15 +148,12 @@
 
 ;;; 32-bit is not done yet. Three slots are still used, instead of two.
 
-;;; Maximum value of N in ANCESTOR_N. Couldn't come up with a better name.
-(defconstant sb-c::layout-inherits-max-optimized-depth 5)
-(sb-xc:defstruct (layout
-             ;; Accept a specific subset of keywords
-             #+64-bit (:constructor %make-layout
-                          (clos-hash classoid flags info bitmap))
-             #-64-bit (:constructor %make-layout
-                          (clos-hash classoid depthoid length flags info bitmap))
-             (:copier nil))
+#-metaspace
+(progn
+(sb-xc:defstruct (layout (:copier nil)
+                         ;; Parsing DEFSTRUCT uses a temporary layout
+                         (:constructor make-temporary-layout
+                             (clos-hash classoid inherits &aux (invalid nil))))
 
   ;; A packed field containing the DEPTHOID, LENGTH, and FLAGS
   #+64-bit (flags 0 :type (signed-byte #.sb-vm:n-word-bits))
@@ -188,83 +203,111 @@
   #-64-bit (length 0 :type layout-length) ; smaller than SB-INT:INDEX
   ;; If this layout has some kind of compiler meta-info, then this is
   ;; it. If a structure, then we store the DEFSTRUCT-DESCRIPTION here.
-  (info nil :type (or null defstruct-description))
-  ;; Map of raw slot indices.
-  (bitmap +layout-all-tagged+ :type layout-bitmap)
-  ;; Per-slot comparator for implementing EQUALP.
-  (equalp-tests #() :type simple-vector)
   ;; If this layout is for an object of metatype STANDARD-CLASS,
-  ;; these are the EFFECTIVE-SLOT-DEFINITION metaobjects.
-  (slot-list nil :type list)
+  ;; then these are the EFFECTIVE-SLOT-DEFINITION metaobjects.
+  ;; The two are mutually exclusive.
+  (%info nil :type (or list defstruct-description))
+  ;; EQUALP comparator for two instances with this layout
+  ;; Could be the generalized function, or a type-specific one
+  ;; if the defstruct was compiled in a policy of SPEED 3.
+  (equalp-impl #'equalp-err :type (sfunction (t t) boolean) :read-only t)
   ;; Information about slots in the class to PCL: this provides fast
   ;; access to slot-definitions and locations by name, etc.
   ;; See MAKE-SLOT-TABLE in pcl/slots-boot.lisp for further details.
   (slot-table #(1 nil) :type simple-vector)
-  ;; inherited layouts or 0, only pertinent to structure classoids.
-  ;; There is no need to store the layout at depths 0 or 1
-  ;; since they're predetermined to be T and STRUCTURE-OBJECT.
-  (ancestor_2 0)
-  (ancestor_3 0)
-  (ancestor_4 0)
-  (ancestor_5 0))
-(declaim (freeze-type layout))
+  (id-word0 0 :type word)
+  (id-word1 0 :type word)
+  (id-word2 0 :type word)
+  #-64-bit (id-word3 0 :type word)
+  #-64-bit (id-word4 0 :type word)
+  #-64-bit (id-word5 0 :type word))
+(declaim (freeze-type layout)))
 
-;;; Applicable only if bit-packed (for 64-bit architectures)
-(defmacro pack-layout-flags (depthoid length flags)
-  `(logior (ash ,(or depthoid -1) (+ 32 sb-vm:n-fixnum-tag-bits))
-           (ash ,(or length 0) 16)
-           ,(or flags 0)))
-
-(defmacro set-layout-inherits (layout inherits &optional depthoid)
-  `(let* ((l ,layout) (i ,inherits) (d ,(or depthoid '(length i))))
-     (declare (ignorable i d))
-     ;; I tried putting a /SHOW here for debugging, but it's just too broken
-     ;; because layouts affect the printer dispatch mechanism.
-     #+nil
-     (let ((*print-pretty* nil))
-       (fresh-line)
-       (write-string "SET-INHERITS ")
-       (write (layout-classoid-name l))
-       (write-string " ")
-       (write (map 'list #'layout-classoid-name i))
-       (terpri))
-     (setf (layout-inherits l) i)
-     #-sb-xc-host
-     (setf (layout-ancestor_2 l) (if (> d 2) (svref i 2) 0)
-           (layout-ancestor_3 l) (if (> d 3) (svref i 3) 0)
-           (layout-ancestor_4 l) (if (> d 4) (svref i 4) 0)
-           (layout-ancestor_5 l) (if (> d 5) (svref i 5) 0))
-     l))
+#+metaspace
+(progn
+;;; Separating the pointer and nonpointer slots of LAYOUT satisfies the
+;;; requirement of a certain garbage collector (WIP), but also opens up the
+;;; possibility of absorbing at least one of the vectorish slots into the
+;;; WRAPPER since it could be reallocated with variable size.
+;;; The SLOT-TABLE slot might be a good candidate for trailing tagged slots.
+(sb-xc:defstruct (wrapper (:copier nil) (:constructor %make-wrapper))
+  (clos-hash (missing-arg) :type (and fixnum unsigned-byte)) ; redundant
+  (classoid (missing-arg) :type classoid)
+  (inherits #() :type simple-vector)
+  (equalp-impl #'equalp-err :type (sfunction (t t) boolean) :read-only t)
+  (slot-table #(1 nil) :type simple-vector)
+  (%info nil :type (or list defstruct-description))
+  (invalid :uninitialized :type (or cons (member nil t :uninitialized)))
+  (friend nil :type layout))
+;;; See LAYOUT for the remarks about each slot.
+;;; LAYOUT points to WRAPPER and vice-versa.
+;;; The most common LAYOUT is 8 words.
+;;; Needing >64 words would be quite unusual - the layout would have
+;;; an enormous depthoid or bitmap or both.
+(sb-xc:defstruct (layout (:copier nil)
+                         ;; Parsing DEFSTRUCT uses a temporary layout
+                         (:constructor %make-temporary-layout (friend clos-hash)))
+  ;; !!! The FRIEND slot in LAYOUT *MUST* BE FIRST !!!
+  (friend nil :type wrapper)
+  (clos-hash (missing-arg) :type (and fixnum unsigned-byte))
+  (flags 0 :type (signed-byte #.sb-vm:n-word-bits))
+  (id-word0 0 :type word)
+  (id-word1 0 :type word)
+  (id-word2 0 :type word)
+  ;; There are zero or more raw words if a type needs to store additional layout-ids,
+  ;; and there are one or more raw words for the GC bitmap.
+  )
+(declaim (freeze-type wrapper layout))
 
 #-sb-xc-host
-(defun make-layout (clos-hash classoid
-                    &key (depthoid -1) (length 0) (flags 0)
-                         (inherits #() inheritsp)
-                         (info nil)
-                         (bitmap (if info (dd-bitmap info) +layout-all-tagged+)))
-  (let ((layout (%make-layout clos-hash classoid
-                              #+64-bit (pack-layout-flags depthoid length flags)
-                              #-64-bit depthoid #-64-bit length #-64-bit flags
-                              info bitmap)))
-    (when inheritsp
-      (set-layout-inherits layout inherits))
-    layout))
+(progn
+(defmacro layout-classoid (layout) `(wrapper-classoid (layout-friend ,layout)))
+(defmacro layout-slot-table (layout) `(wrapper-slot-table (layout-friend ,layout)))
+(defmacro layout-%info (layout) `(wrapper-%info (layout-friend ,layout)))
+(defmacro layout-equalp-impl (layout) `(wrapper-equalp-impl (layout-friend ,layout)))
+(defun layout-inherits (layout) (wrapper-inherits (layout-friend layout)))
+(declaim (inline layout-invalid))
+(defun layout-invalid (layout)
+  (wrapper-invalid (layout-friend layout)))
+(defun (setf layout-invalid) (newval layout)
+  (setf (wrapper-invalid (layout-friend layout)) newval))
+(defun make-temporary-layout (clos-hash classoid inherits)
+  (let* ((wrapper (%make-wrapper :clos-hash clos-hash :classoid classoid
+                                 :inherits inherits :invalid nil
+                                 :friend #.(find-layout t)))
+         (layout (%make-temporary-layout wrapper clos-hash)))
+    (setf (wrapper-friend wrapper) layout)
+    layout))))
+
+#+(and (not metaspace) (not sb-xc-host))
+(progn (defmacro layout-friend (x) x)
+       (defmacro wrapper-friend (x) x)
+       (defmacro wrapper-%info (x) `(layout-%info ,x)))
 
 ;;; The cross-compiler representation of a LAYOUT omits several things:
-;;;   * BITMAP - obtainable via (DD-MAPMAP (LAYOUT-INFO layout)).
+;;;   * BITMAP - obtainable via (DD-BITMAP (LAYOUT-INFO layout)).
 ;;;     GC wants it in the layout to avoid double indirection.
 ;;;   * EQUALP-TESTS - needed only for the target's implementation of EQUALP.
 ;;;   * SLOT-TABLE, and SLOT-LIST - used only by the CLOS implementation.
-;;;   * ANCESTOR_N are optimizations for TYPEP.
+;;;   * ID-WORDn are optimizations for TYPEP.
 ;;; So none of those really make sense on the host.
 ;;; Also, we eschew the packed representation of length+depthoid+flags.
 ;;; FLAGS are not even strictly necessary, since they are for optimizing
 ;;; various type checks.
 #+sb-xc-host
 (progn
+  ;; As far as the host is concerned, LAYOUT and WRAPPER are always the same.
+  ;; We need this deftype because of the DEF!STRUCT for CLASSOID which states
+  ;; that the type of the slot named WRAPPER is WRAPPER.
+  #+metaspace (deftype wrapper () 'layout)
   (defstruct (layout (:include structure!object)
-                     (:constructor make-layout
-                                   (clos-hash classoid &key depthoid length flags inherits info)))
+                     (:constructor host-make-layout
+                                   (id clos-hash classoid &key info depthoid inherits
+                                                          length flags invalid)))
+    (id nil :type (or null fixnum))
+    ;; Cross-compiler-only translation from slot index to symbol naming
+    ;; the accessor to call. (Since access by position is not a thing)
+    (index->accessor-map #() :type simple-vector)
     ;; CLOS-HASH is needed to convert some TYPECASE forms to jump tables.
     ;; Theoretically we don't need this in the cross-compiler, because the
     ;; layout has a classoid which has a name which has a known hash.
@@ -277,15 +320,163 @@
     (depthoid -1 :type layout-depthoid)
     (length 0 :type layout-length)
     (info nil :type (or null defstruct-description)))
+  (defun layout-dd (layout)
+    (the defstruct-description (layout-info layout)))
+  (defun make-temporary-layout (clos-hash classoid inherits)
+    (host-make-layout nil clos-hash classoid :inherits inherits :invalid nil))
   (defun layout-bitmap (layout)
     (if (layout-info layout) (dd-bitmap (layout-info layout)) +layout-all-tagged+)))
 
-(defmacro sb-c::layout-nth-ancestor-slot (n)
-  `(case ,n
-     (2 'layout-ancestor_2)
-     (3 'layout-ancestor_3)
-     (4 'layout-ancestor_4)
-     (5 'layout-ancestor_5)))
+(defun equalp-err (a b)
+  (bug "EQUALP ~S ~S" a b))
+
+(defmacro get-dsd-index (type-name slot-name)
+  (declare (notinline dsd-index)) ; avoid later inlining failure style-warning
+  (dsd-index (find slot-name
+                   (dd-slots (find-defstruct-description type-name))
+                   :key #'dsd-name)))
+
+#-sb-xc-host
+(defun layout-id (layout)
+  ;; If a structure type at depthoid >= 2, then fetch the INDEXth id
+  ;; where INDEX is depthoid - 2. Otherwise fetch the 0th id.
+  ;; There are a few non-structure types at positive depthoid; those do not store
+  ;; their ancestors in the vector; they only store self-id at index 0.
+  ;; This isn't performance-critical. If it were, then we should store self-ID
+  ;; at a fixed index. Using it for type-based dispatch remains a possibility.
+  (let* ((depth (- (layout-depthoid layout) 2))
+         (index (if (or (< depth 0) (not (logtest (layout-flags layout)
+                                                  +structure-layout-flag+)))
+                    0 depth)))
+    (truly-the layout-id
+              #-64-bit
+              (%raw-instance-ref/signed-word
+               layout (+ (get-dsd-index layout id-word0) index))
+              #+64-bit ; use SAP-ref for lack of half-sized slots
+              (with-pinned-objects (layout)
+                (let ((sap (sap+ (int-sap (get-lisp-obj-address layout))
+                                 (- (ash (+ sb-vm:instance-slots-offset
+                                            (get-dsd-index layout id-word0))
+                                         sb-vm:word-shift)
+                                    sb-vm:instance-pointer-lowtag))))
+                  (signed-sap-ref-32 sap (ash index 2)))))))
+
+;;; Applicable only if bit-packed (for 64-bit architectures)
+(defmacro pack-layout-flags (depthoid length flags)
+  `(logior (ash ,depthoid (+ 32 sb-vm:n-fixnum-tag-bits)) (ash ,length 16) ,flags))
+
+(defmacro type-dd-length (type-name)
+  (dd-length (find-defstruct-description type-name)))
+
+(defconstant layout-id-vector-fixed-capacity 7)
+(defmacro calculate-extra-id-words (depthoid)
+  ;; There are 1 or 2 ids per word depending on n-word-bytes.
+  ;; We can always store IDs at depthoids 2,3,4,5,6,7,
+  ;; so depthoid less than or equal to 7 needs no extra words.
+  ;; 0 and 1 for T and STRUCTURE-OBJECT respectively are not stored.
+  `(ceiling (max 0 (- ,depthoid ,layout-id-vector-fixed-capacity))
+            ,(/ sb-vm:n-word-bytes 4)))
+
+;;; See the pictures above DD-BITMAP in src/code/defstruct for the details.
+(defconstant standard-gf-primitive-obj-layout-bitmap
+  #+compact-instance-header  6
+  #-compact-instance-header -4)
+
+;;; For lack of any better to place to write up some detail surrounding
+;;; layout creation for structure types, I'm putting here.
+;;; When you issue a DEFSTRUCT at the REPL, there are *three* instances
+;;; of LAYOUT makde for the new structure.
+;;; 1) The first is one associated with a temporary instance of
+;;; structure-classoid used in parsing the DEFSTRUCT form so that
+;;; we don't signal an UNKNOWN-TYPE condition for something like:
+;;;   (defstruct chain (next nil :type (or null chain)).
+;;; The temporary classoid is garbage immediately after parsing
+;;; and is never installed.
+;;; 2) The next is the actual LAYOUT that ends up being registered.
+;;; 3) The third is a layout created when setting the "compiler layout"
+;;; which contains copies of the length/depthoid/inherits etc
+;;; that we compare against the installed one to make sure they match.
+;;; The third one also gets thrown away.
+#-sb-xc-host
+(progn
+(declaim (inline layout-dd layout-info))
+;; Use LAYOUT-DD to read LAYOUT-INFO if you want to assert that it is non-nil.
+(defun layout-dd (layout)
+  (the defstruct-description (layout-%info layout)))
+(defun layout-info (layout)
+  (let ((info (layout-%info layout)))
+    (if (%instancep info) info)))
+(defun (setf layout-info) (newval layout)
+  ;; The current value must be nil or a defstruct-description,
+  ;; otherwise we'd clobber a non-nil slot list.
+  (aver (not (consp (layout-%info layout))))
+  (setf (layout-%info layout) newval))
+
+(define-load-time-global *layout-id-generator* (cons 0 nil))
+(declaim (type (cons fixnum) *layout-id-generator*))
+(defun make-layout (clos-hash classoid
+                    &key (depthoid -1) (length 0) (flags 0)
+                         (inherits #())
+                         (info nil)
+                         (bitmap (if info (dd-bitmap info) 0))
+                         (invalid :uninitialized)
+                    &aux (id (or (atomic-pop (cdr *layout-id-generator*))
+                                 (atomic-incf (car *layout-id-generator*)))))
+  (unless (typep id '(and layout-id (not (eql 0))))
+    (error "Layout ID limit reached"))
+  (let* ((fixed-words (type-dd-length layout))
+         (extra-id-words ; count of additional words needed to store ancestors
+          (if (logtest flags +structure-layout-flag+)
+              (calculate-extra-id-words depthoid)
+              0))
+         (bitmap-words (ceiling (1+ (integer-length bitmap)) sb-vm:n-word-bits))
+         (nwords (+ fixed-words extra-id-words bitmap-words))
+         (layout (truly-the layout
+                 #+immobile-space
+                 (sb-vm::alloc-immobile-fixedobj
+                  (1+ nwords)
+                  (logior (ash nwords sb-vm:instance-length-shift)
+                          sb-vm:instance-widetag))
+                 #-immobile-space (%make-instance nwords))))
+    (setf (%instance-layout layout) #.(find-layout 'layout))
+    #+metaspace
+    (let ((wrapper (%make-wrapper :clos-hash clos-hash :classoid classoid  :%info info
+                                  :inherits inherits :invalid invalid
+                                  :friend layout)))
+      (setf (layout-friend layout) wrapper))
+    (setf (layout-flags layout) #+64-bit (pack-layout-flags depthoid length flags)
+                                #-64-bit flags)
+    (setf (layout-clos-hash layout) clos-hash
+          (layout-classoid layout) classoid
+          (layout-invalid layout) invalid)
+    #-64-bit (setf (layout-depthoid layout) depthoid (layout-length layout) length)
+    (setf (layout-info layout) info)
+    (setf (layout-slot-table layout) #(1 nil))
+    (set-layout-inherits layout inherits (logtest flags +structure-layout-flag+) id)
+    (let ((bitmap-base (+ fixed-words extra-id-words)))
+      (dotimes (i bitmap-words)
+        (setf (%raw-instance-ref/word layout (+ bitmap-base i))
+              (ldb (byte sb-vm:n-word-bits (* i sb-vm:n-word-bits)) bitmap))))
+    ;; It's not terribly important that we recycle layout IDs, but I have some other
+    ;; changes planned that warrant a finalizer per layout.
+    (unless (built-in-classoid-p classoid)
+      (finalize layout (lambda () (atomic-push id (cdr *layout-id-generator*)))
+                :dont-save t))
+    layout))
+
+(declaim (inline layout-bitmap-words))
+(defun layout-bitmap-words (layout)
+  (declare (layout layout))
+  (- (%instance-length layout) (type-dd-length layout)))
+(defun layout-bitmap (layout)
+  (acond ((layout-info layout) (dd-bitmap it))
+         ;; Instances lacking DD-INFO are CLOS objects, which can't generally have
+         ;; raw slots, except that funcallable-instances may have 2 raw slots -
+         ;; the trampoline and the layout. The trampoline can have a tag, depending
+         ;; on the platform, and the layout is tagged but a special case.
+         ;; In any event, the bitmap is always 1 word, and there are no "extra ID"
+         ;; words preceding it.
+         (t (%raw-instance-ref/signed-word layout (type-dd-length layout))))))
 
 #+(and (not sb-xc-host) 64-bit)
 ;;; LAYOUT-DEPTHOID gets a vop and a stub
@@ -319,10 +510,6 @@
 (defun layout-for-pcl-obj-p (x)
   (logtest (layout-flags x) +pcl-object-layout-flag+))
 
-(declaim (inline sb-fasl:dumpable-layout-p))
-(defun sb-fasl:dumpable-layout-p (x)
-  (and (typep x 'layout) (not (layout-for-pcl-obj-p x))))
-
 ;;; The CLASSOID structure is a supertype of all classoid types.  A
 ;;; CLASSOID is also a CTYPE structure as recognized by the type
 ;;; system.  (FIXME: It's also a type specifier, though this might go
@@ -347,8 +534,11 @@
              #-sb-xc-host (:pure nil))
   ;; the value to be returned by CLASSOID-NAME.
   (name nil :type symbol)
-  ;; the current layout for this class, or NIL if none assigned yet
-  (layout nil :type (or layout null))
+  ;; the current WRAPPER, or LAYOUT, for this classoid,
+  ;; or NIL if none assigned yet.
+  ;; The name of the slot is always WRAPPER even when WRAPPER = LAYOUT.
+  ;; See doc/internals-notes/metaspace for further details.
+  (wrapper nil :type (or null #+metaspace wrapper #-metaspace layout))
   ;; How sure are we that this class won't be redefined?
   ;;   :READ-ONLY = We are committed to not changing the effective
   ;;                slots or superclasses.
@@ -371,6 +561,14 @@
   ;; NIL if none assigned yet
   (pcl-class nil))
 
+;; XC version is defined in cross-misc
+#-sb-xc-host
+(progn
+  (declaim (inline classoid-layout))
+  (defun classoid-layout (classoid)
+    #-metaspace (classoid-wrapper classoid)
+    #+metaspace (awhen (classoid-wrapper classoid) (wrapper-friend it))))
+
 (defun layout-classoid-name (x)
   (classoid-name (layout-classoid x)))
 
@@ -494,6 +692,70 @@
 (declaim (freeze-type built-in-classoid condition-classoid
                       standard-classoid static-classoid))
 
+#-sb-xc-host
+(defun id-to-layout (id)
+  (maphash (lambda (classoid wrapper &aux (layout (wrapper-friend wrapper)))
+             (declare (ignore classoid))
+             (when (eql (layout-id layout) id) (return-from id-to-layout layout)))
+           (classoid-subclasses (find-classoid 't))))
+(export 'id-to-layout)
+
+(defun summarize-layouts ()
+   (flet ((flag-bits (x) (logand (layout-flags x) #xffff)))
+     (let ((prev -1))
+       (dolist (layout (sort (loop for v being each hash-value
+                                of (classoid-subclasses (find-classoid 't))
+                                collect v)
+                             #'> :key #'flag-bits))
+         (let ((flags (flag-bits layout)))
+           (unless (= flags prev)
+             (format t "Layout flags = #b~10,'0b~%" flags)
+             (setq prev flags)))
+         (format t "  ~a~%" layout)))))
+
+#+sb-xc-host
+(progn
+(defun make-layout (hash classoid &rest keys)
+  (apply #'host-make-layout
+         (cdr (assq (classoid-name classoid) *popular-structure-types*))
+         hash classoid
+         (let ((copy (copy-list keys))) (remf copy :bitmap) copy)))
+;; The target reconstructs layouts using FOP-LAYOUT, but the host uses MAKE-LOAD-FORM.
+(defmethod make-load-form ((layout layout) &optional env)
+  (declare (ignore env))
+  (labels ((externalize (layout &aux (classoid (layout-classoid layout))
+                                     (name (classoid-name classoid)))
+             (when (or (layout-invalid layout) (not name))
+               (sb-c:compiler-error "can't dump ~S" layout))
+             (aver (= (layout-flags layout)
+                      (typecase classoid
+                        (structure-classoid +structure-layout-flag+)
+                        (condition-classoid +condition-layout-flag+)
+                        (undefined-classoid
+                         (bug "xc MAKE-LOAD-FORM on undefined layout"))
+                        (t 0))))
+             `(xc-load-layout ',name
+                              ,(layout-depthoid layout)
+                              (vector ,@(map 'list #'externalize (layout-inherits layout)))
+                              ,(layout-length layout)
+                              ,(layout-bitmap layout)
+                              ,(layout-flags layout))))
+    (externalize layout)))
+(defun xc-load-layout (name depthoid inherits length bitmap flags)
+  (let ((classoid (find-classoid name)))
+    (aver (and classoid (not (undefined-classoid-p classoid))))
+    (let ((layout (classoid-layout classoid)))
+      (unless (and (= (layout-depthoid layout) depthoid)
+                   (= (length (layout-inherits layout)) (length inherits))
+                   (every #'eq (layout-inherits layout) inherits)
+                   (= (layout-length layout) length)
+                   (= (layout-bitmap layout) bitmap)
+                   (= (layout-flags layout) flags))
+        (error "XC can't reload layout for ~S with ~S vs ~A"
+               name (list depthoid inherits length bitmap flags) layout))
+      layout)))
+) ; end PROGN
+
 (in-package "SB-C")
 
 ;;; layout for this type being used by the compiler
@@ -511,7 +773,7 @@
   #+sb-xc-host
   (specifier-type 'function)
   #-sb-xc-host
-  (let* ((fdefn (sb-kernel::find-fdefn name))
+  (let* ((fdefn (sb-int:find-fdefn name))
          (fun (and fdefn (fdefn-fun fdefn))))
     (if fun
         (handler-bind ((style-warning #'muffle-warning))
@@ -527,15 +789,3 @@
 (define-info-type (:function :type)
   :type-spec (or ctype (cons (eql function)) (member :generic-function))
   :default #'ftype-from-fdefn)
-
-(defun summarize-layouts ()
-  (let ((prev -1))
-    (dolist (layout (sort (loop for v being each hash-value
-                                of (classoid-subclasses (find-classoid 't))
-                                collect v)
-                          #'< :key #'sb-kernel:layout-flags))
-      (let ((flags (sb-kernel:layout-flags layout)))
-        (unless (= flags prev)
-          (format t "Layout flags = ~d~%" flags)
-          (setq prev flags)))
-      (format t "  ~a~%" layout))))
