@@ -154,7 +154,9 @@
                  (error "Keyword type description is not a two-list: ~S." key))
                (let ((kwd (first key)))
                  (when (find kwd (key-info) :key #'key-info-name)
-                   (error "~@<repeated keyword ~S in lambda list: ~2I~_~S~:>"
+                   (error (sb-format:tokens
+                           "~@<repeated keyword ~S in lambda list: ~2I~_~
+                            ~/sb-impl:print-lambda-list/~:>")
                           kwd lambda-listy-thing))
                  (key-info
                   (make-key-info
@@ -875,7 +877,7 @@
 ;;;  SPECIFIER-TYPE:
 ;;;    disallow (VALUES ...) even if single value, and disallow *
 ;;;  VALUES-SPECIFIER-TYPE:
-;;;    allow VALUES and *
+;;;    allow VALUES, disallow *
 ;;; TYPE-OR-NIL-IF-UNKNOWN:
 ;;;    like SPECIFIER-TYPE, but return NIL if contains unknown
 ;;; all the above are funneled through BASIC-PARSE-TYPESPEC.
@@ -1031,8 +1033,13 @@
 ;;; This takes no CONTEXT (which implies lack of recursion) because
 ;;; you can't reasonably place a VALUES type inside another type.
 (defun values-specifier-type (type-specifier)
-  (dx-let ((context (make-type-context type-specifier)))
-    (basic-parse-typespec type-specifier context)))
+  ;; This catches uses of literal '* where it shouldn't appear, but it
+  ;; accidentally lets other uses slip through. We'd have to catch '*
+  ;; post-type-expansion to be more strict, but it isn't very important.
+  (if (eq type-specifier '*)
+      (error "* is not permitted as a type specifier")
+      (dx-let ((context (make-type-context type-specifier)))
+        (basic-parse-typespec type-specifier context))))
 
 ;;; This is like VALUES-SPECIFIER-TYPE, except that we guarantee to
 ;;; never return a VALUES type.
