@@ -121,8 +121,8 @@
                       (return))
                     (incf start count)))))))))))
 
-(defun pretty-misc (stream op &optional arg1 arg2)
-  (declare (ignore stream op arg1 arg2)))
+(defun pretty-misc (stream op arg)
+  (declare (ignore stream op arg)))
 
 ;;;; logical blocks
 
@@ -756,6 +756,9 @@ line break."
             (pprint-dispatch-entry-priority entry)
             (pprint-dispatch-entry-initial-p entry))))
 
+(defmethod print-object ((table pprint-dispatch-table) stream)
+  (print-unreadable-object (table stream :type t :identity t)))
+
 ;; Return T iff E1 is strictly less preferable than E2.
 (defun entry< (e1 e2)
   (declare (type pprint-dispatch-entry e1 e2))
