@@ -13,7 +13,7 @@
 
 ;;;; miscellaneous non-primitive predicates
 
-#-sb-fluid (declaim (inline streamp))
+(declaim (inline streamp))
 (defun streamp (stream)
   (typep stream 'stream))
 
@@ -89,7 +89,6 @@
   ;; the type it tests for in the Common Lisp type system, and since it's
   ;; only used in the implementation of a few specialized things.)
   (def-type-predicate-wrapper double-float-p)
-  (def-type-predicate-wrapper extended-char-p)
   (def-type-predicate-wrapper fdefn-p)
   (def-type-predicate-wrapper fixnump)
   (def-type-predicate-wrapper floatp)
@@ -139,8 +138,7 @@
   (def-type-predicate-wrapper simple-rank-1-array-*-p)
   (def-type-predicate-wrapper simple-string-p)
   (def-type-predicate-wrapper stringp)
-  (def-type-predicate-wrapper vectorp)
-  (def-type-predicate-wrapper vector-nil-p))
+  (def-type-predicate-wrapper vectorp))
 
 #+(or x86 x86-64 arm arm64)
 (defun fixnum-mod-p (x limit)
@@ -216,10 +214,10 @@
      (cond
        ((<= 0 object 1) 'bit)
        ((< object 0) 'fixnum)
-       (t `(integer 0 ,sb-xc:most-positive-fixnum))))
+       (t `(integer 0 ,most-positive-fixnum))))
     (integer
      (if (>= object 0)
-         `(integer ,(1+ sb-xc:most-positive-fixnum))
+         `(integer ,(1+ most-positive-fixnum))
          'bignum))
     (character
      (typecase object
@@ -235,25 +233,37 @@
            ((eq object nil) 'null)
            ((eq (sb-xc:symbol-package object) *keyword-package*) 'keyword)
            (t 'symbol)))
-    ((or array complex #+sb-simd-pack simd-pack #+sb-simd-pack-256 simd-pack-256)
-     (let ((sb-kernel::*unparse-allow-negation* nil))
-       (declare (special sb-kernel::*unparse-allow-negation*)) ; forward ref
-       (type-specifier (ctype-of object))))
+    (array
+     (let ((etype (specifier-type (array-element-type object))))
+       ;; Obviously :COMPLEXP is known to be T or NIL, but it's not allowed to
+       ;; return (NOT SIMPLE-ARRAY), so use :MAYBE in lieu of T.
+       (type-specifier
+        (make-array-type (array-dimensions object)
+                         :complexp (if (typep object 'simple-array) nil :maybe)
+                         :element-type etype
+                         :specialized-element-type etype))))
+    ((or complex #+sb-simd-pack simd-pack #+sb-simd-pack-256 simd-pack-256)
+     (type-specifier (ctype-of object)))
+    (simple-fun 'compiled-function)
     (t
-     (let* ((classoid (classoid-of object))
-            (name (classoid-name classoid)))
-       (if (%instancep object)
-           (case name
-             (sb-alien-internals:alien-value
-              `(alien
-                ,(sb-alien-internals:unparse-alien-type
-                  (sb-alien-internals:alien-value-type object))))
-             (t
-              (let ((pname (classoid-proper-name classoid)))
-                (if (classoid-p pname)
-                    (classoid-pcl-class pname)
-                    pname))))
-           name)))))
+     (let ((layout (layout-of object)))
+       (if (eq (get-lisp-obj-address layout) 0)
+            ;; An empty instance, e.g. created by with-output-to-string
+            'instance
+            (let* ((classoid (layout-classoid layout))
+                   (name (classoid-name classoid)))
+              (if (%instancep object)
+                  (case name
+                    (sb-alien-internals:alien-value
+                     `(alien
+                       ,(sb-alien-internals:unparse-alien-type
+                         (sb-alien-internals:alien-value-type object))))
+                    (t
+                     (let ((pname (classoid-proper-name classoid)))
+                       (if (classoid-p pname)
+                           (classoid-pcl-class pname)
+                           pname))))
+                  name)))))))
 
 ;;;; equality predicates
 
@@ -387,11 +397,11 @@ length and have identical components. Other arrays must be EQ to be EQUAL."
            (equal x y)))))
 
 ;;; EQUALP comparison of HASH-TABLE values
+;;; Can be called only if both X and Y are definitely hash-tables.
 (defun hash-table-equalp (x y)
-  (declare (type hash-table x y))
+  (declare (type hash-table x y) (explicit-check))
   (or (eq x y)
-      (and (hash-table-p y)
-           (eql (hash-table-count x) (hash-table-count y))
+      (and (eql (hash-table-count x) (hash-table-count y))
            (eql (hash-table-test x) (hash-table-test y))
            (block comparison-of-entries
              (maphash (lambda (key x-value)
@@ -402,75 +412,116 @@ length and have identical components. Other arrays must be EQ to be EQUAL."
                       x)
              t))))
 
-(declaim (inline instance-equalp))
+(macrolet ((slot-ref-equalp ()
+             `(let ((x-el (%instance-ref x i))
+                    (y-el (%instance-ref y i)))
+                (or (eq x-el y-el) (equalp x-el y-el)))))
 (defun instance-equalp (x y)
-  (let ((layout-x (%instance-layout x)))
-    (and
-     (eq layout-x (%instance-layout y))
-     (logtest +structure-layout-flag+ (layout-flags layout-x))
-     (macrolet ((slot-ref-equalp ()
-                  `(let ((x-el (%instance-ref x i))
-                         (y-el (%instance-ref y i)))
-                     (or (eq x-el y-el) (equalp x-el y-el)))))
-       (let ((n (%instance-length x)))
-         (if (eql (layout-bitmap layout-x) sb-kernel:+layout-all-tagged+)
-             (loop for i downfrom (1- n) to sb-vm:instance-data-start
-                   always (slot-ref-equalp))
-             (let ((comparators (layout-equalp-tests layout-x)))
-               (unless (= (length comparators) (- n sb-vm:instance-data-start))
-                 (bug "EQUALP got incomplete instance layout"))
-               ;; See remark at the source code for %TARGET-DEFSTRUCT
-               ;; explaining how to use the vector of comparators.
-               (loop for i downfrom (1- n) to sb-vm:instance-data-start
-                     for test = (data-vector-ref
-                                 comparators (- i sb-vm:instance-data-start))
-                     always (cond ((eql test 0) (slot-ref-equalp))
-                                  ((functionp test) (funcall test i x y))
-                                  (t))))))))))
+  (declare (optimize (safety 0)))
+  (loop for i downfrom (1- (%instance-length x)) to sb-vm:instance-data-start
+        always (slot-ref-equalp)))
+(defun instance-equalp* (comparators x y)
+  (declare (optimize (safety 0))
+           (simple-vector comparators)
+           (type instance x y))
+  ;; See remark at the source code for %TARGET-DEFSTRUCT
+  ;; explaining how to use the vector of comparators.
+  (loop for i downfrom (1- (%instance-length x)) to sb-vm:instance-data-start
+        for test = (data-vector-ref comparators (- i sb-vm:instance-data-start))
+        always (cond ((eql test 0) (slot-ref-equalp))
+                     ((functionp test) (funcall test i x y))
+                     (t)))))
 
-;;; Doesn't work on simple vectors
-(defun array-equalp (x y)
-  (declare (array x y))
-  (let ((rank (array-rank x)))
-    (and
-     (= rank (array-rank y))
-     (dotimes (axis rank t)
-       (unless (= (%array-dimension x axis)
-                  (%array-dimension y axis))
-         (return nil)))
-     (with-array-data ((x x) (start-x) (end-x) :force-inline t
-                                               :array-header-p t)
-       (with-array-data ((y y) (start-y) (end-y) :force-inline t
-                                                 :array-header-p t)
-         (declare (ignore end-y))
-         (let* ((reffers %%data-vector-reffers%%)
-                (getter-x (truly-the function (svref reffers (%other-pointer-widetag x))))
-                (getter-y (truly-the function (svref reffers (%other-pointer-widetag y)))))
-           (loop for x-i fixnum from start-x below end-x
-                 for y-i fixnum from start-y
-                 for x-el = (funcall getter-x x x-i)
-                 for y-el = (funcall getter-y y y-i)
-                 always (or (eq x-el y-el)
-                            (equalp x-el y-el)))))))))
-
-(defun vector-equalp (x y)
-  (declare (vector x y))
-  (let ((length (length x)))
-    (and (= length (length y))
-         (with-array-data ((x x) (start-x) (end-x) :force-inline t
-                                                   :check-fill-pointer t)
-           (with-array-data ((y y) (start-y) (end-y) :force-inline t
-                                                     :check-fill-pointer t)
-             (declare (ignore end-y))
-             (let* ((reffers %%data-vector-reffers%%)
-                    (getter-x (truly-the function (svref reffers (%other-pointer-widetag x))))
-                    (getter-y (truly-the function (svref reffers (%other-pointer-widetag y)))))
-               (loop for x-i fixnum from start-x below end-x
-                     for y-i fixnum from start-y
-                     for x-el = (funcall getter-x x x-i)
-                     for y-el = (funcall getter-y y y-i)
-                     always (or (eq x-el y-el)
-                                (equalp x-el y-el)))))))))
+(macrolet ((numericp (v)
+             (let ((widetags
+                    (map 'list #'sb-vm:saetp-typecode
+                         (remove-if (lambda (x)
+                                      (not (typep (sb-vm:saetp-ctype x) 'numeric-type)))
+                                    sb-vm:*specialized-array-element-type-properties*))))
+               `(%other-pointer-subtype-p ,v ',widetags)))
+           (compare-loop (typespec)
+             `(let ((x (truly-the (simple-array ,typespec 1) x))
+                    (y (truly-the (simple-array ,typespec 1) y)))
+                (loop for x-i fixnum from start-x below end-x
+                      for y-i fixnum from start-y
+                      always (= (aref x x-i) (aref y y-i))))))
+(defun array-equalp (a b)
+  (flet
+      ((data-vector-compare (x y start-x end-x start-y)
+         (declare (index start-x end-x start-y)
+                  (optimize (sb-c::insert-array-bounds-checks 0)))
+         (let ((xtag (%other-pointer-widetag (truly-the (simple-array * 1) x)))
+               (ytag (%other-pointer-widetag (truly-the (simple-array * 1) y))))
+           (case (if (= xtag ytag) xtag 0)
+             (#.sb-vm:simple-vector-widetag
+              (let ((x (truly-the simple-vector x))
+                    (y (truly-the simple-vector y)))
+                (loop for x-i fixnum from start-x below end-x
+                      for y-i fixnum from start-y
+                      always (let ((a (svref x x-i)) (b (svref y y-i)))
+                               (or (eq a b) (equalp a b))))))
+             ;; Special-case the important array types that would cause consing in aref.
+             ;; Though (UNSIGNED-BYTE 62) and (UNSIGNED-BYTE 63) arrays exist,
+             ;; I highly doubt that they occur anywhere except in contrived code
+             ;; that does nothing but test that those types exist. i.e. They are
+             ;; beyond worthless, and are frankly wasteful of space in array dispatch
+             ;; scenarios, or to put it mildy: I disagree with their existence per se.
+             #+64-bit (#.sb-vm:simple-array-unsigned-byte-64-widetag
+                       (compare-loop (unsigned-byte 64)))
+             #+64-bit (#.sb-vm:simple-array-signed-byte-64-widetag
+                       (compare-loop (signed-byte 64)))
+             #-64-bit (#.sb-vm:simple-array-unsigned-byte-32-widetag
+                       (compare-loop (unsigned-byte 32)))
+             #-64-bit (#.sb-vm:simple-array-signed-byte-32-widetag
+                       (compare-loop (signed-byte 32)))
+             ;; SINGLE-FLOAT wouldn't cons on 64-bit, but it should be treated
+             ;; no less efficiently than DOUBLE-FLOAT.
+             (#.sb-vm:simple-array-single-float-widetag
+              (compare-loop single-float))
+             (#.sb-vm:simple-array-double-float-widetag
+              (compare-loop double-float))
+             (#.sb-vm:simple-array-complex-single-float-widetag
+              (compare-loop (complex single-float)))
+             (#.sb-vm:simple-array-complex-double-float-widetag
+              (compare-loop (complex double-float)))
+             (t
+              (let* ((reffers %%data-vector-reffers%%)
+                     (getter-x (truly-the function (svref reffers xtag)))
+                     (getter-y (truly-the function (svref reffers ytag))))
+                ;; The arrays won't both be strings, because EQUALP has a case for that.
+                ;; If they're both numeric, use = as the test.
+                (if (and (numericp x) (numericp y))
+                    (loop for x-i fixnum from start-x below end-x
+                          for y-i fixnum from start-y
+                          always (= (funcall getter-x x x-i) (funcall getter-y y y-i)))
+                    ;; Everything else
+                    (loop for x-i fixnum from start-x below end-x
+                          for y-i fixnum from start-y
+                          for x-el = (funcall getter-x x x-i)
+                          for y-el = (funcall getter-y y y-i)
+                          always (or (eq x-el y-el)
+                                     (equalp x-el y-el))))))))))
+    (if (vectorp (truly-the array a))
+        (and (vectorp (truly-the array b))
+             (= (length a) (length b))
+             (with-array-data ((x a) (start-x) (end-x)
+                               :force-inline t :check-fill-pointer t)
+               (with-array-data ((y b) (start-y) (end-y)
+                                 :force-inline t :check-fill-pointer t)
+                 (declare (ignore end-y))
+                 (data-vector-compare x y start-x end-x start-y))))
+        (let ((rank (array-rank (truly-the array a))))
+          (and (= rank (array-rank (truly-the array b)))
+               (dotimes (axis rank t)
+                 (unless (= (%array-dimension a axis)
+                            (%array-dimension b axis))
+                   (return nil)))
+               (with-array-data ((x a) (start-x) (end-x)
+                                 :force-inline t :array-header-p t)
+                 (with-array-data ((y b) (start-y) (end-y)
+                                   :force-inline t :array-header-p t)
+                   (declare (ignore end-y))
+                   (data-vector-compare x y start-x end-x start-y)))))))))
 
 (defun equalp (x y)
   "Just like EQUAL, but more liberal in several respects.
@@ -486,28 +537,19 @@ length and have identical components. Other arrays must be EQ to be EQUAL."
          (and (consp y)
               (equalp (car x) (car y))
               (equalp (cdr x) (cdr y))))
-        ((pathnamep x)
-         (and (pathnamep y) (pathname= x y)))
-        ((hash-table-p x)
-         (and (hash-table-p y)
-              (hash-table-equalp x y)))
         ((%instancep x)
          (and (%instancep y)
-              (instance-equalp x y)))
-        ((and (simple-vector-p x) (simple-vector-p y))
-         (let ((len (length x)))
-           (and (= len (length y))
-                (loop for i below len ; somewhat faster than the generic loop
-                      always (let ((a (svref x i)) (b (svref y i)))
-                               (or (eq a b) (equalp a b)))))))
-        ((and (bit-vector-p x) (bit-vector-p y))
-         (bit-vector-= x y))
-        ((vectorp x)
-         (and (vectorp y)
-              (vector-equalp x y)))
+              (let ((layout (%instance-layout x)))
+                (and (logtest (logior +structure-layout-flag+ +pathname-layout-flag+)
+                              (layout-flags layout))
+                     (eq (%instance-layout y) layout)
+                     (funcall (layout-equalp-impl layout) x y)))))
         ((arrayp x)
          (and (arrayp y)
-              (array-equalp x y)))
+              ;; string-equal is nearly 2x the speed of array-equalp for comparing strings
+              (cond ((and (stringp x) (stringp y)) (string-equal x y))
+                    ((and (bit-vector-p x) (bit-vector-p y)) (bit-vector-= x y))
+                    (t (array-equalp x y)))))
         (t nil)))
 
 (let ((test-cases `(($0.0 $-0.0 t)
