@@ -157,8 +157,9 @@
         (simple-reader-error stream
                              "~S is not a defined structure type."
                              (car body)))
-      (let ((default-constructor (dd-default-constructor
-                                  (layout-info (classoid-layout classoid)))))
+      (let ((default-constructor
+             (dd-default-constructor
+              (sb-kernel::wrapper-%info (classoid-wrapper classoid)))))
         (unless default-constructor
           (simple-reader-error
            stream
@@ -298,7 +299,7 @@
              ;; Don't refer to the DD-SLOTS unless there is reason to,
              ;; that is, unless some slot might be raw.
              (if (/= +layout-all-tagged+ (layout-bitmap (%instance-layout tree)))
-                 (let ((dd (layout-info (%instance-layout tree))))
+                 (let ((dd (layout-dd (%instance-layout tree))))
                    (dolist (dsd (dd-slots dd))
                      (when (eq (dsd-raw-type dsd) t)
                        (process (%instance-ref tree (dsd-index dsd))))))
@@ -433,7 +434,6 @@
              'sb-kernel::character-decoding-error-in-dispatch-macro-char-comment
              :sub-char sub-char :position (file-position stream) :stream stream)
             (invoke-restart 'attempt-resync))))
-    (let ((stream (in-stream-from-designator stream)))
       (macrolet ((munch (get-char &optional finish)
                    `(do ((level 1)
                          (prev ,get-char char)
@@ -452,7 +452,7 @@
             (prepare-for-fast-read-char stream
               (munch (fast-read-char) (done-with-fast-read-char)))
             ;; fundamental-stream
-            (munch (read-char stream t)))))))
+            (munch (read-char stream t))))))
 
 ;;;; a grab bag of other sharp readmacros: #', #:, and #.
 
