@@ -352,10 +352,6 @@
           (if (and form doc) (cons form doc) (or form doc))))
   doc)
 
-(defun %simple-fun-next (simple-fun) ; DO NOT USE IN NEW CODE
-  (%code-entry-point (fun-code-header simple-fun)
-                     (1+ (%simple-fun-index simple-fun))))
-
 ;;; Return the number of bytes to subtract from the untagged address of SIMPLE-FUN
 ;;; to obtain the untagged address of its code component.
 ;;; See also CODE-FROM-FUNCTION.
@@ -383,16 +379,13 @@
         ;; return it unchanged in all other cases
         info)))
 
-(defun %code-entry-points (code-obj) ; DO NOT USE IN NEW CODE
-  (%code-entry-point code-obj 0))
-
 (declaim (inline code-obj-is-filler-p))
 (defun code-obj-is-filler-p (code-obj)
   ;; See also HOLE-P in the allocator (same thing but using SAPs)
   ;; and filler_obj_p() in the C code
   (eql (sb-vm::%code-boxed-size code-obj) 0))
 
-#+(or sparc alpha hppa ppc64)
+#+(or sparc ppc64)
 (defun code-trailer-ref (code offset)
   (with-pinned-objects (code)
     (sap-ref-32 (int-sap (get-lisp-obj-address code))
@@ -473,11 +466,6 @@
       (values (%primitive sb-c:compute-fun code-obj
                           (%code-fun-offset code-obj fun-index))))))
 
-(defun code-entry-points (code-obj) ; FIXME: obsolete
-  (let ((a (make-array (code-n-entries code-obj))))
-    (dotimes (i (length a) a)
-      (setf (aref a i) (%code-entry-point code-obj i)))))
-
 ;;; Return the 0-based index of SIMPLE-FUN within its code component.
 ;;; Computed via binary search.
 (defun %simple-fun-index (simple-fun)
