@@ -94,12 +94,13 @@
   ;; funcallable instances
   (def %make-funcallable-instance)
   (def %fun-layout)
-  (def %set-funcallable-instance-layout (fin new-value))
+  (def %set-fun-layout (fin new-value))
   (def %funcallable-instance-fun)
   (def (setf %funcallable-instance-fun) (fin new-value))
   (def %funcallable-instance-info (fin i))
   (def %set-funcallable-instance-info (fin i new-value))
-  #+(and compact-instance-header x86-64) (def layout-of)
+  #+compact-instance-header (progn (def layout-of)
+                                   (def %instanceoid-layout))
   #+64-bit (def layout-depthoid)
   ;; lists
   (def %rplaca (x val))
