@@ -213,6 +213,15 @@
 
 (load "src/cold/shebang.lisp")
 
+(defvar *shebang-backend-subfeatures*
+  (let* ((default-subfeatures nil)
+         (customizer-file-name "customize-backend-subfeatures.lisp")
+         (customizer (if (probe-file customizer-file-name)
+                         (compile nil
+                                  (read-from-file customizer-file-name))
+                         #'identity)))
+    (funcall customizer default-subfeatures)))
+
 ;;; When cross-compiling, the *FEATURES* set for the target Lisp is
 ;;; not in general the same as the *FEATURES* set for the host Lisp.
 ;;; In order to refer to target features specifically, we refer to
@@ -261,6 +270,13 @@
           (setq target-feature-list
                 (remove :int4-breakpoints target-feature-list))
           (warn "Removed :INT4-BREAKPOINTS from target features"))
+        (when (or (member :arm64 target-feature-list)
+                  (member :sse4 *shebang-backend-subfeatures*))
+          (push :round-float target-feature-list))
+        (when (and (member :arm64 target-feature-list)
+                  (member :darwin target-feature-list))
+          (push :arm-v8.1 *shebang-backend-subfeatures*))
+
         ;; Putting arch and gc choice first is visually convenient, versus
         ;; having to parse a random place in the line to figure out the value
         ;; of a binary choice {cheney vs gencgc} and architecture.
@@ -270,15 +286,6 @@
                               (remove arch (remove gc target-feature-list)))
                              #'string<))))
 
-(defvar *shebang-backend-subfeatures*
-  (let* ((default-subfeatures nil)
-         (customizer-file-name "customize-backend-subfeatures.lisp")
-         (customizer (if (probe-file customizer-file-name)
-                         (compile nil
-                                  (read-from-file customizer-file-name))
-                         #'identity)))
-    (funcall customizer default-subfeatures)))
-
 ;;; Call for effect of signaling an error if no target picked.
 (target-platform-keyword)
 
@@ -297,9 +304,6 @@
          ("(and pauseless-threadstart (not sb-thread))"
           ":PAUSELESS-THREADSTART requires :SB-THREAD")
          ("(and sb-safepoint (not sb-thread))" ":SB-SAFEPOINT requires :SB-THREAD")
-         ("(and sb-thruption (not sb-safepoint))" ":SB-THRUPTION requires :SB-SAFEPOINT")
-         ("(and unix sb-safepoint-strictly)"
-          ":SB-SAFEPOINT-STRICTLY not supported on this platform")
          ("(and sb-thread (not (or riscv ppc ppc64 x86 x86-64 arm64)))"
           ":SB-THREAD not supported on selected architecture")
          ("(and gencgc cheneygc)"
@@ -310,10 +314,8 @@
           ":GENCGC not supported on selected architecture")
          ("(not (or gencgc cheneygc))"
           "One of :GENCGC or :CHENEYGC must be enabled")
-         ("(and sb-safepoint (not (or arm64 ppc x86 x86-64)))"
-          ":SB-SAFEPOINT not supported on selected architecture")
-         ("(and sb-safepoint-strictly (not sb-safepoint))"
-          ":SB-SAFEPOINT-STRICTLY requires :SB-SAFEPOINT")
+         ("(and sb-safepoint (not (and (or arm64 x86 x86-64) (or darwin linux win32))))"
+          ":SB-SAFEPOINT not supported on selected arch/OS")
          ("(not (or elf mach-o win32))"
           "No execute object file format feature defined")
          ("(and cons-profiling (not sb-thread))" ":CONS-PROFILING requires :SB-THREAD")
