@@ -23,8 +23,7 @@
 (defconstant extended-symbol-size (1+ symbol-size))
 
 #+sb-thread
-(dolist (slot (primitive-object-slots
-               (find 'thread *primitive-objects* :key #'primitive-object-name)))
+(dovector (slot (primitive-object-slots (primitive-object 'thread)))
   (when (slot-special slot)
     (setf (info :variable :wired-tls (slot-special slot))
           (ash (slot-offset slot) word-shift))))
@@ -63,8 +62,7 @@
     ;; The sizer is short_boxed.
     (closure ,(or #+(or x86 x86-64) "closure" "short_boxed") "lose" "short_boxed")
     ;; Like closure, but these can also have a layout pointer in the high header bytes.
-    (funcallable-instance ,(or #+compact-instance-header "funinstance" "short_boxed")
-                          "lose" "short_boxed")
+    (funcallable-instance "funinstance" "lose" "short_boxed")
     ;; These have a scav and trans function, but no size function.
     #-(or x86 x86-64) (return-pc "return_pc_header" "return_pc_header" "lose")
 
@@ -85,7 +83,7 @@
     #+sb-simd-pack-256 (simd-pack-256 "unboxed")
     (filler "unboxed")
 
-    (simple-array "boxed")
+    (simple-array "array")
     (simple-array-unsigned-byte-2 "vector_unsigned_byte_2")
     (simple-array-unsigned-byte-4 "vector_unsigned_byte_4")
     (simple-array-unsigned-byte-7 "vector_unsigned_byte_8")
@@ -117,13 +115,12 @@
     (simple-array-nil "vector_nil")
     (simple-base-string "base_string")
     #+sb-unicode (simple-character-string "character_string")
-    #+sb-unicode (complex-character-string "boxed")
-    (complex-base-string "boxed")
-    (complex-vector-nil "boxed")
+    #+sb-unicode (complex-character-string "array")
+    (complex-base-string "array")
 
-    (complex-bit-vector "boxed")
-    (complex-vector "boxed")
-    (complex-array "boxed"))))
+    (complex-bit-vector "array")
+    (complex-vector "array")
+    (complex-array "array"))))
 
 #+sb-xc-host
 (defun write-gc-tables (stream)
