@@ -61,7 +61,7 @@
     (%test-headers value temp target not-p nil (canonicalize-widetags widetags)
                    :value-tn-ref args))))
 
-(defmacro !define-type-vop (pred-name type-codes
+(defmacro define-type-vop (pred-name type-codes
                              &optional (inherit 'type-predicate))
   (let ((cost (if (> (reduce #'max type-codes :key #'eval) lowtag-limit)
                   7
@@ -74,53 +74,67 @@
                     target not-p ,type-codes
                     :value-tn-ref args)))))
 
-(!define-type-vop fixnump
+#-x86-64 ; defined in compiler/x86-64/type-vops for x86-64
+(progn
+(define-type-vop unbound-marker-p (unbound-marker-widetag))
+
+(define-type-vop characterp (character-widetag))
+
+(define-type-vop single-float-p (single-float-widetag))
+
+(define-type-vop fixnump
   #.fixnum-lowtags
-  #+(or x86-64 x86) simple-type-predicate) ;; save a register
+  #+(or x86) simple-type-predicate) ;; save a register
 
-(!define-type-vop functionp (fun-pointer-lowtag))
+(define-type-vop functionp (fun-pointer-lowtag))
 
-(!define-type-vop listp (list-pointer-lowtag))
+(define-type-vop listp (list-pointer-lowtag))
 
-(!define-type-vop non-null-symbol-p (symbol-widetag))
+(define-type-vop %instancep (instance-pointer-lowtag))
 
-(!define-type-vop %instancep (instance-pointer-lowtag))
+(define-type-vop %other-pointer-p (other-pointer-lowtag))
 
-(!define-type-vop %other-pointer-p (other-pointer-lowtag))
+(define-type-vop non-null-symbol-p (symbol-widetag))
 
-(!define-type-vop bignump (bignum-widetag))
+(define-type-vop closurep (closure-widetag))
 
-(!define-type-vop ratiop (ratio-widetag))
+(define-type-vop simple-fun-p (simple-fun-widetag))
 
-(!define-type-vop complexp
+(define-type-vop funcallable-instance-p (funcallable-instance-widetag))
+) ; end PROGN
+
+(define-type-vop bignump (bignum-widetag))
+
+(define-type-vop ratiop (ratio-widetag))
+
+(define-type-vop complexp
   (complex-widetag complex-single-float-widetag complex-double-float-widetag
                    #+long-float complex-long-float-widetag))
 
-(!define-type-vop complex-rational-p (complex-widetag))
+(define-type-vop complex-rational-p (complex-widetag))
 
-(!define-type-vop complex-float-p
+(define-type-vop complex-float-p
   (complex-single-float-widetag complex-double-float-widetag
                                 #+long-float complex-long-float-widetag))
 
-(!define-type-vop complex-single-float-p (complex-single-float-widetag))
+(define-type-vop complex-single-float-p (complex-single-float-widetag))
 
-(!define-type-vop complex-double-float-p (complex-double-float-widetag))
+(define-type-vop complex-double-float-p (complex-double-float-widetag))
 
-(!define-type-vop single-float-p (single-float-widetag))
+(define-type-vop double-float-p (double-float-widetag))
 
-(!define-type-vop double-float-p (double-float-widetag))
-
-(!define-type-vop simple-string-p
+(define-type-vop simple-string-p
   (#+sb-unicode simple-character-string-widetag
-   simple-base-string-widetag simple-array-nil-widetag))
+   simple-base-string-widetag))
 
+#-x86-64
 (macrolet
     ((define-simple-array-type-vops ()
          `(progn
            ,@(map 'list
                   (lambda (saetp)
                     (let ((primtype (saetp-primitive-type-name saetp)))
-                    `(!define-type-vop
+                    `(define-type-vop
                       ,(symbolicate primtype "-P")
                       (,(saetp-typecode saetp)))))
                   *specialized-array-element-type-properties*))))
@@ -128,58 +142,46 @@
 
 (macrolet
     ((def ()
-       `(!define-type-vop simple-rank-1-array-*-p
+       `(define-type-vop simple-rank-1-array-*-p
          ,(map 'list #'saetp-typecode
                *specialized-array-element-type-properties*))))
   (def)) ; simple-rank-1-array-*-p
 
-(!define-type-vop characterp (character-widetag))
+(define-type-vop system-area-pointer-p (sap-widetag))
 
-(!define-type-vop system-area-pointer-p (sap-widetag))
+(define-type-vop weak-pointer-p (weak-pointer-widetag))
 
-(!define-type-vop weak-pointer-p (weak-pointer-widetag))
+(define-type-vop code-component-p (code-header-widetag))
 
-(!define-type-vop code-component-p (code-header-widetag))
+#-(or x86 x86-64) (define-type-vop lra-p (return-pc-widetag))
 
-#-(or x86 x86-64) (!define-type-vop lra-p (return-pc-widetag))
+(define-type-vop fdefn-p (fdefn-widetag))
 
-(!define-type-vop fdefn-p (fdefn-widetag))
-
-(!define-type-vop closurep (closure-widetag))
-
-(!define-type-vop simple-fun-p (simple-fun-widetag))
-
-(!define-type-vop funcallable-instance-p (funcallable-instance-widetag))
-
-(!define-type-vop array-header-p
+(define-type-vop array-header-p
   (simple-array-widetag
    #+sb-unicode complex-character-string-widetag
    complex-base-string-widetag complex-bit-vector-widetag
-   complex-vector-widetag complex-array-widetag complex-vector-nil-widetag))
+   complex-vector-widetag complex-array-widetag))
 
-(!define-type-vop simple-array-header-p
+(define-type-vop simple-array-header-p
   (simple-array-widetag))
 
-(!define-type-vop stringp
+(define-type-vop stringp
   (#+sb-unicode simple-character-string-widetag
    #+sb-unicode complex-character-string-widetag
-   simple-base-string-widetag complex-base-string-widetag
-   simple-array-nil-widetag complex-vector-nil-widetag))
+   simple-base-string-widetag complex-base-string-widetag))
 
-(!define-type-vop base-string-p
+(define-type-vop base-string-p
   (simple-base-string-widetag complex-base-string-widetag))
 
-(!define-type-vop bit-vector-p
+(define-type-vop bit-vector-p
   (simple-bit-vector-widetag complex-bit-vector-widetag))
 
-(!define-type-vop vector-nil-p
-  (simple-array-nil-widetag complex-vector-nil-widetag))
-
 #+sb-unicode
-(!define-type-vop character-string-p
+(define-type-vop character-string-p
   (simple-character-string-widetag complex-character-string-widetag))
 
-(!define-type-vop vectorp
+(define-type-vop vectorp
   (complex-vector-widetag .
    #.(append
       (map 'list
@@ -200,15 +202,15 @@
 ;;; type.) Thus, there's no point in building up the full machinery of
 ;;; associated backend type predicates and so forth as we do for
 ;;; ordinary type VOPs.
-(!define-type-vop complex-vector-p (complex-vector-widetag))
+(define-type-vop complex-vector-p (complex-vector-widetag))
 
-(!define-type-vop simple-array-p
+(define-type-vop simple-array-p
   (simple-array-widetag .
    #.(map 'list
           #'saetp-typecode
           *specialized-array-element-type-properties*)))
 
-(!define-type-vop arrayp
+(define-type-vop arrayp
   (simple-array-widetag
    complex-array-widetag
    complex-vector-widetag .
@@ -221,7 +223,7 @@
                   (list (saetp-complex-typecode saetp))))
               (coerce *specialized-array-element-type-properties* 'list)))))
 
-(!define-type-vop numberp
+(define-type-vop numberp
   (bignum-widetag
    ratio-widetag
    single-float-widetag
@@ -233,16 +235,16 @@
    #+long-float complex-long-float-widetag
    . #.fixnum-lowtags))
 
-(!define-type-vop rationalp
+(define-type-vop rationalp
   (ratio-widetag bignum-widetag . #.fixnum-lowtags))
 
-(!define-type-vop integerp
+(define-type-vop integerp
   (bignum-widetag . #.fixnum-lowtags))
 
-(!define-type-vop floatp
+(define-type-vop floatp
   (single-float-widetag double-float-widetag #+long-float long-float-widetag))
 
-(!define-type-vop realp
+(define-type-vop realp
   (ratio-widetag
    bignum-widetag
    single-float-widetag
@@ -251,9 +253,18 @@
    . #.fixnum-lowtags))
 
 #+sb-simd-pack
-(!define-type-vop simd-pack-p (simd-pack-widetag))
+(define-type-vop simd-pack-p (simd-pack-widetag))
 #+sb-simd-pack-256
-(!define-type-vop simd-pack-256-p (simd-pack-256-widetag))
-
-(!define-type-vop unbound-marker-p (unbound-marker-widetag))
+(define-type-vop simd-pack-256-p (simd-pack-256-widetag))
 
+;;; Not type vops, but generic over all backends
+(macrolet ((def (name lowtag)
+             `(define-vop ()
+                (:translate ,name)
+                (:policy :fast-safe)
+                (:args (x :scs (descriptor-reg)))
+                (:results (res :scs (unsigned-reg)))
+                (:result-types unsigned-num)
+                (:generator 2 (loadw res x 0 ,lowtag)))))
+  (def function-header-word fun-pointer-lowtag)
+  (def instance-header-word instance-pointer-lowtag))
