@@ -84,6 +84,10 @@
                          sb-vm:bignum-digits-offset
                          index offset))
 
+(defun dd-contains-raw-slots-p (dd)
+  (dolist (dsd (dd-slots dd))
+    (unless (eq (dsd-raw-type dsd) t) (return t))))
+
 (deftransform copy-structure ((instance) * * :result result :node node)
   (let* ((classoid (lvar-type instance))
          (name (and (structure-classoid-p classoid) (classoid-name classoid)))
@@ -104,8 +108,8 @@
                  ;; Also note that VAR-ALLOC can not cope with dynamic-extent except where
                  ;; support has been added (x86oid so far); so requiring an exact type here
                  ;; causes VAR-ALLOC to become FIXED-ALLOC which works on more architectures.
-                 #-c-stack-is-control-stack
-                 (and dd (eql (layout-bitmap layout) sb-kernel:+layout-all-tagged+))
+                 #-c-stack-is-control-stack (and dd (not (dd-contains-raw-slots-p dd)))
+
                  ;; Definitely do this if copying to stack
                  (or (lvar-dynamic-extent result)
                      ;; Or if it's a small fixed number of words
@@ -131,19 +135,24 @@
           (t
            `(%copy-instance-slots (%make-structure-instance ,dd nil) instance)))))
 
+(defun varying-length-struct-p (classoid)
+  ;; This is a nice feature to have in general, but at present it is only possible
+  ;; to make varying length instances of LAYOUT and nothing else.
+  (eq (classoid-name classoid) 'layout))
+
 (deftransform %instance-length ((instance))
   (let ((classoid (lvar-type instance)))
     (if (and (structure-classoid-p classoid)
              (sb-kernel::compiler-layout-ready-p (classoid-name classoid))
              (eq (classoid-state classoid) :sealed)
+             (not (varying-length-struct-p classoid))
              ;; TODO: if sealed with subclasses which add no slots, use the fixed length
              (not (classoid-subclasses classoid)))
-        (dd-length (layout-info (sb-kernel::compiler-layout-or-lose (classoid-name classoid))))
+        (dd-length (layout-dd (sb-kernel::compiler-layout-or-lose (classoid-name classoid))))
         (give-up-ir1-transform))))
 
-;;; The layout is stored in slot 0.
-;;; *** These next two transforms should be the only code, aside from
-;;;     some parts of the C runtime, with knowledge of the layout index.
+;;; *** These transforms should be the only code, aside from the C runtime
+;;;     with knowledge of the layout index.
 #+compact-instance-header
 (define-source-transform function-with-layout-p (x) `(functionp ,x))
 #-compact-instance-header
@@ -153,11 +162,7 @@
   (define-source-transform %set-instance-layout (x val)
     `(%instance-set ,x 0 (the layout ,val)))
   (define-source-transform function-with-layout-p (x)
-    `(funcallable-instance-p ,x))
-  (define-source-transform %fun-layout (x)
-    `(truly-the layout (%funcallable-instance-info ,x 0)))
-  (define-source-transform %set-funcallable-instance-layout (x val)
-    `(setf (%funcallable-instance-info ,x 0) (the layout ,val))))
+    `(funcallable-instance-p ,x)))
 
 ;;;; simplifying HAIRY-DATA-VECTOR-REF and HAIRY-DATA-VECTOR-SET
 
@@ -183,7 +188,7 @@
   "avoid runtime dispatch on array element type"
   (let* ((type (lvar-type array))
          (element-ctype (array-type-upgraded-element-type type))
-         (declared-element-ctype (array-type-declared-element-type type)))
+         (declared-element-ctype (declared-array-element-type type)))
     (declare (type ctype element-ctype))
     (when (eq *wild-type* element-ctype)
       (give-up-ir1-transform
@@ -202,8 +207,7 @@
                   ((type= element-ctype declared-element-ctype)
                    bare-form)
                   (t
-                   `(the ,(type-specifier declared-element-ctype)
-                         ,bare-form))))))))
+                   (the-unwild declared-element-ctype bare-form))))))))
 
 ;;; Transform multi-dimensional array to one dimensional data vector
 ;;; access.
@@ -277,7 +281,7 @@
   "avoid runtime dispatch on array element type"
   (let* ((type (lvar-type array))
          (element-ctype (array-type-upgraded-element-type type))
-         (declared-element-ctype (array-type-declared-element-type type)))
+         (declared-element-ctype (declared-array-element-type type)))
     (declare (type ctype element-ctype))
     (when (eq *wild-type* element-ctype)
       (give-up-ir1-transform
@@ -289,10 +293,9 @@
                   (type ,element-type-specifier new-value))
          ,(if (type= element-ctype declared-element-ctype)
               '(data-vector-set array index new-value)
-              `(truly-the ,(type-specifier declared-element-ctype)
-                 (data-vector-set array index
-                  (the ,(type-specifier declared-element-ctype)
-                       new-value))))))))
+              (truly-the-unwild declared-element-ctype
+                 `(data-vector-set array index
+                   ,(the-unwild declared-element-ctype 'new-value))))))))
 
 ;;; Transform multi-dimensional array to one dimensional data vector
 ;;; access.
@@ -688,15 +691,15 @@
      (values)))
 
 ;;;; transforms for EQL of floating point values
-#-(vop-named sb-vm::eql/single-float)
+(unless (vop-existsp :named sb-vm::eql/single-float)
 (deftransform eql ((x y) (single-float single-float))
-  '(= (single-float-bits x) (single-float-bits y)))
+  '(= (single-float-bits x) (single-float-bits y))))
 
-#-(vop-named sb-vm::eql/double-float)
+(unless (vop-existsp :named sb-vm::eql/double-float)
 (deftransform eql ((x y) (double-float double-float))
   #-64-bit '(and (= (double-float-low-bits x) (double-float-low-bits y))
                   (= (double-float-high-bits x) (double-float-high-bits y)))
-  #+64-bit '(= (double-float-bits x) (double-float-bits y)))
+  #+64-bit '(= (double-float-bits x) (double-float-bits y))))
 
 
 ;;;; modular functions
