@@ -17,7 +17,7 @@
 
 ;; Return a bitmask representing the LIST of lambda list keywords.
 (defmacro lambda-list-keyword-mask (list)
-  (if (constantp list)
+  (if (cl:constantp list)
       ;; When invoked with a quoted constant, some flexibility
       ;; is allowed, in that the input may be a single symbol.
       (let ((val (#+sb-xc constant-form-value #-sb-xc eval list)))
@@ -119,14 +119,19 @@
                            (char= (char name 0) #\&)))))
              (check-suspicious (kind form)
                (and (probably-ll-keyword-p form)
-                    (member form sb-xc:lambda-list-keywords)
+                    (member form lambda-list-keywords)
                     (report-suspicious kind form)))
              (report-suspicious (kind what)
-               (style-warn-once list "suspicious ~A ~S in lambda list: ~S."
+               (style-warn-once list (sb-format:tokens
+                                      "suspicious ~A ~S in lambda list: ~
+                                       ~/sb-impl:print-lambda-list/.")
                                 kind what list)
                nil) ; Avoid "return convention is not fixed" optimizer note
              (need-arg (state)
-               (croak "expecting variable after ~A in: ~S" state list))
+               (croak (sb-format:tokens
+                       "expecting variable after ~A in: ~
+                        ~/sb-impl:print-lambda-list/")
+                      state list))
              (need-symbol (x why)
                (unless (symbolp x)
                  (croak "~A is not a symbol: ~S" why x)))
@@ -179,7 +184,10 @@
                        (symbolp input))
                   (setf rest (list input)))
                  (t
-                  (croak "illegal dotted lambda list: ~S" list)))
+                  (croak (sb-format:tokens
+                          "illegal dotted lambda list: ~
+                           ~/sb-impl:print-lambda-list/")
+                         list)))
            (return))
          (shiftf last-arg arg (pop input))
 
@@ -220,7 +228,10 @@
                           (destructuring-bind "a destructuring lambda list")
                           (defmethod "a specialized lambda list")
                           (t context))))
-                   (croak "~A is not allowed in ~A: ~S" arg where list)))
+                   (croak (sb-format:tokens
+                           "~A is not allowed in ~A: ~
+                            ~/sb-impl:print-lambda-list/")
+                          arg where list)))
 
                ;; &ENVIRONMENT can't intercede between &KEY,&ALLOW-OTHER-KEYS.
                ;; For all other cases it's as if &ENVIRONMENT were never there.
@@ -235,9 +246,14 @@
                ;; a better thing can be said, e.g. &WHOLE must go to the front.
                (cond ((logbitp to-state seen) ; Oops! Been here before.
                       (if (= rest-bits 3)
-                          (croak "~S and ~S are mutually exclusive: ~S"
+                          (croak (sb-format:tokens
+                                  "~S and ~S are mutually exclusive: ~
+                                   ~/sb-impl:print-lambda-list/")
                                  '&body '&rest list)
-                          (croak "repeated ~S in lambda list: ~S" arg list)))
+                          (croak (sb-format:tokens
+                                  "repeated ~S in lambda list: ~
+                                   ~/sb-impl:print-lambda-list/")
+                                 arg list)))
                      ((logbitp state from-states) ; valid transition
                       (setq state to-state
                             seen (logior seen (ash 1 state))
@@ -245,10 +261,11 @@
                      ((logbitp state (bits &whole &rest &more &environment))
                       (need-arg last-arg)) ; Variable expected.
                      (t
-                      (croak (if (state= to-state &whole)
-                                 "~A must appear first in a lambda list: ~S"
-                                 "misplaced ~A in lambda list: ~S")
-                             arg list)))
+                      (croak (sb-format:tokens
+                              "~:[misplaced ~A in lambda list~;~
+                               ~A must appear first in a lambda list~]:
+                               ~/sb-impl:print-lambda-list/")
+                             (state= to-state &whole) arg list)))
                (go LOOP)))
            ;; Fell through, so warn if desired, and fall through some more.
            (unless silent (report-suspicious "variable" arg)))
@@ -256,7 +273,10 @@
          ;; Handle a lambda variable
          (when (logbitp state (bits &allow-other-keys ; Not a collecting state.
                                     :post-env :post-rest :post-more))
-           (croak "expected lambda list keyword at ~S in: ~S" arg list))
+           (croak (sb-format:tokens
+                   "expected lambda list keyword at ~S in: ~
+                    ~/sb-impl:print-lambda-list/")
+                  arg list))
          (let ((item (list arg)))
            (setq tail (if tail (setf (cdr tail) item) (begin-list item))))
          (when (logbitp state (bits &rest &more &whole &environment))
@@ -278,7 +298,9 @@
         (style-warn-once
          list
          (make-condition '&optional-and-&key-in-lambda-list
-                         :format-control "&OPTIONAL and &KEY found in the same lambda list: ~S"
+                         :format-control (sb-format:tokens
+                                          "&OPTIONAL and &KEY found in the same lambda list: ~
+                                           ~/sb-impl:print-lambda-list/")
                          :format-arguments (list list))))
 
       ;; For CONTEXT other than :VALUES-TYPE/:FUNCTION-TYPE we reject
@@ -308,10 +330,13 @@
                      ;; (This is not a regression)
                      (destructuring-bind (var &optional default sup-p) arg
                        (if (and (consp var) (eq what-kind '&key))
-                           (destructuring-bind (keyword-name var) var
-                             (unless (symbolp keyword-name)
-                               (croak "keyword-name in ~S is not a symbol" arg))
-                             (need-bindable var description))
+                           (cond ((singleton-p (cdr var))
+                                  (destructuring-bind (keyword-name var) var
+                                    (unless (symbolp keyword-name)
+                                      (croak "keyword-name in ~S is not a symbol" arg))
+                                    (need-bindable var description)))
+                                 (t
+                                  (croak "invalid &KEY syntax: ~S" var)))
                            (need-bindable var description))
                        ;; Inform the user about a possibly malformed
                        ;; destructuring list (&OPTIONAL (A &OPTIONAL B)).
@@ -413,7 +438,7 @@
                 (when (and (not silent) (vectorp parse)) ; is destructuring
                   (let ((default (and (cdr arg-specifier) ; have an explicit default
                                       (cadr arg-specifier))))
-                    (when (and (constantp default)
+                    (when (and (cl:constantp default)
                                (not (ds-lambda-list-match-p
                                      (#+sb-xc constant-form-value #-sb-xc eval
                                       default)
@@ -1056,7 +1081,7 @@
                             (pop plist)
                             (unless (or (keywordp key)
                                         (and (symbolp key)
-                                             (constantp key)
+                                             (cl:constantp key)
                                              (eq key (symbol-value key))))
                               (signal 'compiler-macro-keyword-problem
                                       :argument key))))
@@ -1276,3 +1301,18 @@
          (unless allow-special
            (lose "special variable"))))
       (values name kind))))
+
+;; This is a variant of destructuring-bind that provides the name
+;; of the containing construct in generated error messages.
+(macrolet (#+sb-xc-host ; Bootstrap NAMED-DS-BIND
+           (named-ds-bind (name lambda-list expression &body body)
+             (declare (ignore name))
+               `(cl:destructuring-bind ,lambda-list ,expression ,@body)))
+  (defmacro named-ds-bind (name lambda-list expression &body body
+                                                       &environment env)
+    (declare (ignore env)) ; could be policy-sensitive (but isn't)
+    `(binding* ,(sb-c::expand-ds-bind lambda-list expression t nil name
+                 (and (eq (car name) :macro)
+                      (eq (cddr name) 'deftype)
+                      ''*))
+       ,@body)))
