@@ -19,6 +19,47 @@
 (defvar *undefined-warnings*)
 (declaim (list *undefined-warnings*))
 
+;;; Delete any undefined warnings for NAME and KIND. This is for the
+;;; benefit of the compiler, but it's sometimes called from stuff like
+;;; type-defining code which isn't logically part of the compiler.
+(declaim (ftype (function ((or symbol cons) keyword) (values))
+                note-name-defined))
+(defun note-name-defined (name kind)
+  #-sb-xc-host (atomic-incf *type-cache-nonce*)
+  ;; We do this BOUNDP check because this function can be called when
+  ;; not in a compilation unit (as when loading top level forms).
+  (when (boundp '*undefined-warnings*)
+    (let ((name (uncross name)))
+      (setq *undefined-warnings*
+            (delete-if (lambda (x)
+                         (and (equal (undefined-warning-name x) name)
+                              (eq (undefined-warning-kind x) kind)))
+                       *undefined-warnings*))))
+  (values))
+
+;;; to be called when a variable is lexically bound
+(declaim (ftype (function (symbol) (values)) note-lexical-binding))
+(defun note-lexical-binding (symbol)
+    ;; This check is intended to protect us from getting silently
+    ;; burned when we define
+    ;;   foo.lisp:
+    ;;     (DEFVAR *FOO* -3)
+    ;;     (DEFUN FOO (X) (+ X *FOO*))
+    ;;   bar.lisp:
+    ;;     (DEFUN BAR (X)
+    ;;       (LET ((*FOO* X))
+    ;;         (FOO 14)))
+    ;; and then we happen to compile bar.lisp before foo.lisp.
+  (when (looks-like-name-of-special-var-p symbol)
+    ;; FIXME: should be COMPILER-STYLE-WARNING?
+    (style-warn 'asterisks-around-lexical-variable-name
+                :format-control
+                "using the lexical binding of the symbol ~
+                 ~/sb-ext:print-symbol-with-prefix/, not the~@
+                 dynamic binding"
+                :format-arguments (list symbol)))
+  (values))
+
 ;;; In the target compiler, a lexenv can hold an alist of condition
 ;;; types (CTYPE . ACTION) such that when signaling condition CTYPE,
 ;;; we perform ACTION which is usually MUFFLE-CONDITION.
@@ -206,8 +247,8 @@
   (setf (classoid-state class) :sealed)
   (let ((subclasses (classoid-subclasses class)))
     (when subclasses
-      (dohash ((subclass layout) subclasses :locked t)
-        (declare (ignore layout))
+      (dohash ((subclass wrapper) subclasses :locked t)
+        (declare (ignore wrapper))
         (setf (classoid-state subclass) :sealed)))))
 
 (defun process-freeze-type-declaration (type-specifier)
@@ -222,6 +263,16 @@
   (warn-if-inline-failed/proclaim name kind)
   (setf (info :function :inlinep name) kind))
 
+(defun process-block-compile-declaration (entries kind)
+  (ecase kind
+    (start-block
+     (finish-block-compilation)
+     (let ((compilation *compilation*))
+       (setf (block-compile compilation) t)
+       (setf (entry-points compilation) entries)))
+    (end-block
+     (finish-block-compilation))))
+
 (defun check-deprecation-declaration (state since form)
   (unless (typep state 'deprecation-state)
     (error 'simple-type-error
@@ -329,8 +380,16 @@
                                     (ftype #'proclaim-ftype))
                             ctype type :declared)))
              (push raw-form *queued-proclaims*)))
+        #-sb-fluid
         (freeze-type
          (map-args #'process-freeze-type-declaration))
+        ((start-block end-block)
+         (when (and *compile-time-eval* (boundp '*compilation*))
+           (if (eq *block-compile-argument* :specified)
+               (process-block-compile-declaration args kind)
+               (compiler-notify "ignoring ~S declaration since ~
+                                :BLOCK-COMPILE is not :SPECIFIED"
+                                kind))))
         (optimize
          (multiple-value-bind (new-policy specified-qualities)
              (process-optimize-decl form *policy*)
@@ -354,7 +413,7 @@
         ((disable-package-locks enable-package-locks)
          (setq *disabled-package-locks*
                (process-package-lock-decl form *disabled-package-locks*)))
-        ((inline notinline maybe-inline)
+        ((#-sb-fluid inline notinline #-sb-fluid maybe-inline)
          (map-args #'process-inline-declaration kind))
         (deprecated
          (destructuring-bind (state since &rest things) args
