@@ -109,29 +109,6 @@
                       ;; If it's a lisp-rep-type, the CTYPE should be one already.
                       (aver (not (compute-lisp-rep-type alien-type)))
                       `(sb-alien::alien-value-typep object ',alien-type)))
-            #+(vop-translates sb-int:fixnump-instance-ref)
-            ((and (type= type (specifier-type 'fixnum))
-                         (let ((use (lvar-uses object)))
-                           (and (combination-p use)
-                                (almost-immediately-used-p object use)
-                                (or (and (eq (lvar-fun-name (combination-fun use))
-                                             '%instance-ref)
-                                         (constant-lvar-p
-                                          (second (combination-args use))))
-                                    (member (lvar-fun-name (combination-fun use))
-                                            '(car cdr))))))
-             ;; FIXME: vopcombine should be able to combine a load
-             ;; and a fixnump test withoot special-casing these three.
-             (case (lvar-fun-name (combination-fun (lvar-uses object)))
-               (%instance-ref
-                (splice-fun-args object '%instance-ref 2)
-                `(lambda (obj i) (fixnump-instance-ref obj i)))
-               (car
-                (splice-fun-args object 'car 1)
-                `(lambda (obj) (fixnump-car obj)))
-               (cdr
-                (splice-fun-args object 'cdr 1)
-                `(lambda (obj) (fixnump-cdr obj)))))
             (t
              (give-up-ir1-transform))))))
 
@@ -333,18 +310,29 @@
                                type)))
                  (rational 'rational)
                  (float (or (numeric-type-format type) 'float))
-                 ((nil) 'real))))
+                 ((nil) 'real)))
+         (low (numeric-type-low type))
+         (high (numeric-type-high type)))
     (ecase (numeric-type-complexp type)
       (:real
-       (cond #+(or x86 x86-64 arm arm64) ;; Not implemented elsewhere yet
-             ((and
-               (eql (numeric-type-class type) 'integer)
-               (or (eql (numeric-type-low type) 0)
-                   (eql (numeric-type-low type) 1))
-               (fixnump (numeric-type-high type)))
+       (cond ((and (eql (numeric-type-class type) 'integer)
+                   (and (fixnump low)
+                        (fixnump high)
+                        (<= (1+ (- high low)) 2)))
+              ;; The fixnum-mod-p case is worse than just EQ testing with
+              ;; only 2 values in the range. (INTEGER 1 2) would have become
+              ;;   (and (not (eq x 0)) (fixnump x) (not (> x 2))).
+              ;; If exactly 1 value, it should have been picked off by TYPE-SINGLETON-P
+              ;; in %SOURCE-TRANSFORM-TYPEP, but even if it wasn't,
+              ;; the OR will drop out due to constraint propagation.
+              `(or (eq ,object ,low) (eq ,object ,high)))
+             #+(or x86 x86-64 arm arm64) ;; Not implemented elsewhere yet
+             ((and (eql (numeric-type-class type) 'integer)
+                   (or (eql low 0) (eql low 1))
+                   (fixnump (numeric-type-high type)))
               (let ((mod-p
                       `(fixnum-mod-p ,object ,(numeric-type-high type))))
-                (if (eql (numeric-type-low type) 1)
+                (if (eql low 1)
                     `(and (not (eq ,object 0))
                           ,mod-p)
                     mod-p)))
@@ -677,10 +665,9 @@
       (cond ((cdr dims)
              (values `(,header-test
                        ,@(when (eq (array-type-dimensions stype) '*)
-                           #+x86-64
-                           `((%array-rank= ,obj ,(length dims)))
-                           #-x86-64
-                           `((= (%array-rank ,obj) ,(length dims))))
+                           (if (vop-existsp :translate %array-rank=)
+                               `((%array-rank= ,obj ,(length dims)))
+                               `((= (%array-rank ,obj) ,(length dims)))))
                        ,@(loop for d in dims
                                for i from 0
                                unless (eq '* d)
@@ -838,59 +825,23 @@
 ;;; object's layout can ever be EQ to that of the ancestor.
 ;;; e.g. a fixnum as representative of class REAL.
 ;;; So in actual practice, you can't make something that is a pure STREAM, etc.
+#-(or x86 x86-64) ; vop-translated for these 2
+(defmacro layout-depthoid-ge (layout depthoid)
+  `(>= (layout-depthoid ,layout) ,depthoid))
 (defun transform-instance-typep (classoid)
   (binding*
       ((name (classoid-name classoid))
        (layout (let ((res (info :type :compiler-layout name)))
                  (when (and res (not (layout-invalid res))) res)))
-       ((primtype-predicate slot-reader)
+       ((lowtag lowtag-test slot-reader)
         (cond ((csubtypep classoid (specifier-type 'funcallable-instance))
-               (values '(function-with-layout-p object)
-                       '(%fun-layout object)))
-              ((or (csubtypep classoid (specifier-type 'instance))
-                   (eq classoid (specifier-type 'logical-pathname))
-                   ;; CONDITION can't be a funcallable-instance
-                   (csubtypep classoid (specifier-type 'condition)))
-               (values '(%instancep object)
-                       '(%instance-layout object)))))
-       (get-layout-or-return-false
-        (if primtype-predicate
-            ;; Test just one of %INSTANCEP or %FUNCALLABLE-INSTANCE-P
-            `(if ,primtype-predicate ,slot-reader (return-from typep nil))
-            ;; But if we don't know which is will be, try both.
-            ;; This is less general than LAYOUT-OF,and therefore
-            ;; a little quicker to fail, because objects with
-            ;; {LIST|OTHER}-POINTER-LOWTAG can't possibly pass.
-            ;; It's a bit disappointing that STREAM uses this slower path,
-            ;; but some people think it should be possible to create
-            ;; funcallable streams. As a countermeasure, it would be possible to emit
-            ;; slightly better code in a vop.
-            `(cond ((%instancep object) (%instance-layout object))
-                   ((function-with-layout-p object) (%fun-layout object))
-                   (t (return-from typep nil)))))
+               (values sb-vm:fun-pointer-lowtag
+                       '(function-with-layout-p object) '(%fun-layout object)))
+              ((csubtypep classoid (specifier-type 'instance))
+               (values sb-vm:instance-pointer-lowtag
+                       '(%instancep object) '(%instance-layout object)))))
        (depthoid (if layout (layout-depthoid layout) -1))
-       (n-layout (gensym))
-        ;; In order to efficiently perform the DEEPER-P test without this hack of using
-        ;; a vop (when available for non-risc machines), we'd have to do two things:
-        ;; - have instcombine combine the read and compare as one instruction
-        ;; - implement half-width structure slots
-        ;; Since both of those are not happening any time soon, ...
-       (deeper-p
-         #+(vop-translates sb-c::layout-depthoid-gt) `(layout-depthoid-gt ,n-layout ,depthoid)
-         #-(vop-translates sb-c::layout-depthoid-gt) `(> (layout-depthoid ,n-layout) ,depthoid))
-       (nth-ancestor ; This is possibly unused (if no compile-time layout, or depthoid -1)
-        ;; Use DATA-VECTOR-REF directly, since that's what SVREF in SAFETY 0 will become.
-        `(locally (declare (optimize (safety 0)))
-           (data-vector-ref (layout-inherits ,n-layout) ,depthoid)))
-       (ancestor-layout-eq
-        ;; Layouts are immediate constants in immobile space. Again, this is something that
-        ;; an instcombine pass might be able to recognize as having a single instruction.
-        #+(and immobile-space x86-64) `(sb-vm::layout-inherits-ref-eq
-                                        (layout-inherits ,n-layout) ,depthoid ,layout)
-        #-(and immobile-space x86-64) `(eq ,nth-ancestor ,layout))
-       ;; For shallow depthoid we can avoid checking the depthoid or reading the 'inherits'
-       ;; slot, because the layout has some number of ancestor layouts directly in it.
-       (ancestor-slot (layout-nth-ancestor-slot depthoid)))
+       (wrapper (make-symbol "LAYOUT")))
 
     ;; Easiest case first: single bit test.
     (cond ((member name '(condition pathname structure-object))
@@ -901,44 +852,34 @@
                              (pathname  +pathname-layout-flag+)
                              (t         +structure-layout-flag+)))))
 
-          ;; Next easiest: Sealed and at most one subclass.
-          ((and (eq (classoid-state classoid) :sealed) layout
-                 (or (not (classoid-subclasses classoid))
-                     (eql (hash-table-count (classoid-subclasses classoid))
-                          1)))
-            ;; It's possible to seal a STANDARD-CLASS, not just a STRUCTURE-CLASS,
-            ;; though probably extremely weird. Also the PRED should be set in
-            ;; that event, but it isn't.
-            ;; The crummy dual expressions for the same result are because
-            ;; (BLOCK (RETURN ...)) seems to emit a forward branch in the
-            ;; passing case, but AND emits a forward branch in the failing
-            ;; case which I believe is the better choice.
-            (let ((other-layout (and (classoid-subclasses classoid)
-                                     (dohash ((classoid layout)
-                                              (classoid-subclasses classoid)
-                                              :locked t)
-                                             (declare (ignore classoid))
-                                             (return layout)))))
-              (flet ((check-layout (layout-getter)
-                       (cond (other-layout
-                              ;; It's faster to compare two layouts than
-                              ;; doing whatever is done below
-                              `(let ((object-layout ,layout-getter))
-                                 (or (eq object-layout ',layout)
-                                     (eq object-layout ',other-layout))))
-                             #+(vop-named sb-vm::layout-eq)
-                             ((equal layout-getter '(%instance-layout object))
-                              `(sb-vm::layout-eq object ',layout))
-                             (t
-                              `(eq ,layout-getter ',layout)))))
-                (if primtype-predicate
-                    `(and ,primtype-predicate ,(check-layout slot-reader))
-                    `(block typep ,(check-layout get-layout-or-return-false))))))
+          ;; Next easiest: Sealed and no subtypes. Typically for DEFSTRUCT only.
+          ;; Even if you don't seal a DEFCLASS, we're allowed to assume that things
+          ;; won't change, as per CLHS 3.2.2.3 on Semantic Constraints:
+          ;;  "Classes defined by defclass in the compilation environment must be defined
+          ;;  at run time to have the same superclasses and same metaclass."
+          ;; I think that means we should know the lowtag always. Nonetheless, this isn't
+          ;; an important scenario, and only if you _do_ seal a class could this case be
+          ;; reached; users rarely seal their classes since the standard doesn't say how.
+          ((and layout
+                (eq (classoid-state classoid) :sealed)
+                (not (classoid-subclasses classoid)))
+           (if lowtag-test
+               `(and ,lowtag-test ,(if (vop-existsp :translate layout-eq)
+                                       `(layout-eq object ,layout ,lowtag)
+                                       `(eq ,slot-reader ,layout)))
+               ;; `(eq ,layout
+               ;;      (if-vop-existsp (:translate %instanceoid-layout)
+               ;;        (%instanceoid-layout object)
+               ;;        ;; Slightly quicker than LAYOUT-OF. See also %PCL-INSTANCE-P
+               ;;        (cond ((%instancep object) (%instance-layout object))
+               ;;              ((funcallable-instance-p object) (%fun-layout object))
+               ;;              (t ,(find-layout 't)))))
+               (bug "Unexpected metatype for ~S" layout)))
 
           ;; All other structure types
           ((and (typep classoid 'structure-classoid) layout)
             ;; structure type tests; hierarchical layout depths
-            (aver (equal primtype-predicate '(%instancep object)))
+            (aver (eql lowtag sb-vm:instance-pointer-lowtag))
             ;; we used to check for invalid layouts here, but in fact that's both unnecessary and
             ;; wrong; it's unnecessary because structure classes can't be redefined, and it's wrong
             ;; because it is quite legitimate to pass an object with an invalid layout
@@ -947,44 +888,42 @@
                     ;; If we allowed structure classes to be mixed in to standard-object,
                     ;; this might have to change to consider object invalidation. Probably would
                     ;; want to track structure classoids that would render this code inadmissible.
-                    (let ((,n-layout (%instance-layout object)))
-                        ,(cond ((<= 2 depthoid layout-inherits-max-optimized-depth)
-                                `(or (eq (,ancestor-slot ,n-layout) ,layout)
-                                     (eq ,n-layout ,layout)))
-                               ((dd-constructors (layout-info layout))
-                                `(cond ((eq ,n-layout ,layout) t)
-                                       (,deeper-p ,ancestor-layout-eq)))
-                               (t ; abstract base type deeper than optimized max.
-                                ;; Assume that no layout is EQ to the base layout,
-                                ;; and unconditionally fetch and dereference layout-inherits.
-                                `(eq (if ,deeper-p ,nth-ancestor ,n-layout) ,layout))))))
+                  ,(if (<= depthoid sb-kernel::layout-id-vector-fixed-capacity)
+                       `(%structure-is-a (%instance-layout object) ,layout)
+                       `(let ((,wrapper (%instance-layout object)))
+                          (and (layout-depthoid-ge ,wrapper ,depthoid)
+                               (%structure-is-a ,wrapper ,layout))))))
 
-          ((> depthoid 0) ; fixed-depth ancestors of non-structure types: STREAM, FILE-STREAM,
-           ;; SEQUENCE, CONDITION; all are abstract base types.
+          ((> depthoid 0)
+           ;; fixed-depth ancestors of non-structure types:
+           ;; STREAM, FILE-STREAM, STRING-STREAM, and SEQUENCE.
             #+sb-xc-host (when (typep classoid 'static-classoid)
                            ;; should have use :SEALED code above
                            (bug "Non-frozen static classoids?"))
-            ;; quasi-hierarchical layout for other things: STREAM, FILE-STREAM,
-            ;; SEQUENCE, CONDITION; all are abstract base types.
-            (let ((guts `(,@(unless (eq name 'condition)
-                              `((when (zerop (layout-clos-hash ,n-layout))
-                                 (setq ,n-layout
-                                  (truly-the layout
-                                             (update-object-layout-or-invalid
-                                              object ',layout))))))
-                          ;; the Nth ancestor for N=1 is not directly stored in the layout.
-                          ;; But we don't have to check the layout-inherits length because
-                          ;; it has as least two physical data elements even if it has T
-                          ;; as the sole ancestor.
-                          (or ,(if (eql depthoid 1)
-                                   ancestor-layout-eq
-                                   `(eq (,ancestor-slot ,n-layout) ,layout))
-                              ;; So that (TYPEP (MAKE-CONDITION 'CONDITION) 'CONDITION) => T
-                              ,@(when (eq name 'condition) `((eq ,n-layout ,layout)))))))
-              (if primtype-predicate
-                  `(and ,primtype-predicate (let ((,n-layout ,slot-reader)) ,@guts))
-                  `(block typep
-                     (let ((,n-layout ,get-layout-or-return-false)) ,@guts)))))
+            (let ((guts `((when (zerop (layout-clos-hash ,wrapper))
+                            (setq ,wrapper (update-object-layout object)))
+                          ,(ecase name
+                            (stream
+                             `(logtest (layout-flags ,wrapper) ,+stream-layout-flag+))
+                            (file-stream
+                             `(logtest (layout-flags ,wrapper) ,+file-stream-layout-flag+))
+                            (string-stream
+                             `(logtest (layout-flags ,wrapper) ,+string-stream-layout-flag+))
+                            ;; Testing the type EXTENDED-SEQUENCE tests for #<LAYOUT of SEQUENCE>.
+                            ;; It can only arise from a direct invocation of TRANSFORM-INSTANCE-TYPEP,
+                            ;; because the lisp type is not a classoid. It's done this way to define
+                            ;; the logic once only, instead of both here and src/code/pred.lisp.
+                            (sequence
+                             `(logtest (layout-flags ,wrapper) ,+sequence-layout-flag+))))))
+              (if lowtag-test
+                  `(and ,lowtag-test (let ((,wrapper ,slot-reader)) ,@guts))
+                  (if-vop-existsp (:translate %instanceoid-layout)
+                    `(let ((,wrapper (%instanceoid-layout object))) ,@guts)
+                    `(block typep
+                       (let ((,wrapper (cond ((%instancep object) (%instance-layout object))
+                                             ((funcallable-instance-p object) (%fun-layout object))
+                                             (t (return-from typep nil)))))
+                         ,@guts))))))
 
           (t
             `(classoid-cell-typep ',(find-classoid-cell name :create t)
@@ -1007,6 +946,10 @@
   (let ((ctype (careful-specifier-type type)))
     (if ctype
         (or
+         ;; It's purely a waste of compiler resources to wait for IR1 to
+         ;; see these 2 edge cases that can be decided right now.
+         (cond ((eq ctype *universal-type*) t)
+               ((eq ctype *empty-type*) nil))
          (and (not (intersection-type-p ctype))
               (multiple-value-bind (constantp value) (type-singleton-p ctype)
                 (and constantp
@@ -1034,8 +977,6 @@
            (args-type
             (compiler-warn "illegal type specifier for TYPEP: ~S" type)
             (return-from %source-transform-typep (values nil t)))
-           (t nil))
-         (typecase ctype
            (numeric-type
             (source-transform-numeric-typep object ctype))
            (classoid
@@ -1067,6 +1008,44 @@
           (values `(let ((,name ,object))
                      (%typep-wrapper ,transform ,name ',type)))))))
 
+;;; These things will be removed by the tree shaker, so no #+ needed.
+(defvar *interesting-types* nil)
+(defun involves-alien-p (ctype)
+  (sb-kernel::map-type
+   (lambda (type)
+     (when (alien-type-type-p type) (return-from involves-alien-p t)))
+   ctype))
+(defun dump/restore-interesting-types (op)
+  (declare (ignorable op))
+  #+collect-typep-regression-dataset
+  (ecase op
+   (write
+    (when *interesting-types*
+      (let ((list (sort (loop for k being each hash-key of *interesting-types* collect k)
+                        #'string< :key #'write-to-string)))
+        (with-open-file (f "interesting-types.lisp-expr" :direction :output
+                         :if-exists :supersede  :if-does-not-exist :create)
+          (let ((*package* #+sb-xc-host (find-package "XC-STRICT-CL")
+                           #-sb-xc-host #.(find-package "SB-KERNEL"))
+                (*print-pretty* nil)
+                (*print-length* nil)
+                (*print-level* nil)
+                (*print-readably* t))
+            (dolist (item list)
+              (write (uncross item) :stream f)
+              (terpri f)))))))
+   (read
+    (unless (hash-table-p *interesting-types*)
+      (setq *interesting-types* (make-hash-table :test 'equal :synchronized t)))
+    (with-open-file (f "interesting-types.lisp-expr" :if-does-not-exist nil)
+      (when f
+        (let ((*package* (find-package "SB-KERNEL")))
+          (loop (let ((expr (read f nil f)))
+                  (when (eq expr f) (return))
+                  (format t "Read ~a~%" expr)
+                  (setf (gethash expr *interesting-types*) t))))))
+    *interesting-types*)))
+
 (define-source-transform typep (object spec &optional env)
   ;; KLUDGE: It looks bad to only do this on explicitly quoted forms,
   ;; since that would overlook other kinds of constants. But it turns
@@ -1083,8 +1062,24 @@
         ;; during block compilation, we give ourselves a better chance
         ;; at open-coding the type test.
         (let ((type (cadr spec)))
+          ;;
+          #+collect-typep-regression-dataset
+          (let ((parse (specifier-type type)))
+            ;; alien types aren't externalizable as trees of symbols,
+            ;; and some classoid types aren't defined at the start of warm build,
+            ;; making it impossible to re-parse a dump produced late in the build.
+            ;; Luckily there are no cases involving compund types and classoids.
+            (unless (or (involves-alien-p parse)
+                        (or (classoid-p parse)
+                            (and (cons-type-p parse)
+                                 (classoid-p (cons-type-car-type parse)))))
+              (let ((table *interesting-types*))
+                (unless (hash-table-p table)
+                  (setq table (dump/restore-interesting-types 'read)))
+                (setf (gethash type table) t))))
+          ;;
           (if (and (block-compile *compilation*)
-                   (unknown-type-p (specifier-type type)))
+                   (contains-unknown-type-p (careful-specifier-type type)))
               (values nil t)
               (source-transform-typep object type))))
       (values nil t)))
@@ -1320,3 +1315,6 @@
           ((csubtypep (lvar-type object) this-type)
            nil)
           ((give-up-ir1-transform)))))
+
+;;; BIGNUMP is simpler than INTEGERP, so if we can rule out FIXNUM then ...
+(deftransform integerp ((x) ((not fixnum)) * :important nil) '(bignump x))
