@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 ;;;; data object ref/set stuff
 
@@ -30,7 +30,7 @@
   (:results)
   (:vop-var vop)
   (:generator 1
-    (cond ((emit-gc-barrier-store-p name)
+    (cond ((emit-code-page-write-barrier-p name)
            ;; Don't assume that the immediate is acceptable to 'push'
            ;; (only MOV can take a 64-bit immediate)
            (inst mov temp-reg-tn (encode-value-if-immediate value))
@@ -38,43 +38,40 @@
            ;; Push the slot index into code, then code itself
            (cond ((= lowtag fun-pointer-lowtag)
                   ;; compute code address similarly to CODE-FROM-FUNCTION vop
-                  (inst mov (reg-in-size temp-reg-tn :dword)
-                        (make-ea-for-object-slot-half object 0 fun-pointer-lowtag))
-                  (inst shr (reg-in-size temp-reg-tn :dword) n-widetag-bits)
+                  (inst mov :dword temp-reg-tn (ea (- fun-pointer-lowtag) object))
+                  (inst shr :dword temp-reg-tn n-widetag-bits)
                   ;; now temp-reg-tn holds the difference in words from code to fun.
                   (inst push temp-reg-tn)
-                  (inst add (make-ea :qword :base rsp-tn) offset) ; for particular slot
+                  (inst add :qword (ea 0 rsp-tn) offset) ; for particular slot
                   ;; finish computing the code address
                   (inst neg temp-reg-tn)
                   (inst lea temp-reg-tn
-                        (make-ea :qword :base object :index temp-reg-tn :scale n-word-bytes
-                                 :disp (- other-pointer-lowtag fun-pointer-lowtag)))
+                        (ea (- other-pointer-lowtag fun-pointer-lowtag)
+                            object temp-reg-tn n-word-bytes))
                   (inst push temp-reg-tn))
                  (t ; is code already
                   (inst push offset)
                   (inst push object)))
            (invoke-asm-routine 'call 'code-header-set vop))
+          ((equal name '(setf %funcallable-instance-fun))
+           (gen-cell-set (object-slot-ea object offset lowtag) value nil vop t))
           (t
-           (gen-cell-set (make-ea-for-object-slot object offset lowtag)
-                         value nil)))))
+           (gen-cell-set (object-slot-ea object offset lowtag) value nil)))))
 
 ;; INIT-SLOT has to know about the :COMPACT-INSTANCE-HEADER feature.
 (define-vop (init-slot set-slot)
   (:info name dx-p offset lowtag)
   (:generator 1
     (progn name dx-p)
-    (cond #!+compact-instance-header
+    (cond #+compact-instance-header
           ((and (eq name '%make-structure-instance) (eql offset :layout))
         ;; The layout is in the upper half of the header word.
-           (inst mov
-              (make-ea :dword :base object :disp (- 4 instance-pointer-lowtag))
-              (if (sc-is value immediate)
-                  (make-fixup (tn-value value) :layout)
-                  (reg-in-size value :dword))))
+           (inst mov :dword (ea (- 4 instance-pointer-lowtag) object)
+                 (if (sc-is value immediate)
+                     (make-fixup (tn-value value) :layout)
+                     value)))
           ((sc-is value immediate)
-           (move-immediate (make-ea :qword
-                                    :base object
-                                    :disp (- (* offset n-word-bytes) lowtag))
+           (move-immediate (ea (- (* offset n-word-bytes) lowtag) object)
                            (encode-value-if-immediate value)
                            temp-reg-tn (not dx-p)))
           (t
@@ -92,13 +89,18 @@
   (:results (result :scs (descriptor-reg any-reg)))
   (:generator 5
      (move rax old)
-     (inst cmpxchg (make-ea :qword :base object
-                            :disp (- (* offset n-word-bytes) lowtag))
-           new :lock)
+     (inst cmpxchg :lock (ea (- (* offset n-word-bytes) lowtag) object) new)
      (move result rax)))
 
 ;;;; symbol hacking VOPs
 
+(define-vop (make-unbound-marker)
+  (:args)
+  (:results (result :scs (descriptor-reg any-reg)))
+  (:generator 1
+    (inst mov result (logior (+ sb-vm:static-space-start #x100)
+                             unbound-marker-widetag))))
+
 (define-vop (%set-symbol-global-value)
   (:args (object :scs (descriptor-reg immediate))
          (value :scs (descriptor-reg any-reg immediate)))
@@ -107,7 +109,7 @@
     (gen-cell-set (cond ((sc-is object immediate)
                          (symbol-slot-ea (tn-value object) symbol-value-slot))
                         (t
-                         (make-ea-for-object-slot object symbol-value-slot
+                         (object-slot-ea object symbol-value-slot
                                                   other-pointer-lowtag)))
                   value nil)))
 
@@ -132,7 +134,7 @@
   (:generator 9
     (let ((err-lab (generate-error-code vop 'unbound-symbol-error object)))
       (loadw value object symbol-value-slot other-pointer-lowtag)
-      (inst cmp (reg-in-size value :dword) unbound-marker-widetag)
+      (inst cmp :byte value unbound-marker-widetag)
       (inst jmp :e err-lab))))
 
 ;; Return the DISP field to use in an EA relative to thread-base-tn
@@ -143,9 +145,9 @@
 
 (deftransform %compare-and-swap-symbol-value ((symbol old new)
                                               ((constant-arg symbol) t t))
-  (if (eq (info :variable :kind (sb!c::lvar-value symbol)) :global)
+  (if (eq (info :variable :kind (sb-c::lvar-value symbol)) :global)
       `(%cas-symbol-global-value symbol old new)
-      (sb!c::give-up-ir1-transform)))
+      (sb-c::give-up-ir1-transform)))
 
 (macrolet (;; Logic common to thread-aware SET and CAS. CELL is assigned
            ;; to the location that should be accessed to modify SYMBOL's
@@ -155,22 +157,18 @@
            ;; (2) conditionally make CELL point to the symbol itself
            (compute-virtual-symbol ()
              `(progn
-                (inst mov (reg-in-size cell :dword) (tls-index-of symbol))
+                (inst mov :dword cell (tls-index-of symbol))
                 (inst lea cell
-                      (make-ea :qword :base thread-base-tn :index cell
-                               :disp (- other-pointer-lowtag
-                                        (ash symbol-value-slot word-shift))))
-                (inst cmp (access-value-slot cell :dword) ; TLS reference
+                      (ea (- other-pointer-lowtag (ash symbol-value-slot word-shift))
+                          thread-base-tn cell))
+                (inst cmp :dword (symbol-value-slot-ea cell) ; TLS reference
                       no-tls-value-marker-widetag)
                 (inst cmov :e cell symbol))) ; now possibly get the symbol
            (access-wired-tls-val (sym) ; SYM is a symbol
              `(thread-tls-ea (load-time-tls-offset ,sym)))
-           (access-value-slot (sym &optional (size :qword)) ; SYM is a TN
-             (ecase size
-               (:dword `(make-ea-for-object-slot-half
-                         ,sym symbol-value-slot other-pointer-lowtag))
-               (:qword `(make-ea-for-object-slot
-                         ,sym symbol-value-slot other-pointer-lowtag)))))
+           (symbol-value-slot-ea (sym) ; SYM is a TN
+             `(ea (- (* symbol-value-slot n-word-bytes) other-pointer-lowtag)
+                  ,sym)))
 
   (define-vop (%compare-and-swap-symbol-value)
     (:translate %compare-and-swap-symbol-value)
@@ -179,7 +177,7 @@
            (new :scs (descriptor-reg any-reg)))
     (:temporary (:sc descriptor-reg :offset rax-offset
                  :from (:argument 1) :to (:result 0)) rax)
-    #!+sb-thread
+    #+sb-thread
     (:temporary (:sc descriptor-reg :to (:result 0)) cell)
     (:results (result :scs (descriptor-reg any-reg)))
     (:policy :fast-safe)
@@ -192,13 +190,13 @@
     ;; tls-index=0, because that would succeed, assigning NEW to each
     ;; symbol in existence having otherwise no thread-local value.
       (let ((unbound (generate-error-code vop 'unbound-symbol-error symbol)))
-        #!+sb-thread (progn (compute-virtual-symbol)
+        #+sb-thread (progn (compute-virtual-symbol)
                             (move rax old)
-                            (inst cmpxchg (access-value-slot cell) new :lock))
-        #!-sb-thread (progn (move rax old)
+                            (inst cmpxchg :lock (symbol-value-slot-ea cell) new))
+        #-sb-thread (progn (move rax old)
                             ;; is the :LOCK is necessary?
-                            (inst cmpxchg (access-value-slot symbol) new :lock))
-        (inst cmp (reg-in-size rax :dword) unbound-marker-widetag)
+                            (inst cmpxchg :lock (symbol-value-slot-ea symbol) new))
+        (inst cmp :byte rax unbound-marker-widetag)
         (inst jmp :e unbound)
         (move result rax))))
 
@@ -213,14 +211,14 @@
     (:policy :fast-safe)
     (:generator 10
       (move rax old)
-      (inst cmpxchg
+      (inst cmpxchg :lock
             (if (sc-is symbol immediate)
                 (symbol-slot-ea (tn-value symbol) symbol-value-slot)
-                (access-value-slot symbol))
-            new :lock)
+                (symbol-value-slot-ea symbol))
+            new)
       (move result rax)))
 
-  #!+sb-thread
+  #+sb-thread
   (progn
     ;; TODO: SET could be shorter for any known wired-tls symbol.
     (define-vop (set)
@@ -232,7 +230,7 @@
         ;; a register, so we can't conditionally move into the TLS and
         ;; conditionally move in the opposite flag sense to the symbol.
         (compute-virtual-symbol)
-        (gen-cell-set (access-value-slot cell) value nil)))
+        (gen-cell-set (symbol-value-slot-ea cell) value nil)))
 
     ;; This code is tested by 'codegen.impure.lisp'
     (defun emit-symeval (value symbol symbol-reg check-boundp vop)
@@ -240,49 +238,50 @@
              (known-symbol (and known-symbol-p (tn-value symbol))))
         ;; In order from best to worst.
         (cond
-         ((symbol-always-has-tls-value-p known-symbol) ; e.g. *HANDLER-CLUSTERS*
-          (inst mov value (access-wired-tls-val known-symbol)))
-         (t
-          (cond
-            ((symbol-always-has-tls-index-p known-symbol) ; e.g. CL:*PRINT-BASE*
-             ;; Known nonzero TLS index, but possibly no per-thread value.
-             ;; The TLS value and global value can be loaded independently.
-             (inst mov value (access-wired-tls-val known-symbol))
-             (when (sc-is symbol constant)
-               (inst mov symbol-reg symbol))) ; = MOV Rxx, [RIP-N]
+          ((symbol-always-has-tls-value-p known-symbol)
+           (setq symbol-reg nil)
+           (inst mov value (access-wired-tls-val known-symbol)))
+          (t
+           (cond
+             ((symbol-always-has-tls-index-p known-symbol) ; e.g. CL:*PRINT-BASE*
+              ;; Known nonzero TLS index, but possibly no per-thread value.
+              ;; The TLS value and global value can be loaded independently.
+              (inst mov value (access-wired-tls-val known-symbol))
+              (when (sc-is symbol constant)
+                (inst mov symbol-reg symbol))) ; = MOV Rxx, [RIP-N]
 
-            (known-symbol-p ; unknown TLS index, possibly 0
-             (sc-case symbol
-               (immediate
-               ;; load the TLS index from the symbol. TODO: use [RIP-n] mode
-               ;; for immobile code to make it automatically relocatable.
-               (inst mov (reg-in-size value :dword)
-                     ;; slot index 1/2 is the high half of the header word.
-                     (symbol-slot-ea known-symbol 1/2 :dword))
-               ;; read the TLS value using that index
-               (inst mov value (thread-tls-ea value :qword)))
-               (constant
+             (known-symbol-p           ; unknown TLS index, possibly 0
+              (sc-case symbol
+                (immediate
+                 ;; load the TLS index from the symbol. TODO: use [RIP-n] mode
+                 ;; for immobile code to make it automatically relocatable.
+                 (inst mov :dword value
+                       ;; slot index 1/2 is the high half of the header word.
+                       (symbol-slot-ea known-symbol 1/2))
+                 ;; read the TLS value using that index
+                 (inst mov value (thread-tls-ea value)))
+                (constant
 
-               ;; These reads are inextricably data-dependent
-               (inst mov symbol-reg symbol) ; = MOV REG, [RIP-N]
-               (inst mov (reg-in-size value :dword) (tls-index-of symbol-reg))
-               (inst mov value (thread-tls-ea value :qword)))))
+                 ;; These reads are inextricably data-dependent
+                 (inst mov symbol-reg symbol) ; = MOV REG, [RIP-N]
+                 (inst mov :dword value (tls-index-of symbol-reg))
+                 (inst mov value (thread-tls-ea value)))))
 
-            (t ; SYMBOL-VALUE of a random symbol
-             (inst mov (reg-in-size symbol-reg :dword) (tls-index-of symbol))
-             (inst mov value (thread-tls-ea symbol-reg :qword))
-             (setq symbol-reg symbol)))
+             (t                      ; SYMBOL-VALUE of a random symbol
+              (inst mov :dword symbol-reg (tls-index-of symbol))
+              (inst mov value (thread-tls-ea symbol-reg))
+              (setq symbol-reg symbol)))
 
-          ;; Load the global value if the TLS value didn't exist
-          (inst cmp (reg-in-size value :dword) no-tls-value-marker-widetag)
-          (inst cmov :e value
-                (if (and known-symbol-p (sc-is symbol immediate))
-                    (symbol-slot-ea known-symbol symbol-value-slot) ; MOV Rxx, imm32
-                    (access-value-slot symbol-reg)))))
+           ;; Load the global value if the TLS value didn't exist
+           (inst cmp :dword value no-tls-value-marker-widetag)
+           (inst cmov :e value
+                 (if (and known-symbol-p (sc-is symbol immediate))
+                     (symbol-slot-ea known-symbol symbol-value-slot) ; MOV Rxx, imm32
+                     (symbol-value-slot-ea symbol-reg)))))
 
         (when check-boundp
           (assemble ()
-            (inst cmp (reg-in-size value :dword) unbound-marker-widetag)
+            (inst cmp :byte value unbound-marker-widetag)
             (let* ((immediatep (sc-is symbol immediate))
                    (staticp (and immediatep (static-symbol-p known-symbol)))
                    (*location-context* (make-restart-location RETRY value)))
@@ -296,7 +295,7 @@
                                 (lambda ()
                                   (load-immediate vop symbol symbol-reg)))
                             vop 'unbound-symbol-error
-                            (if (and immediatep (not staticp))
+                            (if (or (not symbol-reg) (and immediatep (not staticp)))
                                 symbol
                                 symbol-reg))))
             RETRY))))
@@ -332,17 +331,17 @@
       (:policy :fast-safe)
       (:args (object :scs (descriptor-reg)))
       (:conditional :ne)
-      (:temporary (:sc dword-reg) temp)
+      (:temporary (:sc unsigned-reg) temp)
       (:generator 9
-        (inst mov temp (tls-index-of object))
-        (inst mov temp (thread-tls-ea temp :dword))
-        (inst cmp temp no-tls-value-marker-widetag)
-        (inst cmov :e temp (access-value-slot object :dword))
-        (inst cmp temp unbound-marker-widetag))))
+        (inst mov :dword temp (tls-index-of object))
+        (inst mov :dword temp (thread-tls-ea temp))
+        (inst cmp :dword temp no-tls-value-marker-widetag)
+        (inst cmov :dword :e temp (symbol-value-slot-ea object))
+        (inst cmp :byte temp unbound-marker-widetag))))
 
 ) ; END OF MACROLET
 
-#!-sb-thread
+#-sb-thread
 (progn
   (define-vop (symbol-value symbol-global-value)
     (:translate symeval))
@@ -355,7 +354,7 @@
     (:args (symbol :scs (descriptor-reg)))
     (:conditional :ne)
     (:generator 9
-      (inst cmp (make-ea-for-object-slot-half
+      (inst cmp :byte (object-slot-ea
                  symbol symbol-value-slot other-pointer-lowtag)
             unbound-marker-widetag))))
 
@@ -365,12 +364,79 @@
   (:args (symbol :scs (descriptor-reg)))
   (:results (res :scs (any-reg)))
   (:result-types positive-fixnum)
+  (:args-var args)
   (:generator 2
+    (loadw res symbol symbol-hash-slot other-pointer-lowtag)
     ;; The symbol-hash slot of NIL holds NIL because it is also the
-    ;; cdr slot, so we have to zero the fixnum tag bit(s) to make sure
+    ;; car slot, so we have to zero the fixnum tag bit(s) to make sure
     ;; it is a fixnum.  The lowtag selection magic that is required to
     ;; ensure this is explained in the comment in objdef.lisp
+    (unless (not-nil-tn-ref-p args)
+      (inst and res (lognot fixnum-tag-mask)))))
+
+;;; Combine SYMBOL-HASH and the lisp fallback code into one vop.
+(define-vop ()
+  (:policy :fast-safe)
+  (:translate ensure-symbol-hash)
+  (:args (symbol :scs (descriptor-reg)))
+  (:results (res :scs (any-reg)))
+  (:result-types positive-fixnum)
+  (:vop-var vop)
+  (:generator 5
+    (when (location= res symbol) (move temp-reg-tn symbol)) ; save a backup
     (loadw res symbol symbol-hash-slot other-pointer-lowtag)
+    (inst test :dword res res)
+    (inst jmp :ne good)
+    (inst push (if (location= res symbol) temp-reg-tn symbol))
+    (invoke-asm-routine 'call 'ensure-symbol-hash vop)
+    (inst pop res)
+    GOOD
+    (inst and res (lognot fixnum-tag-mask)))) ; redundant (but ok) if asm routine used
+
+(define-vop ()
+  (:policy :fast-safe)
+  (:translate sb-impl::install-hash-table-lock)
+  (:args (arg :scs (descriptor-reg)))
+  (:results (res :scs (descriptor-reg)))
+  (:vop-var vop)
+  (:generator 5
+    (inst push arg)
+    (invoke-asm-routine 'call 'sb-impl::install-hash-table-lock vop)
+    (inst pop res)))
+
+(eval-when (:compile-toplevel)
+  ;; assumption: any object can be read 1 word past its base pointer
+  (assert (= sb-vm:symbol-hash-slot 1)))
+
+(define-vop (symbol-hash*)
+  (:policy :fast-safe)
+  (:translate symbol-hash*)
+  (:args (symbol :scs (descriptor-reg)))
+  (:info satisfies)
+  (:arg-types * (:constant (member symbolp non-null-symbol-p)))
+  (:results (res :scs (any-reg)))
+  (:result-types positive-fixnum)
+  (:generator 2
+    (loadw res symbol symbol-hash-slot other-pointer-lowtag)
+    (when (eq satisfies 'symbolp) ; mask to a fixnum
+      (inst and res (lognot fixnum-tag-mask)))))
+(define-vop (symbol-hash*-random) ; this vop needs a temp; the above doesn't
+  (:policy :fast-safe)
+  (:translate symbol-hash*)
+  (:args (object :scs (descriptor-reg)))
+  (:info satisfies)
+  ;; arg can not target the temp because they both have to be live
+  ;; in order that the tagged pointer not disappear.
+  ;; But temp and output could be in the same register.
+  (:temporary (:sc unsigned-reg :to (:result 0)) base-ptr)
+  (:arg-types * (:constant (eql nil)))
+  (:results (res :scs (any-reg)))
+  (:result-types positive-fixnum)
+  (:ignore satisfies)
+  (:generator 4
+    (inst mov base-ptr object)
+    (inst and base-ptr (lognot lowtag-mask))
+    (inst mov res (ea n-word-bytes base-ptr)) ; 1 word beyond the header
     (inst and res (lognot fixnum-tag-mask))))
 
 ;;;; fdefinition (FDEFN) objects
@@ -388,13 +454,13 @@
   (:generator 10
     (loadw value object fdefn-fun-slot other-pointer-lowtag)
     ;; byte comparison works because lowtags of function and nil differ
-    (inst cmp (reg-in-size value :byte) (logand nil-value #xff))
+    (inst cmp :byte value (logand nil-value #xff))
     (let* ((*location-context* (make-restart-location RETRY value))
            (err-lab (generate-error-code vop 'undefined-fun-error object)))
       (inst jmp :e err-lab))
     RETRY))
 
-#!-immobile-code
+#-immobile-code
 (define-vop (set-fdefn-fun)
   (:policy :fast-safe)
   (:translate (setf fdefn-fun))
@@ -404,14 +470,42 @@
   (:results (result :scs (descriptor-reg)))
   (:generator 38
     (inst mov raw (make-fixup 'closure-tramp :assembly-routine))
-    (inst cmp (make-ea :byte :base function :disp (- fun-pointer-lowtag))
+    (inst cmp :byte (ea (- fun-pointer-lowtag) function)
           simple-fun-widetag)
     (inst cmov :e raw
-          (make-ea :qword :base function
-                   :disp (- (* simple-fun-self-slot n-word-bytes) fun-pointer-lowtag)))
+          (ea (- (* simple-fun-self-slot n-word-bytes) fun-pointer-lowtag) function))
     (storew function fdefn fdefn-fun-slot other-pointer-lowtag)
     (storew raw fdefn fdefn-raw-addr-slot other-pointer-lowtag)
     (move result function)))
+#+immobile-code
+(progn
+(define-vop (set-fdefn-fun)
+  (:args (fdefn :scs (descriptor-reg))
+         (function :scs (descriptor-reg))
+         (raw-word :scs (unsigned-reg)))
+  (:vop-var vop)
+  (:generator 38
+    (pseudo-atomic ()
+      (inst push fdefn)
+      (invoke-asm-routine 'call 'touch-gc-card vop)
+      (inst mov (ea (- (ash fdefn-fun-slot word-shift) other-pointer-lowtag) fdefn)
+            function)
+      (inst mov (ea (- (ash fdefn-raw-addr-slot word-shift) other-pointer-lowtag) fdefn)
+            raw-word)
+      ;; Ensure that the header contains a JMP instruction, not INT3.
+      ;; This store is aligned
+      (inst mov :word (ea (- 2 other-pointer-lowtag) fdefn) #x25FF))))
+(define-vop (set-undefined-fdefn-fun)
+  ;; Do not set the raw-addr slot and do not change the header
+  (:args (fdefn :scs (descriptor-reg))
+         (function :scs (descriptor-reg)))
+  (:vop-var vop)
+  (:generator 38
+    (pseudo-atomic ()
+      (inst push fdefn)
+      (invoke-asm-routine 'call 'touch-gc-card vop)
+      (inst mov (ea (- (ash fdefn-fun-slot word-shift) other-pointer-lowtag) fdefn)
+            function)))))
 
 (define-vop (fdefn-makunbound)
   (:policy :fast-safe)
@@ -419,28 +513,22 @@
   (:args (fdefn :scs (descriptor-reg) :target result))
   (:results (result :scs (descriptor-reg)))
   (:vop-var vop)
-  #!+immobile-code (:temporary (:sc unsigned-reg) temp)
   (:generator 38
-    #!+immobile-code
-    (let ((tramp (make-fixup 'undefined-fdefn :assembly-routine)))
-     (if (sb!c::code-immobile-p vop)
-         (inst lea temp (make-ea :qword :base rip-tn :disp tramp))
-         (inst mov temp tramp))
-     ;; Compute displacement from the call site
-     (inst sub (reg-in-size temp :dword) (reg-in-size fdefn :dword))
-     (inst sub (reg-in-size temp :dword)
-           (+ (- other-pointer-lowtag) (ash fdefn-raw-addr-slot word-shift) 5))
-     ;; Compute the encoding of a "CALL rel32" instruction
-     (inst shl temp 8)
-     (inst or (reg-in-size temp :byte) #xE8)
-     ;; Store
-     (storew nil-value fdefn fdefn-fun-slot other-pointer-lowtag)
-     (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag))
-    #!-immobile-code
-    (progn
-     (storew nil-value fdefn fdefn-fun-slot other-pointer-lowtag)
-     (storew (make-fixup 'undefined-tramp :assembly-routine)
-             fdefn fdefn-raw-addr-slot other-pointer-lowtag))
+    ;; Change the JMP instruction to INT3 so that a trap occurs in the fdefn
+    ;; itself, otherwise we've no way of knowing what function name was invoked.
+    (inst mov :word (ea (- 2 other-pointer-lowtag) fdefn)
+          (logand undefined-fdefn-header #xFFFF))
+    ;; Once the opcode is written, the values in 'fun' and 'raw-addr' become irrelevant.
+    ;; These stores act primarily to clear the reference from a GC perspective.
+    (storew nil-value fdefn fdefn-fun-slot other-pointer-lowtag)
+    ;; With #+immobile-code we never call via the raw-addr slot for undefined
+    ;; functions if the single instruction "call <fdefn>" form is used. The INT3
+    ;; raises sigtrap which we catch, then load RAX with the address of the fdefn
+    ;; and resume at undefined-tramp. However, CALL-SYMBOL jumps via raw-addr if
+    ;; its callable object was not a function. In that case RAX holds a symbol,
+    ;; so we're OK because we can identify the undefined function.
+    (storew (make-fixup 'undefined-tramp :assembly-routine)
+            fdefn fdefn-raw-addr-slot other-pointer-lowtag)
     (move result fdefn)))
 
 ;;;; binding and unbinding
@@ -450,7 +538,7 @@
 ;;; symbol.
 ;;; See the "Chapter 9: Specials" of the SBCL Internals Manual.
 
-#!+sb-thread
+#+sb-thread
 (progn
 (define-vop (dynbind) ; bind a symbol in a PROGV form
   (:args (val :scs (any-reg descriptor-reg))
@@ -460,10 +548,10 @@
   (:vop-var vop)
   (:generator 10
     (load-binding-stack-pointer bsp)
-    (inst mov (reg-in-size tls-index :dword) (tls-index-of symbol))
+    (inst mov :dword tls-index (tls-index-of symbol))
     (inst add bsp (* binding-size n-word-bytes))
     (store-binding-stack-pointer bsp)
-    (inst test (reg-in-size tls-index :dword) (reg-in-size tls-index :dword))
+    (inst test :dword tls-index tls-index)
     (inst jmp :ne TLS-INDEX-VALID)
     (inst mov tls-index symbol)
     (invoke-asm-routine 'call 'alloc-tls-index vop)
@@ -474,7 +562,10 @@
     (inst mov (thread-tls-ea tls-index) val)))
 
 (define-vop (bind) ; bind a known symbol
-  (:args (val :scs (any-reg descriptor-reg)))
+  (:args (val :scs (any-reg descriptor-reg)
+              :load-if (not (let ((imm (encode-value-if-immediate val)))
+                              (or (fixup-p imm)
+                                  (plausible-signed-imm32-operand-p imm))))))
   (:temporary (:sc unsigned-reg) bsp tmp)
   (:info symbol)
   (:generator 10
@@ -489,26 +580,25 @@
       (storew tmp bsp binding-value-slot)
       ;; Indices are small enough to be written as :DWORDs which avoids
       ;; a REX prefix if 'bsp' happens to be any of the low 8 registers.
-      (inst mov (make-ea :dword :base bsp
-                         :disp (ash binding-symbol-slot word-shift)) tls-index)
-      (inst mov tls-cell val)))))
+      (inst mov :dword (ea (ash binding-symbol-slot word-shift) bsp) tls-index)
+      (inst mov :qword tls-cell (encode-value-if-immediate val))))))
 
-#!-sb-thread
+#-sb-thread
 (define-vop (dynbind)
   (:args (val :scs (any-reg descriptor-reg))
          (symbol :scs (descriptor-reg)))
   (:temporary (:sc unsigned-reg) temp bsp)
   (:generator 5
-    (load-symbol-value bsp *binding-stack-pointer*)
+    (load-binding-stack-pointer bsp)
     (loadw temp symbol symbol-value-slot other-pointer-lowtag)
     (inst add bsp (* binding-size n-word-bytes))
-    (store-symbol-value bsp *binding-stack-pointer*)
+    (store-binding-stack-pointer bsp)
     (storew temp bsp (- binding-value-slot binding-size))
     (storew symbol bsp (- binding-symbol-slot binding-size))
     (storew val symbol symbol-value-slot other-pointer-lowtag)))
 
-#!+sb-thread
-(define-vop (unbind)
+#+sb-thread
+(define-vop (unbind-n)
   (:temporary (:sc unsigned-reg) temp bsp)
   (:temporary (:sc complex-double-reg) zero)
   (:info symbols)
@@ -525,27 +615,25 @@
           (loadw temp bsp binding-value-slot)
           (inst mov tls-cell temp)
           ;; Zero out the stack.
-          (inst movapd (make-ea :qword :base bsp) zero))
+          (inst movapd (ea bsp) zero))
     (store-binding-stack-pointer bsp)))
 
-#!-sb-thread
+#-sb-thread
 (define-vop (unbind)
   (:temporary (:sc unsigned-reg) symbol value bsp)
   (:generator 0
-    (load-symbol-value bsp *binding-stack-pointer*)
+    (load-binding-stack-pointer bsp)
     (loadw symbol bsp (- binding-symbol-slot binding-size))
     (loadw value bsp (- binding-value-slot binding-size))
     (storew value symbol symbol-value-slot other-pointer-lowtag)
     (storew 0 bsp (- binding-symbol-slot binding-size))
     (storew 0 bsp (- binding-value-slot binding-size))
     (inst sub bsp (* binding-size n-word-bytes))
-    (store-symbol-value bsp *binding-stack-pointer*)))
+    (store-binding-stack-pointer bsp)))
 
-(define-vop (unbind-to-here)
-  (:args (where :scs (descriptor-reg any-reg)))
-  (:temporary (:sc unsigned-reg) symbol value bsp)
-  (:temporary (:sc complex-double-reg) zero)
-  (:generator 0
+(defun unbind-to-here (where symbol value bsp
+                       zero)
+  (assemble ()
     (load-binding-stack-pointer bsp)
     (inst cmp where bsp)
     (inst jmp :e DONE)
@@ -554,30 +642,36 @@
     (inst sub bsp (* binding-size n-word-bytes))
     ;; on sb-thread symbol is actually a tls-index, and it fits into
     ;; 32-bits.
-    #!+sb-thread
-    (let ((tls-index (reg-in-size symbol :dword)))
-      (inst mov tls-index
-            (make-ea :dword :base bsp :disp (* binding-symbol-slot n-word-bytes)))
-      (inst test tls-index tls-index))
-    #!-sb-thread
+    #+sb-thread
+    (progn
+      (inst mov :dword symbol (ea (* binding-symbol-slot n-word-bytes) bsp))
+      (inst test :dword symbol symbol))
+    #-sb-thread
     (progn
       (loadw symbol bsp binding-symbol-slot)
       (inst test symbol symbol))
     (inst jmp :z SKIP)
     (loadw value bsp binding-value-slot)
-    #!-sb-thread
+    #-sb-thread
     (storew value symbol symbol-value-slot other-pointer-lowtag)
-    #!+sb-thread
+    #+sb-thread
     (inst mov (thread-tls-ea symbol) value)
 
     SKIP
-    (inst movapd (make-ea :qword :base bsp) zero)
+    (inst movapd (ea bsp) zero)
 
     (inst cmp where bsp)
     (inst jmp :ne LOOP)
     (store-binding-stack-pointer bsp)
 
     DONE))
+
+(define-vop (unbind-to-here)
+  (:args (where :scs (descriptor-reg any-reg)))
+  (:temporary (:sc unsigned-reg) symbol value bsp)
+  (:temporary (:sc complex-double-reg) zero)
+  (:generator 0
+    (unbind-to-here where symbol value bsp zero)))
 
 ;;;; closure indexing
 
@@ -623,48 +717,18 @@
 
 ;;;; structure hackery
 
-(define-vop (instance-length)
+(define-vop ()
   (:policy :fast-safe)
   (:translate %instance-length)
   (:args (struct :scs (descriptor-reg)))
-  ;; Explicitly fixnumizing admits a 32-bit left-shift which might encode
-  ;; in 1 less byte. (Other backends return an UNSIGNED-REG here.)
   (:results (res :scs (any-reg)))
   (:result-types positive-fixnum)
   (:generator 4
-    (let ((res (reg-in-size res :dword)))
-      (inst movzx res (make-ea :word :base struct
-                               :disp (1+ (- instance-pointer-lowtag))))
-      (inst shl res n-fixnum-tag-bits))))
-
-#!+compact-instance-header
-(progn
- (define-vop (%instance-layout)
-   (:translate %instance-layout)
-   (:policy :fast-safe)
-   (:args (object :scs (descriptor-reg)))
-   (:results (res :scs (descriptor-reg)))
-   (:variant-vars lowtag)
-   (:variant instance-pointer-lowtag)
-   (:generator 1
-    (inst mov (reg-in-size res :dword) (make-ea :dword :base object :disp (- 4 lowtag)))))
- (define-vop (%set-instance-layout)
-   (:translate %set-instance-layout)
-   (:policy :fast-safe)
-   (:args (object :scs (descriptor-reg))
-          (value :scs (any-reg descriptor-reg) :target res))
-   (:results (res :scs (any-reg descriptor-reg)))
-   (:variant-vars lowtag)
-   (:variant instance-pointer-lowtag)
-   (:generator 2
-    (inst mov (make-ea :dword :base object :disp (- 4 lowtag)) (reg-in-size value :dword))
-    (move res value)))
- (define-vop (%funcallable-instance-layout %instance-layout)
-   (:translate %funcallable-instance-layout)
-   (:variant fun-pointer-lowtag))
- (define-vop (%set-funcallable-instance-layout %set-instance-layout)
-   (:translate %set-funcallable-instance-layout)
-   (:variant fun-pointer-lowtag)))
+    (inst mov :dword res (ea (- instance-pointer-lowtag) struct))
+    ;; Returning fixnum/any-reg elides some REX prefixes due to the shifts
+    ;; being small. Maybe the asm optimizer could figure it out now?
+    (inst shr :dword res (- instance-length-shift n-fixnum-tag-bits))
+    (inst and :dword res (fixnumize instance-length-mask))))
 
 (define-full-reffer instance-index-ref * instance-slots-offset
   instance-pointer-lowtag (any-reg descriptor-reg) * %instance-ref)
@@ -684,7 +748,7 @@
 (define-full-reffer code-header-ref * 0 other-pointer-lowtag
   (any-reg descriptor-reg) * code-header-ref)
 
-;; CODE-HEADER-SET of a constant index is never used, so we specifically
+;; CODE-HEADER-SET of a constant index is seldom used, so we specifically
 ;; ask not to get it defined. (It's not impossible to use, it's just not
 ;; helpful to have - either you want to access CODE-FIXUPS or CODE-DEBUG-INFO,
 ;; which do things their own way, or you want a variable index.)
@@ -694,22 +758,18 @@
 
 ;;;; raw instance slot accessors
 
-(flet ((make-ea-for-raw-slot (object index)
+(flet ((instance-slot-ea (object index)
          (etypecase index
            (integer
-              (make-ea :qword
-                       :base object
-                       :disp (+ (* (+ instance-slots-offset index) n-word-bytes)
-                                (- instance-pointer-lowtag))))
+              (ea (+ (* (+ instance-slots-offset index) n-word-bytes)
+                     (- instance-pointer-lowtag))
+                  object))
            (tn
-              (make-ea :qword
-                       :base object
-                       :index index
-                       :scale (ash 1 (- word-shift n-fixnum-tag-bits))
-                       :disp (+ (* instance-slots-offset n-word-bytes)
-                                (- instance-pointer-lowtag)))))))
+              (ea (+ (* instance-slots-offset n-word-bytes)
+                     (- instance-pointer-lowtag))
+                  object index (ash 1 (- word-shift n-fixnum-tag-bits)))))))
   (macrolet
-      ((def (suffix result-sc result-type inst &optional (inst/c inst))
+      ((def (suffix result-sc result-type inst &optional (inst/c (list 'quote inst)))
          `(progn
             (define-vop (,(symbolicate "RAW-INSTANCE-REF/" suffix))
               (:translate ,(symbolicate "%RAW-INSTANCE-REF/" suffix))
@@ -719,7 +779,7 @@
               (:results (value :scs (,result-sc)))
               (:result-types ,result-type)
               (:generator 5
-                (inst ,inst value (make-ea-for-raw-slot object index))))
+                (inst ,inst value (instance-slot-ea object index))))
             (define-vop (,(symbolicate "RAW-INSTANCE-REF-C/" suffix))
               (:translate ,(symbolicate "%RAW-INSTANCE-REF/" suffix))
               (:policy :fast-safe)
@@ -733,7 +793,7 @@
               (:results (value :scs (,result-sc)))
               (:result-types ,result-type)
               (:generator 4
-                (inst ,inst/c value (make-ea-for-raw-slot object index))))
+                (inst* ,inst/c value (instance-slot-ea object index))))
             (define-vop (,(symbolicate "RAW-INSTANCE-SET/" suffix))
               (:translate ,(symbolicate "%RAW-INSTANCE-SET/" suffix))
               (:policy :fast-safe)
@@ -744,7 +804,7 @@
               (:results (result :scs (,result-sc)))
               (:result-types ,result-type)
               (:generator 5
-                (inst ,inst (make-ea-for-raw-slot object index) value)
+                (inst ,inst (instance-slot-ea object index) value)
                 (move result value)))
             (define-vop (,(symbolicate "RAW-INSTANCE-SET-C/" suffix))
               (:translate ,(symbolicate "%RAW-INSTANCE-SET/" suffix))
@@ -759,7 +819,7 @@
               (:results (result :scs (,result-sc)))
               (:result-types ,result-type)
               (:generator 4
-                (inst ,inst/c (make-ea-for-raw-slot object index) value)
+                (inst* ,inst/c (instance-slot-ea object index) value)
                 (move result value)))
             (define-vop (,(symbolicate "RAW-INSTANCE-INIT/" suffix))
               (:args (object :scs (descriptor-reg))
@@ -767,7 +827,7 @@
               (:arg-types * ,result-type)
               (:info index)
               (:generator 4
-                (inst ,inst/c (make-ea-for-raw-slot object index) value))))))
+                (inst* ,inst/c (instance-slot-ea object index) value))))))
     (def word unsigned-reg unsigned-num mov)
     (def signed-word signed-reg signed-num mov)
     (def single single-reg single-float movss)
@@ -786,7 +846,7 @@
     (:results (result :scs (unsigned-reg)))
     (:result-types unsigned-num)
     (:generator 5
-      (inst xadd (make-ea-for-raw-slot object index) diff :lock)
+      (inst xadd :lock (instance-slot-ea object index) diff)
       (move result diff)))
 
   (define-vop (raw-instance-atomic-incf-c/word)
@@ -802,7 +862,7 @@
     (:results (result :scs (unsigned-reg)))
     (:result-types unsigned-num)
     (:generator 4
-      (inst xadd (make-ea-for-raw-slot object index) diff :lock)
+      (inst xadd :lock (instance-slot-ea object index) diff)
       (move result diff))))
 
 ;;;;
@@ -818,7 +878,7 @@
   (move rdx old-hi)
   (move rbx new-lo)
   (move rcx new-hi)
-  (inst cmpxchg16b memory-operand :lock)
+  (inst cmpxchg16b :lock memory-operand)
   ;; RDX:RAX hold the actual old contents of memory.
   ;; Manually analyze result lifetimes to avoid clobbering.
   (cond ((and (location= result-lo rdx) (location= result-hi rax))
@@ -843,24 +903,23 @@
                  (new-hi :scs (descriptor-reg any-reg) :target ecx))
           (:results (result-lo :scs (descriptor-reg any-reg))
                     (result-hi :scs (descriptor-reg any-reg)))
-          (:temporary (:sc unsigned-reg :offset eax-offset
+          (:temporary (:sc unsigned-reg :offset rax-offset
                        :from (:argument 2) :to (:result 0)) eax)
-          (:temporary (:sc unsigned-reg :offset edx-offset
+          (:temporary (:sc unsigned-reg :offset rdx-offset
                        :from (:argument 3) :to (:result 0)) edx)
-          (:temporary (:sc unsigned-reg :offset ebx-offset
+          (:temporary (:sc unsigned-reg :offset rbx-offset
                        :from (:argument 4) :to (:result 0)) ebx)
-          (:temporary (:sc unsigned-reg :offset ecx-offset
+          (:temporary (:sc unsigned-reg :offset rcx-offset
                        :from (:argument 5) :to (:result 0)) ecx)
           (:generator 7
            (generate-dblcas ,memory-operand
                             expected-old-lo expected-old-hi new-lo new-hi
                             eax ebx ecx edx result-lo result-hi)))))
   (define-cmpxchg-vop compare-and-exchange-pair
-      (make-ea :dword :base object :disp (- list-pointer-lowtag))
+      (ea (- list-pointer-lowtag) object)
       ((:translate %cons-cas-pair)))
   (define-cmpxchg-vop compare-and-exchange-pair-indexed
-      (make-ea :dword :base object :disp offset :index index
-                      :scale (ash n-word-bytes (- n-fixnum-tag-bits)))
+      (ea offset object index (ash n-word-bytes (- n-fixnum-tag-bits)))
       ((:variant-vars offset))
       ((index :scs (descriptor-reg any-reg) :to :eval))))
 
