@@ -16,8 +16,7 @@
 (defun generate-fixnum-test (value)
   "Set the Z flag if VALUE is fixnum"
   (inst test :byte
-        (cond ;; This is hooey. None of the type-vops presently allow
-              ;; control-stack as a storage class.
+        (cond ((ea-p value) value) ; merged a memory load + fixnump vop
               ((sc-is value control-stack)
                (ea (frame-byte-offset (tn-offset value)) rbp-tn))
               (t
@@ -131,6 +130,8 @@
                                   (= (cdar headers) complex-array-widetag)))
                          (ea (- lowtag) value)
                          temp))
+         (first (car headers))
+         (second (cadr headers))
          (untagged))
 
     (multiple-value-bind (equal less-or-equal greater-or-equal when-true
@@ -176,45 +177,56 @@
         (inst cmp :byte temp widetag)
         (inst jmp :e when-false))
 
-      ;; FIXME: this backend seems to be missing the special logic for
-      ;;        testing exactly two widetags differing only in a single bit,
-      ;;        which through evolution is almost totally unworkable anyway...
+      (cond
+       ((and (fixnump first)
+             (fixnump second)
+             (not (cddr headers))
+             (= (logcount (logxor first second)) 1))
+        ;; Two widetags differing at one bit. Use one cmp and branch.
+        ;; Start by ORing in the bit that they differ on.
+        (let ((diff-bit (logxor first second)))
+          (aver (not (ea-p widetag-tn))) ; can't clobber a header
+          (inst or :byte widetag-tn diff-bit)
+          (inst cmp :byte widetag-tn (logior first diff-bit))
+          (if not-p (inst jmp :ne target) (inst jmp :eq target))))
+       (t
       ;; Compared to x86 we additionally optimize the cases of a
       ;; range starting with BIGNUM-WIDETAG (= min widetag)
       ;; or ending with COMPLEX-ARRAY-WIDETAG (= max widetag)
-      (do ((remaining headers (cdr remaining)))
-          ((null remaining))
-        (let ((header (car remaining))
-              (last (null (cdr remaining))))
-          (cond
-           ((atom header)
-            (inst cmp :byte widetag-tn header)
-            (if last
-                (inst jmp equal target)
-                (inst jmp :e when-true)))
-           (t
-             (let ((start (car header))
-                   (end (cdr header)))
-               (cond
-                 ((= start bignum-widetag)
-                  (inst cmp :byte widetag-tn end)
-                  (if last
-                      (inst jmp less-or-equal target)
-                      (inst jmp :be when-true)))
-                 ((= end complex-array-widetag)
-                  (inst cmp :byte widetag-tn start)
-                  (if last
-                      (inst jmp greater-or-equal target)
-                      (inst jmp :b when-false)))
-                 ((not last)
-                  (inst cmp :byte temp start)
-                  (inst jmp :b when-false)
-                  (inst cmp :byte temp end)
-                  (inst jmp :be when-true))
-                 (t
-                  (inst sub :byte temp start)
-                  (inst cmp :byte temp (- end start))
-                  (inst jmp less-or-equal target))))))))
+        (do ((remaining headers (cdr remaining)))
+            ((null remaining))
+          (let ((header (car remaining))
+                (last (null (cdr remaining))))
+            (cond
+             ((atom header)
+              (inst cmp :byte widetag-tn header)
+              (if last
+                  (inst jmp equal target)
+                  (inst jmp :e when-true)))
+             (t
+               (let ((start (car header))
+                     (end (cdr header)))
+                 (cond
+                   ((= start bignum-widetag)
+                    (inst cmp :byte widetag-tn end)
+                    (if last
+                        (inst jmp less-or-equal target)
+                        (inst jmp :be when-true)))
+                   ((= end complex-array-widetag)
+                    (inst cmp :byte widetag-tn start)
+                    (if last
+                        (inst jmp greater-or-equal target)
+                        (inst jmp :b when-false)))
+                   ((not last)
+                    (inst cmp :byte temp start)
+                    (inst jmp :b when-false)
+                    (inst cmp :byte temp end)
+                    (inst jmp :be when-true))
+                   (t
+                    (inst sub :byte temp start)
+                    (inst cmp :byte temp (- end start))
+                    (inst jmp less-or-equal target))))))))))
+
       (emit-label drop-through))))
 
 ;;;; other integer ranges
@@ -243,7 +255,7 @@
     ;;    a <= x <= a + 2^n - 1
     ;; is equivalent to unsigned
     ;;    ((x-a) >> n) = 0
-    (inst mov temp #.(- sb-xc:most-negative-fixnum))
+    (inst mov temp #.(- most-negative-fixnum))
     (inst add temp value)
     (inst shr temp n-fixnum-bits)))
 
@@ -385,31 +397,85 @@
        (inst jmp (if not-p :a :be) target)
        (emit-label skip))))
 
+;;; SINGLE-FLOAT-P, CHARACTERP, UNBOUND-MARKER-P produce a flag result
+;;; and never need a temporary.
+(macrolet ((define (name widetag)
+             `(define-vop (,name simple-type-predicate)
+                (:translate ,name)
+                (:info)
+                (:conditional :z)
+                (:generator 1 (inst cmp :byte value ,widetag)))))
+  (define single-float-p single-float-widetag)
+  (define characterp character-widetag)
+  (define unbound-marker-p unbound-marker-widetag))
+
 (define-vop (pointerp)
   (:args (value :scs (any-reg descriptor-reg) :target temp))
   (:temporary (:sc unsigned-reg :from (:argument 0)) temp)
-  (:conditional)
-  (:info target not-p)
+  (:conditional :z)
   (:policy :fast-safe)
   (:translate pointerp)
   (:generator 3
     (inst lea :dword temp (ea -3 value))
-    (inst test :byte temp #b11)
-    (inst jmp (if not-p :nz :z) target)))
+    (inst test :byte temp #b11)))
+
+;;; FUNCTIONP, LISTP, %INSTANCEP, %OTHER-POINTER-P produce a flag result
+(macrolet ((define (name lowtag)
+             `(define-vop (,name pointerp)
+                (:translate ,name)
+                (:generator 2
+                  (inst lea :dword temp (ea (- ,lowtag) value))
+                  (inst test :byte temp lowtag-mask)))))
+  (define functionp fun-pointer-lowtag)
+  (define listp list-pointer-lowtag)
+  (define %instancep instance-pointer-lowtag)
+  (define %other-pointer-p other-pointer-lowtag))
+
+(macrolet ((define (name widetag)
+             `(define-vop (,name type-predicate)
+                (:translate ,name)
+                (:info)
+                (:conditional :z)
+                (:generator 4
+                  (inst lea temp (ea (- fun-pointer-lowtag) value))
+                  (inst test :byte temp lowtag-mask)
+                  (inst jmp :ne out)
+                  (inst cmp :byte (ea temp) ,widetag)
+                  out))))
+  (define closurep closure-widetag)
+  (define simple-fun-p simple-fun-widetag)
+  (define funcallable-instance-p funcallable-instance-widetag))
 
 ;;;; list/symbol types
 ;;;
 ;;; symbolp (or symbol (eq nil))
 ;;; consp (and list (not (eq nil)))
 
+(defun test-other-ptr (arg arg-ref widetag temp label)
+  (inst cmp :byte
+        (cond ((other-pointer-tn-ref-p arg-ref)
+               (ea (- other-pointer-lowtag) arg))
+              (t
+               (%lea-for-lowtag-test temp arg other-pointer-lowtag :qword)
+               (inst test :byte temp lowtag-mask)
+               (inst jmp :ne label)
+               (ea temp)))
+        widetag))
+
 (define-vop (symbolp type-predicate)
   (:translate symbolp)
-  (:generator 12
-    (let ((is-symbol-label (if not-p DROP-THRU target)))
-      (inst cmp value nil-value)
-      (inst jmp :e is-symbol-label)
-      (test-type value temp target not-p (symbol-widetag)))
-    DROP-THRU))
+  (:info)
+  (:conditional :z)
+  (:generator 5
+    ;; SYMBOLP would have been IR1-transformed to NON-NULL-SYMBOL-P when possible
+    (inst cmp value nil-value)
+    (inst jmp :e out)
+    (test-other-ptr value args symbol-widetag temp out)
+    out))
+
+(define-vop (non-null-symbol-p symbolp)
+  (:translate non-null-symbol-p)
+  (:generator 3 (test-other-ptr value args symbol-widetag temp out) out))
 
 (define-vop (consp type-predicate)
   (:translate consp)
@@ -431,41 +497,124 @@
   (:generator 2
    (inst cmp :byte (ea (- other-pointer-lowtag) x) widetag)))
 
-;;; TRANSFORM-INSTANCE-TYPEP checks for this vop by name and will try to use it,
-;;; so don't define it if inapplicable.
 #+compact-instance-header
 (progn
-(defknown layout-eq (instance t) boolean (flushable))
-(define-vop (layout-eq)
-  (:translate layout-eq)
+ (define-vop ()
+   (:translate %instance-layout)
+   (:policy :fast-safe)
+   (:args (object :scs (descriptor-reg)))
+   (:results (res :scs (descriptor-reg)))
+   (:variant-vars lowtag)
+   (:variant instance-pointer-lowtag)
+   (:generator 1
+    (inst mov :dword res (ea (- 4 lowtag) object))))
+ (define-vop ()
+   (:translate %set-instance-layout)
+   (:policy :fast-safe)
+   (:args (object :scs (descriptor-reg))
+          (value :scs (any-reg descriptor-reg) :target res))
+   (:results (res :scs (any-reg descriptor-reg)))
+   (:vop-var vop)
+   (:generator 1
+     (inst mov :dword (ea (- 4 instance-pointer-lowtag) object) value)
+     (move res value)))
+ (define-vop (%fun-layout %instance-layout)
+   (:translate %fun-layout)
+   (:variant fun-pointer-lowtag))
+ (define-vop (%set-fun-layout %set-instance-layout)
+   (:translate %set-fun-layout)
+   (:generator 50
+     (pseudo-atomic ()
+       (inst push object)
+       (invoke-asm-routine 'call 'touch-gc-card vop)
+       (inst mov :dword (ea (- 4 fun-pointer-lowtag) object) value))
+     (move res value)))
+(define-vop ()
+  (:translate sb-c::layout-eq)
   (:policy :fast-safe)
   (:conditional :e)
-  (:args (instance :scs (descriptor-reg))
+  (:args (object :scs (descriptor-reg))
          (layout :scs (descriptor-reg immediate)))
+  (:arg-types * * (:constant t))
+  (:info lowtag)
   (:generator 1
-    (inst cmp :dword
-          (ea (- 4 instance-pointer-lowtag) instance)
+    (inst cmp :dword (ea (- 4 lowtag) object)
           (if (sc-is layout immediate)
               (make-fixup (tn-value layout) :layout)
               layout)))))
 
-(defknown layout-inherits-ref-eq (simple-vector index t) boolean (flushable))
-(define-vop (layout-inherits-ref-eq)
-  (:translate layout-inherits-ref-eq)
-  (:policy :fast-safe)
-  (:conditional :e)
-  (:args (vector :scs (descriptor-reg))
-         (index :scs (any-reg descriptor-reg immediate))
-         (thing :scs (descriptor-reg immediate)))
-  (:generator 1
-    (inst cmp :dword
-          (if (sc-is index immediate)
-              (ea (+ (- other-pointer-lowtag)
-                     (ash (+ vector-data-offset (tn-value index)) word-shift))
-                  vector)
-              (ea (+ (- other-pointer-lowtag)
-                     (ash vector-data-offset word-shift))
-                  vector index (ash 1 (- word-shift n-fixnum-tag-bits))))
-          (if (sc-is thing immediate)
-              (make-fixup (tn-value thing) :layout)
-              thing))))
+;;; Return the DISP part of an EA based on MEM-OP, which is comprised
+;;; of a vop name and its codegen info.
+;;; Return NIL if the value can't be encoded in a machine instruction.
+(defun valid-memref-byte-disp (mem-op)
+  (ecase (car mem-op)
+    ((instance-index-ref-c
+      raw-instance-ref-c/word
+      raw-instance-ref-c/signed-word)
+     (destructuring-bind (index) (cdr mem-op)
+       (- (ash (+ index instance-slots-offset) word-shift)
+          instance-pointer-lowtag)))
+    (slot
+     (destructuring-bind (name index lowtag) (cdr mem-op)
+       (declare (ignore name))
+       (- (ash index word-shift) lowtag)))
+    (data-vector-ref-with-offset/simple-vector-c
+     (destructuring-bind (index offset) (cdr mem-op)
+       (let ((disp (- (ash (+ vector-data-offset index offset) word-shift)
+                      other-pointer-lowtag)))
+         (if (typep disp '(signed-byte 32)) disp))))))
+
+(define-vop (fixnump simple-type-predicate)
+  (:translate fixnump)
+  (:args-var arg-ref)
+  (:args (value :scs (any-reg descriptor-reg) :load-if (tn-ref-memory-access arg-ref)))
+  (:info)
+  (:conditional :z)
+  ;; the compiler is very sensitive to this cost here as regards boxing. DON'T TOUCH !!!
+  (:generator 3
+   (awhen (tn-ref-memory-access arg-ref)
+     (setq value (ea (cdr it) value)))
+   (generate-fixnum-test value)))
+
+(macrolet
+    ((define-simple-array-type-vops ()
+         `(progn
+           ,@(map 'list
+                  (lambda (saetp &aux (primtype (saetp-primitive-type-name saetp))
+                                      (name (symbolicate primtype "-P")))
+                    `(define-vop (,name symbolp)
+                       (:translate ,name)
+                       (:generator 4
+                        (test-other-ptr value args ,(saetp-typecode saetp) temp out)
+                        out)))
+                  *specialized-array-element-type-properties*))))
+  (define-simple-array-type-vops))
+
+;;; Try to absorb a memory load into FIXNUMP.
+(defoptimizer (sb-c::vop-optimize fixnump) (vop)
+  ;; Ensure that the fixnump vop does not try to absorb more than one memref.
+  ;; That is, if the initial IR2 matches (fixnump (memref (memref))) which is simplified
+  ;; to (memref+fixnump (memref x)), it would seem to allow matching of the pattern
+  ;; again if this optimizer is reapplied, because the "new" fixnump vop is superficially
+  ;; the same, except for the attachment of extra data to its input.
+  (unless (tn-ref-memory-access (sb-c::vop-args vop))
+    (let ((prev (sb-c::previous-vop-is
+                 vop
+                 '(instance-index-ref-c slot
+                   data-vector-ref-with-offset/simple-vector-c))))
+      (aver (not (sb-c::vop-results vop))) ; is a :CONDITIONAL vop
+      (when (and prev (eq (vop-block prev) (vop-block vop)))
+        (let ((arg (sb-c::vop-args vop)))
+          (when (and (eq (tn-ref-tn (sb-c::vop-results prev)) (tn-ref-tn arg))
+                     (sb-c::very-temporary-p (tn-ref-tn arg)))
+            (binding* ((mem-op (cons (vop-name prev) (vop-codegen-info prev)))
+                       (disp (valid-memref-byte-disp mem-op) :exit-if-null)
+                       (arg-ref
+                        (sb-c::reference-tn (tn-ref-tn (sb-c::vop-args prev)) nil))
+                       (new (sb-c::emit-and-insert-vop
+                             (sb-c::vop-node vop) (vop-block vop) (sb-c::vop-info vop)
+                             arg-ref nil prev (vop-codegen-info vop))))
+              (setf (tn-ref-memory-access arg-ref) `(:read . ,disp))
+              (sb-c::delete-vop prev)
+              (sb-c::delete-vop vop)
+              new)))))))
