@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 ;;;; allocator for the array header
 
@@ -26,37 +26,18 @@
   (:generator 13
     (inst lea bytes
           (make-ea :dword :base rank
-                   :disp (+ (* (1+ array-dimensions-offset) n-word-bytes)
+                   :disp (+ (* array-dimensions-offset n-word-bytes)
                             lowtag-mask)))
     (inst and bytes (lognot lowtag-mask))
-    (inst lea header (make-ea :dword :base rank
-                              :disp (fixnumize (1- array-dimensions-offset))))
-    (inst shl header n-widetag-bits)
+    ;; rank 1 is stored as 0, 2 is stored as 1, ...
+    (inst lea header (make-ea :dword :disp (fixnumize -1) :base rank))
+    (inst and header (fixnumize array-rank-mask))
+    (inst shl header array-rank-byte-pos)
     (inst or  header type)
-    (inst shr header 2)
-    (pseudo-atomic
-     (allocation result bytes node nil other-pointer-lowtag)
+    (inst shr header n-fixnum-tag-bits)
+    (pseudo-atomic ()
+     (allocation nil bytes other-pointer-lowtag node nil result)
      (storew header result 0 other-pointer-lowtag))))
-
-(define-vop (make-array-header/c)
-  (:translate make-array-header)
-  (:policy :fast-safe)
-  (:arg-types (:constant t) (:constant t))
-  (:info type rank)
-   (:results (result :scs (descriptor-reg)))
-  (:node-var node)
-  (:generator 12
-    (let* ((header-size (+ rank
-                           (1- array-dimensions-offset)))
-           (bytes (logandc2 (+ (* (1+ header-size) n-word-bytes)
-                               lowtag-mask)
-                            lowtag-mask))
-           (header (logior (ash header-size
-                                n-widetag-bits)
-                           type)))
-     (pseudo-atomic
-      (allocation result bytes node nil other-pointer-lowtag)
-      (storew header result 0 other-pointer-lowtag)))))
 
 ;;;; additional accessors and setters for the array header
 (define-full-reffer %array-dimension *
@@ -67,82 +48,120 @@
   array-dimensions-offset other-pointer-lowtag
   (any-reg) positive-fixnum %set-array-dimension)
 
-(define-vop (array-rank-vop)
+(define-vop ()
   (:translate %array-rank)
   (:policy :fast-safe)
   (:args (x :scs (descriptor-reg)))
   (:results (res :scs (unsigned-reg)))
   (:result-types positive-fixnum)
   (:generator 6
-    (loadw res x 0 other-pointer-lowtag)
-    (inst shr res n-widetag-bits)
-    (inst sub res (1- array-dimensions-offset))))
+    (inst movzx res (make-ea :byte :disp (- 2 other-pointer-lowtag) :base x))
+    ;; not all registers have an addressable low byte,
+    ;; so the simple trick used on x86-64 won't work.
+    (inst inc res)
+    (inst and res array-rank-mask)))
+
+(define-vop ()
+  (:translate %array-rank=)
+  (:policy :fast-safe)
+  (:args (array :scs (descriptor-reg)))
+  (:info rank)
+  (:arg-types * (:constant t))
+  (:conditional :e)
+  (:generator 2
+    (inst cmp (make-ea :byte :disp (- 2 other-pointer-lowtag) :base array)
+          (encode-array-rank rank))))
+
+(define-vop (array-vectorp simple-type-predicate)
+  ;; SIMPLE-TYPE-PREDICATE says that it takes stack locations, but that's no good.
+  (:args (array :scs (any-reg descriptor-reg)))
+  (:translate vectorp)
+  (:conditional :z)
+  (:info)
+  (:guard (lambda (node)
+            (let ((arg (car (sb-c::combination-args node))))
+              (csubtypep (sb-c::lvar-type arg) (specifier-type 'array)))))
+  (:generator 1
+    (inst cmp (make-ea :byte :disp (- 2 other-pointer-lowtag) :base array)
+          (encode-array-rank 1))))
 
 ;;;; bounds checking routine
-
-;;; Note that the immediate SC for the index argument is disabled
-;;; because it is not possible to generate a valid error code SC for
-;;; an immediate value.
-;;;
-;;; FIXME: As per the KLUDGE note explaining the :IGNORE-FAILURE-P
-;;; flag in build-order.lisp-expr, compiling this file causes warnings
-;;;    Argument FOO to VOP CHECK-BOUND has SC restriction
-;;;    DESCRIPTOR-REG which is not allowed by the operand type:
-;;;      (:OR POSITIVE-FIXNUM)
-;;; CSR's message "format ~/ /" on sbcl-devel 2002-03-12 contained
-;;; a possible patch, described as
-;;;   Another patch is included more for information than anything --
-;;;   removing the descriptor-reg SCs from the CHECK-BOUND vop in
-;;;   x86/array.lisp seems to allow that file to compile without error[*],
-;;;   and build; I haven't tested rebuilding capability, but I'd be
-;;;   surprised if there were a problem.  I'm not certain that this is the
-;;;   correct fix, though, as the restrictions on the arguments to the VOP
-;;;   aren't the same as in the sparc and alpha ports, where, incidentally,
-;;;   the corresponding file builds without error currently.
-;;; Since neither of us (CSR or WHN) was quite sure that this is the
-;;; right thing, I've just recorded the patch here in hopes it might
-;;; help when someone attacks this problem again:
-;;;   diff -u -r1.7 array.lisp
-;;;   --- src/compiler/x86/array.lisp 11 Oct 2001 14:05:26 -0000      1.7
-;;;   +++ src/compiler/x86/array.lisp 12 Mar 2002 12:23:37 -0000
-;;;   @@ -76,10 +76,10 @@
-;;;      (:translate %check-bound)
-;;;      (:policy :fast-safe)
-;;;      (:args (array :scs (descriptor-reg))
-;;;   -        (bound :scs (any-reg descriptor-reg))
-;;;   -        (index :scs (any-reg descriptor-reg #+nil immediate) :target result))
-;;;   +        (bound :scs (any-reg))
-;;;   +        (index :scs (any-reg #+nil immediate) :target result))
-;;;      (:arg-types * positive-fixnum tagged-num)
-;;;   -  (:results (result :scs (any-reg descriptor-reg)))
-;;;   +  (:results (result :scs (any-reg)))
-;;;      (:result-types positive-fixnum)
-;;;      (:vop-var vop)
-;;;      (:save-p :compute-only)
 (define-vop (check-bound)
   (:translate %check-bound)
   (:policy :fast-safe)
-  (:args (array :scs (descriptor-reg))
-         (bound :scs (any-reg))
-         (index :scs (any-reg #+nil immediate) :target result))
-  (:arg-types * positive-fixnum tagged-num)
-  (:results (result :scs (any-reg)))
-  (:result-types positive-fixnum)
+  (:args (array :scs (descriptor-reg constant))
+         (bound :scs (any-reg descriptor-reg)
+                :load-if (not (and (sc-is bound immediate)
+                                   (typep (tn-value bound)
+                                          'sc-offset)
+                                   (not (sc-is index immediate)))))
+         (index :scs (any-reg descriptor-reg)
+                :load-if (not (and (sc-is index immediate)
+                                   (typep (tn-value index)
+                                          'sc-offset)))))
+  (:variant-vars %test-fixnum)
+  (:variant t)
   (:vop-var vop)
   (:save-p :compute-only)
-  (:generator 5
+  (:generator 6
     (let ((error (generate-error-code vop 'invalid-array-index-error
                                       array bound index))
+          (bound (if (sc-is bound immediate)
+                     (let ((value (tn-value bound)))
+                       (cond ((and %test-fixnum
+                                   (power-of-two-limit-p (1- value)))
+                              (lognot (fixnumize (1- value))))
+                             ((sc-is index any-reg descriptor-reg)
+                              (fixnumize value))
+                             (t
+                              value)))
+                     bound))
           (index (if (sc-is index immediate)
-                   (fixnumize (tn-value index))
-                   index)))
-      (inst cmp bound index)
-      ;; We use below-or-equal even though it's an unsigned test,
-      ;; because negative indexes appear as large unsigned numbers.
-      ;; Therefore, we get the <0 and >=bound test all rolled into one.
-      (inst jmp :be error)
-      (unless (and (tn-p index) (location= result index))
-        (inst mov result index)))))
+                     (let ((value (tn-value index)))
+                       (if (sc-is bound any-reg descriptor-reg)
+                           (fixnumize value)
+                           value))
+                     index)))
+      (cond ((typep bound '(integer * -1))
+             ;; Power of two bound, can be checked for fixnumness at
+             ;; the same time as it always occupies a consecutive bit
+             ;; range, everything else, including the tag, has to be
+             ;; zero.
+             (inst test index (if (eql bound -1)
+                                  index ;; zero?
+                                  bound))
+             (inst jmp :ne error))
+            (t
+             (when (and %test-fixnum (not (integerp index)))
+               (%test-fixnum index nil error t))
+             (cond ((integerp bound)
+                    (inst cmp index bound)
+                    (inst jmp :nb error))
+                   (t
+                    (inst cmp bound index)
+                    (inst jmp :be error))))))))
+
+(define-vop (check-bound/fast check-bound)
+  (:policy :fast)
+  (:variant nil)
+  (:variant-cost 4))
+
+(define-vop (check-bound/fixnum check-bound)
+  (:args (array)
+         (bound)
+         (index :scs (any-reg)))
+  (:arg-types * * tagged-num)
+  (:variant nil)
+  (:variant-cost 4))
+
+(define-vop (check-bound/untagged check-bound)
+  (:args (array)
+         (bound :scs (unsigned-reg signed-reg))
+         (index :scs (unsigned-reg signed-reg)))
+  (:arg-types * (:or unsigned-num signed-num)
+                (:or unsigned-num signed-num))
+  (:variant nil)
+  (:variant-cost 5))
 
 ;;;; accessors/setters
 
@@ -166,7 +185,7 @@
     signed-reg)
   (def-full-data-vector-frobs simple-array-unsigned-byte-31 unsigned-num
     unsigned-reg)
-  #!+sb-unicode
+  #+sb-unicode
   (def-full-data-vector-frobs simple-character-string character character-reg))
 
 (define-full-compare-and-swap %compare-and-swap-svref simple-vector
@@ -177,11 +196,54 @@
 ;;;; integer vectors whose elements are smaller than a byte, i.e.,
 ;;;; bit, 2-bit, and 4-bit vectors
 
+(define-vop (data-vector-ref-with-offset/simple-bit-vector-c)
+  (:translate data-vector-ref-with-offset)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg)))
+  (:arg-types simple-bit-vector
+              (:constant (integer 0 #x7fffffff)) (:constant (integer 0 0)))
+  (:info index offset)
+  (:ignore offset)
+  (:results (result :scs (any-reg)))
+  (:result-types positive-fixnum)
+  (:generator 3
+    (multiple-value-bind (dword-index bit) (floor index 32)
+      (inst mov result
+                (make-ea :dword :base object
+                         :disp (+ (* dword-index 4)
+                                  (- (* vector-data-offset n-word-bytes)
+                                     other-pointer-lowtag))))
+      (let ((right-shift (- bit n-fixnum-tag-bits)))
+        (cond ((plusp right-shift)
+               (inst shr result right-shift))
+              ((minusp right-shift) ; = left shift
+               (inst shl result (- right-shift))))))
+    (inst and result (fixnumize 1))))
+
+(define-vop (data-vector-ref-with-offset/simple-bit-vector)
+  (:translate data-vector-ref-with-offset)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg))
+         (index :scs (unsigned-reg)))
+  (:info offset)
+  (:ignore offset)
+  (:arg-types simple-bit-vector positive-fixnum (:constant (integer 0 0)))
+  (:results (result :scs (any-reg)))
+  (:result-types positive-fixnum)
+  (:generator 4
+    (inst bt (make-ea :dword :base object
+                      :disp (- (* vector-data-offset n-word-bytes)
+                               other-pointer-lowtag))
+             index)
+    (inst sbb result result)
+    (inst and result (fixnumize 1))))
+
 (macrolet ((def-small-data-vector-frobs (type bits)
              (let* ((elements-per-word (floor n-word-bits bits))
                     (bit-shift (1- (integer-length elements-per-word))))
     `(progn
-       (define-vop (,(symbolicate 'data-vector-ref-with-offset/ type))
+      ,@(unless (= bits 1)
+       `((define-vop (,(symbolicate 'data-vector-ref-with-offset/ type))
          (:note "inline array access")
          (:translate data-vector-ref-with-offset)
          (:policy :fast-safe)
@@ -224,7 +286,7 @@
              (unless (zerop extra)
                (inst shr result (* extra ,bits)))
              (unless (= extra ,(1- elements-per-word))
-               (inst and result ,(1- (ash 1 bits)))))))
+               (inst and result ,(1- (ash 1 bits)))))))))
        (define-vop (,(symbolicate 'data-vector-set-with-offset/ type))
          (:note "inline array store")
          (:translate data-vector-set-with-offset)
@@ -317,7 +379,7 @@
 
 ;;; And the float variants.
 
-(defun make-ea-for-float-ref (object index offset element-size
+(defun float-ref-ea (object index offset element-size
                               &key (scale 1) (complex-offset 0))
   (sc-case index
     (immediate
@@ -347,7 +409,7 @@
   (:result-types single-float)
   (:generator 5
    (with-empty-tn@fp-top(value)
-     (inst fld (make-ea-for-float-ref object index offset 4)))))
+     (inst fld (float-ref-ea object index offset 4)))))
 
 (define-vop (data-vector-set-with-offset/simple-array-single-float)
   (:note "inline array store")
@@ -366,14 +428,14 @@
   (:generator 5
     (cond ((zerop (tn-offset value))
            ;; Value is in ST0.
-           (inst fst (make-ea-for-float-ref object index offset 4))
+           (inst fst (float-ref-ea object index offset 4))
            (unless (zerop (tn-offset result))
              ;; Value is in ST0 but not result.
              (inst fst result)))
           (t
            ;; Value is not in ST0.
            (inst fxch value)
-           (inst fst (make-ea-for-float-ref object index offset 4))
+           (inst fst (float-ref-ea object index offset 4))
            (cond ((zerop (tn-offset result))
                   ;; The result is in ST0.
                   (inst fst value))
@@ -398,7 +460,7 @@
   (:result-types double-float)
   (:generator 7
    (with-empty-tn@fp-top(value)
-     (inst fldd (make-ea-for-float-ref object index offset 8 :scale 2)))))
+     (inst fldd (float-ref-ea object index offset 8 :scale 2)))))
 
 (define-vop (data-vector-set-with-offset/simple-array-double-float)
   (:note "inline array store")
@@ -417,14 +479,14 @@
   (:generator 20
     (cond ((zerop (tn-offset value))
            ;; Value is in ST0.
-           (inst fstd (make-ea-for-float-ref object index offset 8 :scale 2))
+           (inst fstd (float-ref-ea object index offset 8 :scale 2))
            (unless (zerop (tn-offset result))
                    ;; Value is in ST0 but not result.
                    (inst fstd result)))
           (t
            ;; Value is not in ST0.
            (inst fxch value)
-           (inst fstd (make-ea-for-float-ref object index offset 8 :scale 2))
+           (inst fstd (float-ref-ea object index offset 8 :scale 2))
            (cond ((zerop (tn-offset result))
                   ;; The result is in ST0.
                   (inst fstd value))
@@ -451,11 +513,11 @@
   (:generator 5
     (let ((real-tn (complex-single-reg-real-tn value)))
       (with-empty-tn@fp-top (real-tn)
-        (inst fld (make-ea-for-float-ref object index offset 8 :scale 2))))
+        (inst fld (float-ref-ea object index offset 8 :scale 2))))
     (let ((imag-tn (complex-single-reg-imag-tn value)))
       (with-empty-tn@fp-top (imag-tn)
         ;; FIXME
-        (inst fld (make-ea-for-float-ref object index offset 8
+        (inst fld (float-ref-ea object index offset 8
                                          :scale 2 :complex-offset 4))))))
 
 (define-vop (data-vector-set-with-offset/simple-array-complex-single-float)
@@ -477,14 +539,14 @@
           (result-real (complex-single-reg-real-tn result)))
       (cond ((zerop (tn-offset value-real))
              ;; Value is in ST0.
-             (inst fst (make-ea-for-float-ref object index offset 8 :scale 2))
+             (inst fst (float-ref-ea object index offset 8 :scale 2))
              (unless (zerop (tn-offset result-real))
                ;; Value is in ST0 but not result.
                (inst fst result-real)))
             (t
              ;; Value is not in ST0.
              (inst fxch value-real)
-             (inst fst (make-ea-for-float-ref object index offset 8 :scale 2))
+             (inst fst (float-ref-ea object index offset 8 :scale 2))
              (cond ((zerop (tn-offset result-real))
                     ;; The result is in ST0.
                     (inst fst value-real))
@@ -496,7 +558,7 @@
     (let ((value-imag (complex-single-reg-imag-tn value))
           (result-imag (complex-single-reg-imag-tn result)))
       (inst fxch value-imag)
-      (inst fst (make-ea-for-float-ref object index offset 8
+      (inst fst (float-ref-ea object index offset 8
                                        :scale 2 :complex-offset 4))
       (unless (location= value-imag result-imag)
         (inst fst result-imag))
@@ -517,10 +579,10 @@
   (:generator 7
     (let ((real-tn (complex-double-reg-real-tn value)))
       (with-empty-tn@fp-top (real-tn)
-        (inst fldd (make-ea-for-float-ref object index offset 16 :scale 4)))
+        (inst fldd (float-ref-ea object index offset 16 :scale 4)))
     (let ((imag-tn (complex-double-reg-imag-tn value)))
       (with-empty-tn@fp-top (imag-tn)
-        (inst fldd (make-ea-for-float-ref object index offset 16
+        (inst fldd (float-ref-ea object index offset 16
                                           :scale 4 :complex-offset 8)))))))
 
 (define-vop (data-vector-set-with-offset/simple-array-complex-double-float)
@@ -542,7 +604,7 @@
           (result-real (complex-double-reg-real-tn result)))
       (cond ((zerop (tn-offset value-real))
              ;; Value is in ST0.
-             (inst fstd (make-ea-for-float-ref object index offset 16
+             (inst fstd (float-ref-ea object index offset 16
                                                :scale 4))
              (unless (zerop (tn-offset result-real))
                ;; Value is in ST0 but not result.
@@ -550,7 +612,7 @@
             (t
              ;; Value is not in ST0.
              (inst fxch value-real)
-             (inst fstd (make-ea-for-float-ref object index offset 16
+             (inst fstd (float-ref-ea object index offset 16
                                                :scale 4))
              (cond ((zerop (tn-offset result-real))
                     ;; The result is in ST0.
@@ -563,7 +625,7 @@
     (let ((value-imag (complex-double-reg-imag-tn value))
           (result-imag (complex-double-reg-imag-tn result)))
       (inst fxch value-imag)
-      (inst fstd (make-ea-for-float-ref object index offset 16
+      (inst fstd (float-ref-ea object index offset 16
                                         :scale 4 :complex-offset 8))
       (unless (location= value-imag result-imag)
         (inst fstd result-imag))
@@ -640,8 +702,8 @@
   (define-data-vector-frobs simple-array-signed-byte-8 tagged-num
     movsx nil signed-reg)
   (define-data-vector-frobs simple-base-string character
-                            #!+sb-unicode movzx #!-sb-unicode mov
-                            #!+sb-unicode nil #!-sb-unicode t character-reg))
+                            #+sb-unicode movzx #-sb-unicode mov
+                            #+sb-unicode nil #-sb-unicode t character-reg))
 
 ;;; {un,}signed-byte-16
 (macrolet ((define-data-vector-frobs (ptype element-type ref-inst &rest scs)
@@ -723,8 +785,7 @@
   (:results (result :scs (unsigned-reg)))
   (:result-types unsigned-num)
   (:generator 4
-    (inst xadd (make-ea :dword :base array
-                        :scale 1 :index index
+    (inst xadd (make-ea :dword :base array :index index
                         :disp (- (* vector-data-offset n-word-bytes)
                                  other-pointer-lowtag))
           diff :lock)
