@@ -121,7 +121,16 @@
                           (find-class ',class) ,class)))
                classes)))
 
-(defun !wrapper-p (x) (and (sb-kernel::layout-p x) (layout-for-pcl-obj-p x)))
+(defmacro wrapper-info (x) `(sb-kernel::layout-%info ,x))
+(declaim (inline wrapper-slot-list))
+(defun wrapper-slot-list (wrapper)
+  (let ((info (wrapper-info wrapper)))
+    (if (listp info) info)))
+(defun (setf wrapper-slot-list) (newval wrapper)
+  ;; The current value must be a list, otherwise we'd clobber
+  ;; a defstruct-description.
+  (aver (listp (wrapper-info wrapper)))
+  (setf (wrapper-info wrapper) newval))
 
 (defun !bootstrap-meta-braid ()
   (let* ((*create-classes-from-internal-structure-definitions-p* nil)
@@ -209,8 +218,8 @@
                   (error "Slot allocation ~S is not supported in bootstrap."
                          (getf slot :allocation))))
 
-              (when (!wrapper-p wrapper)
-                (setf (layout-slot-list wrapper) slots))
+              (when (layout-for-pcl-obj-p wrapper)
+                (setf (wrapper-slot-list wrapper) slots))
 
               (setq proto (if (eq meta 'funcallable-standard-class)
                               (allocate-standard-funcallable-instance wrapper name)
@@ -226,10 +235,10 @@
                      standard-effective-slot-definition-wrapper t))
 
               (setf (layout-slot-table wrapper) (make-slot-table class slots t))
-              (when (!wrapper-p wrapper)
-                (setf (layout-slot-list wrapper) slots))
+              (when (layout-for-pcl-obj-p wrapper)
+                (setf (wrapper-slot-list wrapper) slots))
 
-              (case meta
+              (ecase meta
                 ((standard-class funcallable-standard-class)
                  (!bootstrap-initialize-class
                   meta
@@ -359,8 +368,8 @@
             (make-slot-table class slots
                              (member metaclass-name
                                      '(standard-class funcallable-standard-class))))
-      (when (!wrapper-p wrapper)
-        (setf (layout-slot-list wrapper) slots)))
+      (when (layout-for-pcl-obj-p wrapper)
+        (setf (wrapper-slot-list wrapper) slots)))
 
     ;; For all direct superclasses SUPER of CLASS, make sure CLASS is
     ;; a direct subclass of SUPER.  Note that METACLASS-NAME doesn't
@@ -655,6 +664,8 @@
         (add-method gf class-method)))
     gf))
 
+(define-load-time-global *simple-stream-root-classoid* :unknown)
+
 (defun assign-layout-bitmap (layout)
   ;; We decide only at class finalization time whether it is funcallable.
   ;; Picking the right bitmap could probably be done sooner given the metaclass,
@@ -662,16 +673,28 @@
   ;; The big comment above MAKE-IMMOBILE-FUNINSTANCE in src/code/x86-64-vm
   ;; explains why we differentiate between SGF and everything else.
   (let ((inherits (layout-inherits layout)))
+    (when (find #.(find-layout 'stream) inherits)
+      (let ((flags 0))
+        (dovector (layout inherits)
+          (when (eq (layout-classoid layout) *simple-stream-root-classoid*)
+            (setq flags (logior flags +simple-stream-layout-flag+)))
+          (case layout
+            (#.(find-layout 'file-stream)
+             (setq flags (logior flags +file-stream-layout-flag+)))
+            (#.(find-layout 'string-stream)
+             (setq flags +string-stream-layout-flag+))))
+        (setf (layout-flags layout)
+              (logior flags +stream-layout-flag+ (layout-flags layout)))))
+    (when (find #.(find-layout 'sequence) inherits)
+      (setf (layout-flags layout) (logior (layout-flags layout) +sequence-layout-flag+)))
     (when (find #.(find-layout 'function) inherits)
-      (let ((bitmap
+      (setf (%raw-instance-ref/signed-word layout (sb-kernel::type-dd-length layout))
              #+immobile-code ; there are two possible bitmaps
              (if (or (find *sgf-wrapper* inherits) (eq layout *sgf-wrapper*))
                  sb-kernel::standard-gf-primitive-obj-layout-bitmap
                  +layout-all-tagged+)
              ;; there is only one possible bitmap otherwise
-             #-immobile-code sb-kernel::standard-gf-primitive-obj-layout-bitmap))
-        (setf (%raw-instance-ref/word layout (sb-kernel::type-dd-length layout))
-              (logand sb-ext:most-positive-word bitmap))))))
+             #-immobile-code sb-kernel::standard-gf-primitive-obj-layout-bitmap))))
 
 ;;; Set the inherits from CPL, and register the layout. This actually
 ;;; installs the class in the Lisp type system.
