@@ -38,7 +38,7 @@
            has already been partially loaded. This may not work, you may~%~
            need to get a fresh lisp (reboot) and then load PCL."))
 
-#-sb-fluid (declaim (inline gdefinition))
+(declaim (inline gdefinition))
 (defun gdefinition (spec)
   ;; This is null layer right now, but once FDEFINITION stops bypasssing
   ;; fwrappers/encapsulations we can do that here.
@@ -195,8 +195,8 @@
                (let ((subs (classoid-subclasses class)))
                  (/noshow subs)
                  (when subs
-                   (dohash ((sub v) subs)
-                     (declare (ignore v))
+                   (dohash ((sub wrapper) subs)
+                     (declare (ignore wrapper))
                      (/noshow sub)
                      (when (member class (direct-supers sub) :test #'eq)
                        (res sub)))))
@@ -219,7 +219,8 @@
                                  ;; SYSTEM-CLASSes) from the
                                  ;; BUILT-IN-CLASS list
                                  '(t function stream sequence
-                                     file-stream string-stream)))
+                                   file-stream string-stream
+                                   slot-object)))
                        sb-kernel::+!built-in-classes+))))
 (/noshow "done setting up SB-PCL::*BUILT-IN-CLASSES*")
 
@@ -330,6 +331,7 @@
    (lambda-list :initform () :initarg :lambda-list :reader method-lambda-list)
    (%function :initform nil :initarg :function :reader method-function)
    (%documentation :initform nil :initarg :documentation)
+   (%cache :initform nil :accessor method-em-cache)
    ;; True IFF method is known to have no CALL-NEXT-METHOD in it, or
    ;; just a plain (CALL-NEXT-METHOD).
    (simple-next-method-call
@@ -362,7 +364,7 @@
 
 (defun make-gf-hash-table ()
   (make-hash-table :test 'eq
-                   :hash-function #'sb-impl::fsc-instance-hash ; stable hash
+                   :hash-function #'fsc-instance-hash ; stable hash
                    :weakness :key
                    :synchronized t))
 
@@ -568,14 +570,12 @@
 ;; gethash on a live key should get the identical specializer, but since
 ;; nothing referenced the old specializer, consing a new one is fine.
 (defglobal *eql-specializer-table*
-  (make-hash-table :test 'eql :weakness :value))
+  (sb-impl::make-system-hash-table :test 'eql :weakness :value :synchronized nil))
 
 (defun intern-eql-specializer (object)
   ;; Avoid style-warning about compiler-macro being unavailable.
   (declare (notinline make-instance))
-  ;; Need to lock, so that two threads don't get non-EQ specializers
-  ;; for an EQL object.
-  (with-locked-system-table (*eql-specializer-table*)
+  (with-system-mutex ((hash-table-lock *eql-specializer-table*))
     (ensure-gethash object *eql-specializer-table*
                     (make-instance 'eql-specializer :object object))))
 
