@@ -42,15 +42,6 @@
                          (cell-error-name condition)
                          (type-of (unbound-slot-instance condition))))))))
 
-(defmethod wrapper-fetcher ((class standard-class))
-  'std-instance-wrapper)
-
-(defmethod slots-fetcher ((class standard-class))
-  'std-instance-slots)
-
-(defmethod raw-instance-allocator ((class standard-class))
-  'allocate-standard-instance)
-
 ;;; These three functions work on std-instances and fsc-instances. These are
 ;;; instances for which it is possible to change the wrapper and the slots.
 ;;;
@@ -58,23 +49,6 @@
 ;;; structure protocol are promoted to the implementation-specific class
 ;;; std-class. Many of these methods call these four functions.
 
-(defun %swap-wrappers-and-slots (i1 i2)
-  (cond ((std-instance-p i1)
-         (let ((w1 (std-instance-wrapper i1))
-               (s1 (std-instance-slots i1)))
-           (setf (std-instance-wrapper i1) (std-instance-wrapper i2))
-           (setf (std-instance-slots i1) (std-instance-slots i2))
-           (setf (std-instance-wrapper i2) w1)
-           (setf (std-instance-slots i2) s1)))
-        ((fsc-instance-p i1)
-         (let ((w1 (fsc-instance-wrapper i1))
-               (s1 (fsc-instance-slots i1)))
-           (setf (fsc-instance-wrapper i1) (fsc-instance-wrapper i2))
-           (setf (fsc-instance-slots i1) (fsc-instance-slots i2))
-           (setf (fsc-instance-wrapper i2) w1)
-           (setf (fsc-instance-slots i2) s1)))
-        (t
-         (error "unrecognized instance type"))))
 
 ;;;; STANDARD-INSTANCE-ACCESS
 
@@ -121,7 +95,7 @@
                  (return-from slot-value
                    (if cell
                        (funcall (slot-info-reader (cdr cell)) object)
-                       (values (slot-missing (wrapper-class* wrapper) object
+                       (values (slot-missing (wrapper-class wrapper) object
                                              slot-name 'slot-value)))))
                 ;; this next test means CONSP, but the transform that weakens
                 ;; CONSP to LISTP isn't working here for some reason.
@@ -129,15 +103,15 @@
                  (cdr location))
                 (t
                  (bug "Bogus slot cell in SLOT-VALUE: ~S" cell)))))
-    (if (eq +slot-unbound+ value)
-        (slot-unbound (wrapper-class* wrapper) object slot-name)
+    (if (unbound-marker-p value)
+        (slot-unbound (wrapper-class wrapper) object slot-name)
         value)))
 
 (defun set-slot-value (object slot-name new-value)
   (let* ((wrapper (valid-wrapper-of object))
          (cell (or (find-slot-cell wrapper slot-name)
                    (return-from set-slot-value
-                     (progn (slot-missing (wrapper-class* wrapper)
+                     (progn (slot-missing (wrapper-class wrapper)
                                           object slot-name 'setf new-value)
                             new-value))))
          (location (car cell))
@@ -169,7 +143,7 @@
   (let* ((wrapper (valid-wrapper-of object))
          (cell (or (find-slot-cell wrapper slot-name)
                    (return-from slot-value
-                     (values (slot-missing (wrapper-class* wrapper) object slot-name
+                     (values (slot-missing (wrapper-class wrapper) object slot-name
                                            'cas (list old-value new-value))))))
          (location (car cell))
          (info (cdr cell))
@@ -188,9 +162,8 @@
                       (cas (cdr location) old-value new-value))
                      (t
                       (bug "Bogus slot-cell in (CAS SLOT-VALUE): ~S" cell)))))
-      (if (and (eq +slot-unbound+ old)
-               (neq old old-value))
-          (slot-unbound (wrapper-class* wrapper) object slot-name)
+      (if (and (unbound-marker-p old) (neq old old-value))
+          (slot-unbound (wrapper-class wrapper) object slot-name)
           old))))
 
 (defun slot-boundp (object slot-name)
@@ -206,14 +179,14 @@
                  (return-from slot-boundp
                    (if cell
                        (funcall (slot-info-boundp (cdr cell)) object)
-                       (and (slot-missing (wrapper-class* wrapper) object
+                       (and (slot-missing (wrapper-class wrapper) object
                                           slot-name 'slot-boundp)
                             t))))
                 ((listp location) ; forcibly transform CONSP to LISTP
                  (cdr location))
                 (t
                  (bug "Bogus slot cell in SLOT-VALUE: ~S" cell)))))
-    (not (eq +slot-unbound+ value))))
+    (not (unbound-marker-p value))))
 
 (defun slot-makunbound (object slot-name)
   (let* ((wrapper (valid-wrapper-of object))
@@ -226,10 +199,10 @@
                      +slot-unbound+)))
           ((not location)
            (if cell
-               (let ((class (wrapper-class* wrapper)))
+               (let ((class (wrapper-class wrapper)))
                  (slot-makunbound-using-class class object
                                               (find-slot-definition class slot-name)))
-               (slot-missing (wrapper-class* wrapper) object slot-name
+               (slot-missing (wrapper-class wrapper) object slot-name
                              'slot-makunbound)))
           ((listp location) ; forcibly transform CONSP to LISTP
            (setf (cdr location) +slot-unbound+))
@@ -248,6 +221,11 @@
       (slot-value object slot-name)
       (load-time-value (make-unprintable-object "unbound slot") t)))
 
+(defmacro safely-get-slots (method-name x)
+  `(cond ((std-instance-p ,x) (std-instance-slots ,x))
+         ((fsc-instance-p ,x) (fsc-instance-slots ,x))
+         (t (bug "unrecognized instance type in ~S" ',method-name))))
+
 (defmethod slot-value-using-class ((class std-class)
                                    (object standard-object)
                                    (slotd standard-effective-slot-definition))
@@ -258,20 +236,14 @@
          (value
           (typecase location
             (fixnum
-             (cond ((std-instance-p object)
-                    (clos-slots-ref (std-instance-slots object)
-                                    location))
-                   ((fsc-instance-p object)
-                    (clos-slots-ref (fsc-instance-slots object)
-                                    location))
-                   (t (bug "unrecognized instance type in ~S"
-                           'slot-value-using-class))))
+             (clos-slots-ref (safely-get-slots slot-value-using-class object)
+                             location))
             (cons
              (cdr location))
             (t
              (instance-structure-protocol-error slotd
                                                 'slot-value-using-class)))))
-    (if (eq value +slot-unbound+)
+    (if (unbound-marker-p value)
         (values (slot-unbound class object (slot-definition-name slotd)))
         value)))
 
@@ -290,14 +262,9 @@
                         new-value)))
     (typecase location
       (fixnum
-       (cond ((std-instance-p object)
-              (setf (clos-slots-ref (std-instance-slots object) location)
-                    new-value))
-             ((fsc-instance-p object)
-              (setf (clos-slots-ref (fsc-instance-slots object) location)
-                    new-value))
-             (t (bug "unrecognized instance type in ~S"
-                     '(setf slot-value-using-class)))))
+       (setf (clos-slots-ref (safely-get-slots (setf slot-value-using-class) object)
+                             location)
+             new-value))
       (cons
        (setf (cdr location) new-value))
       (t
@@ -311,24 +278,17 @@
   ;; FIXME: Do we need this? SLOT-BOUNDP checks for obsolete
   ;; instances. Are users allowed to call this directly?
   (check-obsolete-instance object)
-  (let* ((location (slot-definition-location slotd))
-         (value
+  (let ((location (slot-definition-location slotd)))
+    (not (unbound-marker-p
           (typecase location
             (fixnum
-             (cond ((std-instance-p object)
-                          (clos-slots-ref (std-instance-slots object)
-                                          location))
-                   ((fsc-instance-p object)
-                    (clos-slots-ref (fsc-instance-slots object)
-                                    location))
-                   (t (bug "unrecognized instance type in ~S"
-                           'slot-boundp-using-class))))
+             (clos-slots-ref (safely-get-slots slot-boundp-using-class object)
+                             location))
             (cons
              (cdr location))
             (t
              (instance-structure-protocol-error slotd
-                                                'slot-boundp-using-class)))))
-    (not (eq value +slot-unbound+))))
+                                                'slot-boundp-using-class)))))))
 
 (defmethod slot-makunbound-using-class
            ((class std-class)
@@ -338,14 +298,9 @@
   (let ((location (slot-definition-location slotd)))
     (typecase location
       (fixnum
-       (cond ((std-instance-p object)
-              (setf (clos-slots-ref (std-instance-slots object) location)
-                    +slot-unbound+))
-             ((fsc-instance-p object)
-              (setf (clos-slots-ref (fsc-instance-slots object) location)
-                    +slot-unbound+))
-             (t (bug "unrecognized instance type in ~S"
-                     'slot-makunbound-using-class))))
+       (setf (clos-slots-ref (safely-get-slots slot-makunbound-using-class object)
+                             location)
+             +slot-unbound+))
       (cons
        (setf (cdr location) +slot-unbound+))
       (t
@@ -388,7 +343,7 @@
     (declare (type function function))
     ;; FIXME: Is this really necessary? Structure slots should surely
     ;; never be unbound!
-    (if (eq value +slot-unbound+)
+    (if (unbound-marker-p value)
         (values (slot-unbound class object (slot-definition-name slotd)))
         value)))
 
@@ -415,7 +370,7 @@
 (defmethod slot-missing
            ((class t) instance slot-name operation &optional new-value)
   (error "~@<When attempting to ~A, the slot ~S is missing from the ~
-          object ~S.~@:>"
+          object ~S.~@[~%~a~]~@:>"
          (ecase operation
            (slot-value "read the slot's value (slot-value)")
            (setf (format nil
@@ -424,7 +379,16 @@
            (slot-boundp "test to see whether slot is bound (SLOT-BOUNDP)")
            (slot-makunbound "make the slot unbound (SLOT-MAKUNBOUND)"))
          slot-name
-         instance))
+         instance
+         (let ((slot (and (typep class 'slot-class)
+                          (find slot-name (class-slots class)
+                                :key #'slot-definition-name
+                                :test #'string-equal))))
+           (when slot
+             (format nil "It has a slot ~/sb-ext:print-symbol-with-prefix/, while ~
+                         ~/sb-ext:print-symbol-with-prefix/ is requested."
+                     (slot-definition-name slot)
+                     slot-name)))))
 
 (defmethod slot-unbound ((class t) instance slot-name)
   (restart-case
@@ -448,12 +412,9 @@
        ;; In the vast majority of cases location corresponds to the position
        ;; in list. The only exceptions are when there are non-local slots
        ;; before the one we want.
-       (let* ((slots (layout-slot-list (layout-of instance)))
-              (guess (nth position slots)))
-         (if (eql position (slot-definition-location guess))
-             (slot-definition-name guess)
-             (slot-definition-name
-              (car (member position (class-slots instance) :key #'slot-definition-location))))))
+       (slot-definition-name
+        (find position (wrapper-slot-list (layout-of instance))
+              :key #'slot-definition-location)))
       (cons
        (car position))))))
 
@@ -487,11 +448,11 @@
     (error 'simple-reference-error
            :format-control "~S called on ~S, which is not yet finalized."
            :format-arguments (list 'class-slots class)
-           :references (list '(:amop :generic-function class-slots)))))
+           :references '((:amop :generic-function class-slots)))))
 
 (defun %set-slots (object names &rest values)
   (mapc (lambda (name value)
-          (if (eq value +slot-unbound+)
+          (if (unbound-marker-p value)
               ;; SLOT-MAKUNBOUND-USING-CLASS might do something nonstandard.
               (slot-makunbound object name)
               (setf (slot-value object name) value)))
