@@ -181,13 +181,11 @@
             (dovector (super-layout (layout-inherits layout))
               (let* ((super (layout-classoid super-layout))
                      (subclasses (or (classoid-subclasses super)
-                                     (or (classoid-subclasses super)
-                                         (setf (classoid-subclasses super)
-                                               (make-hash-table :hash-function #'type-hash-value
-                                                                :test 'eq
-                                                                :synchronized t))))))
-                (when subclasses
-                  (setf (gethash classoid subclasses) layout))))))))
+                                     (setf (classoid-subclasses super)
+                                           (make-hash-table :hash-function #'type-hash-value
+                                                            :test 'eq
+                                                            :synchronized t)))))
+                (setf (gethash classoid subclasses) (sb-kernel::layout-friend layout))))))))
     subclass))
 (defmethod remove-direct-subclass ((class class) (subclass class))
   (with-slots (direct-subclasses) class
@@ -561,7 +559,9 @@
                 (set-layout-inherits layout
                                      (order-layout-inherits
                                       (map 'simple-vector #'class-wrapper
-                                           (reverse (rest cpl))))))
+                                           (reverse (rest cpl))))
+                                     nil 0))
+              (assign-layout-bitmap layout)
               (register-layout layout :invalidate t))))
         (mapc #'make-preliminary-layout (class-direct-subclasses class))))))
 
@@ -774,7 +774,7 @@
        ;;  (LAMBDA () (SB-PCL::FAST-MAKE-INSTANCE #<STRUCTURE-CLASS THING>))
        ;; So maybe we can figure out how to bundle two lambdas together?
        (lambda ()
-         (let* ((dd (layout-info (class-wrapper class)))
+         (let* ((dd (layout-dd (class-wrapper class)))
                 (f (%make-structure-instance-allocator dd nil)))
            (if (functionp f)
                (funcall (setf (slot-value class 'defstruct-constructor) f))
@@ -1082,7 +1082,7 @@
            (owrapper (class-wrapper class))
            (nwrapper
              (cond ((and owrapper
-                         (slot-layouts-compatible-p (layout-slot-list owrapper)
+                         (slot-layouts-compatible-p (wrapper-slot-list owrapper)
                                                     instance-slots class-slots custom-slots))
                     owrapper)
                    ((or (not owrapper)
@@ -1098,7 +1098,7 @@
                     (class-wrapper class)))))
       (%update-lisp-class-layout class nwrapper)
       (setf (slot-value class 'slots) eslotds
-            (layout-slot-list nwrapper) eslotds
+            (wrapper-slot-list nwrapper) eslotds
             (layout-slot-table nwrapper) (make-slot-table class eslotds)
             (layout-length nwrapper) nslots
             (slot-value class 'wrapper) nwrapper)
@@ -1452,7 +1452,7 @@
                 (eq (layout-invalid owrapper) t))
         (let ((nwrapper (make-wrapper (layout-length owrapper)
                                       class)))
-          (setf (layout-slot-list nwrapper) (layout-slot-list owrapper))
+          (setf (wrapper-slot-list nwrapper) (wrapper-slot-list owrapper))
           (setf (layout-slot-table nwrapper) (layout-slot-table owrapper))
           (%update-lisp-class-layout class nwrapper)
           (setf (slot-value class 'wrapper) nwrapper)
@@ -1478,7 +1478,7 @@
         (if (class-has-a-forward-referenced-superclass-p class)
             (return-from make-instances-obsolete class)
             (%update-cpl class (compute-class-precedence-list class))))
-      (setf (layout-slot-list nwrapper) (layout-slot-list owrapper))
+      (setf (wrapper-slot-list nwrapper) (wrapper-slot-list owrapper))
       (setf (layout-slot-table nwrapper) (layout-slot-table owrapper))
       (%update-lisp-class-layout class nwrapper)
       (setf (slot-value class 'wrapper) nwrapper)
@@ -1577,9 +1577,9 @@
                 (plist ())
                 (safe (safe-p class))
                 ((new-instance-slots nil new-custom-slots)
-                 (classify-slotds (layout-slot-list nwrapper)))
+                 (classify-slotds (wrapper-slot-list nwrapper)))
                 ((old-instance-slots old-class-slots old-custom-slots)
-                 (classify-slotds (layout-slot-list owrapper)))
+                 (classify-slotds (wrapper-slot-list owrapper)))
                 (layout (mapcar (lambda (slotd)
                                   ;; Get the names only once.
                                   (cons (slot-definition-name slotd) slotd))
@@ -1672,7 +1672,7 @@
              (old-class (wrapper-class old-wrapper))
              (old-slots (get-slots instance))
              (safe (safe-p new-class))
-             (new-wrapper-slots (layout-slot-list new-wrapper)))
+             (new-wrapper-slots (wrapper-slot-list new-wrapper)))
     (if (functionp copy)
         (setf (%fun-layout copy) new-wrapper
               (%fsc-instance-slots copy) new-slots)
