@@ -32,7 +32,7 @@
 ;;; This is called in BRAID when we are making wrappers for classes
 ;;; whose slots are not initialized yet, and which may be built-in
 ;;; classes.
-(defun !boot-make-wrapper (length name &optional (bitmap -1))
+(defun !boot-make-wrapper (length name &optional (bitmap +layout-all-tagged+))
   (let ((found (find-classoid name nil)))
     (cond
      (found
@@ -40,12 +40,12 @@
         (aver layout)
         layout))
      (t
-      (set-layout-valid
        (make-layout (hash-layout-name name)
                     (make-standard-classoid :name name)
                     :length length
                     :flags +pcl-object-layout-flag+
-                    :bitmap bitmap))))))
+                    :bitmap bitmap
+                    :invalid nil)))))
 
 ;;; In SBCL, as in CMU CL, the layouts (a.k.a wrappers) for built-in
 ;;; and structure classes already exist when PCL is initialized, so we
@@ -71,9 +71,9 @@
                                               :name (and (symbolp name) name)))
                      (t
                       (bug "Got to T branch in ~S" 'make-wrapper))))))
-       (set-layout-valid
-        (make-layout (hash-layout-name name)
-                     classoid :length length :flags +pcl-object-layout-flag+))))
+        (make-layout (hash-layout-name name) classoid
+                     :bitmap +layout-all-tagged+
+                     :invalid nil :length length :flags +pcl-object-layout-flag+)))
     (t
      (let* ((found (find-classoid (slot-value class 'name)))
             (layout (classoid-layout found)))
@@ -245,17 +245,17 @@
                                                            (class-wrapper class)))))))
              (check-wrapper-validity instance))
             ((consp state)
-             (ecase (car state)
-               (:flush
-                (let ((new (cdr state)))
+             (let ((new (the layout (cdr state))))
+               (ecase (car state)
+                 (:flush
                   (cond ((std-instance-p instance)
                          (setf (%instance-layout instance) new))
                         ((fsc-instance-p instance)
                          (setf (%fun-layout instance) new))
                         (t
-                         (bug "unrecognized instance type")))))
-               (:obsolete
-                (%obsolete-instance-trap owrapper (cdr state) instance))))))))
+                         (bug "unrecognized instance type"))))
+                 (:obsolete
+                  (%obsolete-instance-trap owrapper new instance)))))))))
 
 (declaim (inline check-obsolete-instance))
 (defun check-obsolete-instance (instance)
