@@ -34,6 +34,7 @@
 #include "arch.h"
 #include "code.h"
 #include "private-cons.inc"
+#include "getallocptr.h"
 
 /* So you need to debug? */
 #if 0
@@ -69,7 +70,7 @@ tv_diff(struct timeval *x, struct timeval *y)
 #endif
 
 void *
-gc_general_alloc(sword_t bytes, int page_type_flag, int quick_p) {
+gc_general_alloc(sword_t bytes, int page_type_flag) {
     lispobj *new=new_space_free_pointer;
     new_space_free_pointer+=(bytes/N_WORD_BYTES);
     return new;
@@ -82,6 +83,15 @@ lispobj  copy_large_object(lispobj object, sword_t nwords, int page_type_flag) {
     return copy_object(object,nwords);
 }
 
+/*
+ * This flag is needed for compatibility with gencgc.
+ * In theory, it says to splat a nonzero byte pattern over newly allocated
+ * memory before giving the block to Lisp, to verify that Lisp is able to deal
+ * with non-pre-zeroed memory.
+ * In practice, that's not how cheneygc works.
+ */
+char gc_allocate_dirty = 0;
+
 /* Note: The generic GC interface we're implementing passes us a
  * last_generation argument. That's meaningless for us, since we're
  * not a generational GC. So we ignore it. */
@@ -98,7 +108,7 @@ collect_garbage(generation_index_t ignore)
     unsigned long size_retained;
     lispobj *current_static_space_free_pointer;
     sigset_t old;
-    struct thread *th=arch_os_get_current_thread();
+    struct thread *th=get_sb_vm_thread();
 
 #ifdef PRINTNOISE
     printf("[Collecting garbage ... \n");
@@ -118,7 +128,7 @@ collect_garbage(generation_index_t ignore)
     /* Set up from space and new space pointers. */
 
     from_space = current_dynamic_space;
-    from_space_free_pointer = dynamic_space_free_pointer;
+    from_space_free_pointer = get_alloc_pointer();
 
 #ifdef PRINTNOISE
     fprintf(stderr,"from_space = %lx\n",
@@ -143,8 +153,7 @@ collect_garbage(generation_index_t ignore)
     printf("Scavenging interrupt handlers (%d bytes) ...\n",
            (int)sizeof(interrupt_handlers));
 #endif
-    scavenge((lispobj *) interrupt_handlers,
-             sizeof(interrupt_handlers) / sizeof(lispobj));
+    scavenge(lisp_sig_handlers, NSIG);
 
 #ifdef PRINTNOISE
     printf("Scavenging the control stack ...\n");
@@ -191,15 +200,11 @@ collect_garbage(generation_index_t ignore)
     /* Maybe FIXME: it's possible that we could significantly reduce
      * RSS by zeroing the from_space or madvise(MADV_DONTNEED) or
      * similar os-dependent tricks here */
-#ifdef LISP_FEATURE_HPUX
-    /* hpux cant handle unmapping areas that are not 100% mapped */
-    clear_auto_gc_trigger();
-#endif
     os_zero((os_vm_address_t) from_space,
             (os_vm_size_t) dynamic_space_size);
 
     current_dynamic_space = new_space;
-    dynamic_space_free_pointer = new_space_free_pointer;
+    set_alloc_pointer((lispobj)new_space_free_pointer);
 
 #ifdef PRINTNOISE
     size_discarded = (from_space_free_pointer - from_space) * sizeof(lispobj);
@@ -326,7 +331,7 @@ print_garbage(lispobj *from_space, lispobj *from_space_free_pointer)
 
 /* weak pointers */
 
-static sword_t
+sword_t
 scav_weak_pointer(lispobj *where, lispobj object)
 {
     /* Do not let GC scavenge the value slot of the weak pointer */
@@ -340,7 +345,7 @@ lispobj *
 search_dynamic_space(void *pointer)
 {
     lispobj *start = (lispobj *) current_dynamic_space;
-    lispobj *end = (lispobj *) dynamic_space_free_pointer;
+    lispobj *end = get_alloc_pointer();
     if ((pointer < (void *)start) || (pointer >= (void *)end))
         return NULL;
     return gc_search_space(start, pointer);
@@ -352,8 +357,7 @@ search_dynamic_space(void *pointer)
 void
 gc_init(void)
 {
-    weakobj_init();
-    scavtab[WEAK_POINTER_WIDETAG] = scav_weak_pointer;
+    gc_common_init();
 }
 
 /* noise to manipulate the gc trigger stuff */
@@ -368,10 +372,10 @@ void set_auto_gc_trigger(os_vm_size_t dynamic_usage)
 
     addr = os_round_up_to_page((os_vm_address_t)current_dynamic_space
                                + dynamic_usage);
-    if (addr < (os_vm_address_t)dynamic_space_free_pointer)
+    if (addr < (os_vm_address_t)get_alloc_pointer())
         lose("set_auto_gc_trigger: tried to set gc trigger too low! (%ld < 0x%08lx)",
              (unsigned long)dynamic_usage,
-             (unsigned long)((os_vm_address_t)dynamic_space_free_pointer
+             (unsigned long)((os_vm_address_t)get_alloc_pointer()
                              - (os_vm_address_t)current_dynamic_space));
 
     if (dynamic_usage > dynamic_space_size)
@@ -379,12 +383,21 @@ void set_auto_gc_trigger(os_vm_size_t dynamic_usage)
              (unsigned long)dynamic_usage);
     length = os_trunc_size_to_page(dynamic_space_size - dynamic_usage);
 
-#if defined(SUNOS) || defined(SOLARIS) || defined(LISP_FEATURE_HPUX)
-    os_invalidate(addr, length);
+    // range has to fall entirely within either semispace
+    uword_t semispace_0_end = DYNAMIC_0_SPACE_START + dynamic_space_size;
+    uword_t semispace_1_end = DYNAMIC_1_SPACE_START + dynamic_space_size;
+    uword_t end = (uword_t)addr + length - 1;
+    if (((uword_t)addr >= DYNAMIC_0_SPACE_START && end < semispace_0_end) ||
+        ((uword_t)addr >= DYNAMIC_1_SPACE_START && end < semispace_1_end)) {
+#if defined(SUNOS) || defined(SOLARIS)
+        os_invalidate(addr, length);
 #else
-    os_protect(addr, length, 0);
+        os_protect(addr, length, 0);
 #endif
-
+    } else {
+        lose("auto_gc_trigger can't protect %p..%p (not owned)\n",
+             (void*)addr, (char*)end-1);
+    }
     current_auto_gc_trigger = (lispobj *)addr;
 }
 
@@ -399,7 +412,7 @@ void clear_auto_gc_trigger(void)
     addr = (os_vm_address_t)current_auto_gc_trigger;
     length = dynamic_space_size + (os_vm_address_t)current_dynamic_space - addr;
 
-#if defined(SUNOS) || defined(SOLARIS) || defined(LISP_FEATURE_HPUX)
+#if defined(SUNOS) || defined(SOLARIS)
     /* don't want to force whole space into swapping mode... */
     os_validate(NOT_MOVABLE, addr, length);
 #else
@@ -424,7 +437,7 @@ boolean
 cheneygc_handle_wp_violation(os_context_t *context, void *addr)
 {
     if(!foreign_function_call_active && gc_trigger_hit(addr)){
-        struct thread *thread=arch_os_get_current_thread();
+        struct thread *thread=get_sb_vm_thread();
         clear_auto_gc_trigger();
         /* Don't flood the system with interrupts if the need to gc is
          * already noted. This can happen for example when SUB-GC
