@@ -192,7 +192,7 @@ static void sigmask_logandc(sigset_t *dest, const sigset_t *source)
  * maybe all deferrables. */
 
 #if defined LISP_FEATURE_DARWIN && defined LISP_FEATURE_SB_THREAD
-pthread_key_t sigwait_bug_mitigation;
+pthread_key_t foreign_thread_ever_lispified;
 int sigwait_bug_mitigation_count;
 #endif
 
@@ -203,10 +203,14 @@ static void
 resignal_to_lisp_thread(int signal, os_context_t *context)
 {
 #if defined LISP_FEATURE_DARWIN && defined LISP_FEATURE_SB_THREAD
-    if (signal == SIG_STOP_FOR_GC && pthread_getspecific(sigwait_bug_mitigation)) {
-        // fprintf(stderr, "ignored a STOP_FOR_GC signal\n");
+    if (signal == SIG_STOP_FOR_GC && pthread_getspecific(foreign_thread_ever_lispified)) {
+        // This may be error-prone, I'm not sure.  Suppose there is a lingering
+        // stop-for-gc signal after we've demoted a lisp thread back to being
+        // a foreign thread. Suppose that thread then calls into lisp again so it re-promoted
+        // to a lisp thread. Is the next stop-for-gc signal real, or to be ignored?
+        // It'll be treated as real even if it was the lingering signal which ought to have
+        // been ignored. That probably won't happen, but "probably" is not a guarantee.
         __sync_fetch_and_add(&sigwait_bug_mitigation_count, 1);
-        pthread_setspecific(sigwait_bug_mitigation, 0);
         return;
     }
 #endif
@@ -267,7 +271,7 @@ resignal_to_lisp_thread(int signal, os_context_t *context)
  * simple-base-string and won't move, this is slightly ok */
 __attribute__((unused)) static char* cur_thread_name()
 {
-    struct thread* th = arch_os_get_current_thread();
+    struct thread* th = get_sb_vm_thread();
     struct thread_instance *lispthread =
         (void*)(th->lisp_thread - INSTANCE_POINTER_LOWTAG);
     struct vector* name = VECTOR(lispthread->name);
@@ -313,7 +317,7 @@ void dump_eventlog()
 static void record_signal(int sig, void* context)
 {
     event3("sig%d @%p in %d", sig, (void*)*os_context_pc_addr(context),
-           (int)arch_os_get_current_thread()->os_kernel_tid);
+           (int)get_sb_vm_thread()->os_kernel_tid);
 }
 #define RECORD_SIGNAL(sig,ctxt) if(sig!=SIGSEGV)record_signal(sig,ctxt);
 #else
@@ -364,9 +368,14 @@ sigset_tostring(const sigset_t *sigset, char* result, int result_length)
 
 /* Deferrables, blockables, gc signals. */
 
-static void
-sigaddset_deferrable(sigset_t *s)
-{
+#ifdef LISP_FEATURE_SB_SAFEPOINT
+static void sigaddset_deferrable(sigset_t *s) {
+    sigaddset(s, SIGURG);
+}
+static void sigaddset_async(sigset_t *s) {
+#else
+static void sigaddset_deferrable(sigset_t *s) {
+#endif
     sigaddset(s, SIGHUP);
     sigaddset(s, SIGINT);
     sigaddset(s, SIGTERM);
@@ -399,8 +408,12 @@ sigaddset_gc(sigset_t __attribute__((unused)) *sigset)
 void
 sigaddset_blockable(sigset_t *sigset)
 {
+#ifdef LISP_FEATURE_SB_SAFEPOINT
+    sigaddset_async(sigset);
+#else
     sigaddset_deferrable(sigset);
     sigaddset_gc(sigset);
+#endif
     // SIGPIPE is *NOT* an asynchronous signal. In normal usage you receive this signal
     // synchronously in response to a system call. As such we do not place it in the
     // deferrable set, but it _is_ blockable. If you're doing something wherein SIGPIPE
@@ -439,6 +452,13 @@ deferrables_blocked_p(sigset_t *sigset)
         thread_sigmask(SIG_BLOCK, 0, &current);
         sigset = &current;
     }
+#ifdef LISP_FEATURE_SB_SAFEPOINT
+    /* The only signal whose mask bit we manipulate is SIGURG.
+     * All other deferrable signals remain permanently in a blocked state.
+     * Therefore the answer to the question of whether deferrables
+     * are blocked is simply whether SIGURG is blocked */
+    return sigismember(sigset, SIGURG);
+#else
     /* SIGPROF must not be here. Some people use an async-signal-safe profiler
      * which not only doesn't rely on signal deferral, but wants to manipulate the
      * blocked/unblocked bit completely independently of SBCL's requirements.
@@ -468,6 +488,7 @@ deferrables_blocked_p(sigset_t *sigset)
     char buf[3*64]; // assuming worst case 64 signals present in sigset
     sigset_tostring(sigset, buf, sizeof buf);
     lose("deferrable signals partially blocked: {%s}", buf);
+#endif
 }
 
 void
@@ -569,11 +590,10 @@ unblock_deferrable_signals(sigset_t *where)
     check_gc_signals_unblocked_or_lose(where);
 #endif
     sigset_t localmask, *sigset;
-    if (arch_os_get_current_thread()->state_word.user_thread_p) {
+    if (get_sb_vm_thread()->state_word.user_thread_p) {
         sigset = &deferrable_sigset;
     } else {
-        /* ASSUMPTION: the baseline signal mask for non-user threads
-         * is the deferrable mask minus SIGALRM.
+        /* ASSUMPTION: system threads never want to receive SIGALRM.
          * Actually, if we get here in the finalizer thread, things are
          * in bad shape - stack exhaustion or something */
         localmask = deferrable_sigset;
@@ -616,7 +636,7 @@ they are not safe to interrupt at all, this is a pretty severe occurrence.\n");
 inline static void
 check_interrupts_enabled_or_lose(os_context_t *context)
 {
-    __attribute__((unused)) struct thread *thread = arch_os_get_current_thread();
+    __attribute__((unused)) struct thread *thread = get_sb_vm_thread();
     if (read_TLS(INTERRUPTS_ENABLED,thread) == NIL)
         lose("interrupts not enabled");
     if (arch_pseudo_atomic_atomic(context))
@@ -633,7 +653,7 @@ void
 maybe_save_gc_mask_and_block_deferrables(sigset_t *sigset)
 {
 #ifndef LISP_FEATURE_WIN32
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     struct interrupt_data *data = &thread_interrupt_data(thread);
     sigset_t oldset;
     /* Obviously, this function is called when signals may not be
@@ -689,7 +709,7 @@ in_leaving_without_gcing_race_p(struct thread __attribute__((unused)) *thread)
 static void
 check_interrupt_context_or_lose(os_context_t *context)
 {
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     struct interrupt_data *data = &thread_interrupt_data(thread);
     int interrupt_deferred_p = (data->pending_handler != 0);
     int interrupt_pending = (read_TLS(INTERRUPT_PENDING,thread) != NIL);
@@ -832,7 +852,7 @@ build_fake_control_stack_frames(struct thread __attribute__((unused)) *th,
 void fake_foreign_function_call_noassert(os_context_t *context)
 {
     int context_index;
-    struct thread *thread=arch_os_get_current_thread();
+    struct thread *thread=get_sb_vm_thread();
 
 #ifdef reg_ALLOC
 #ifdef LISP_FEATURE_SB_THREAD
@@ -897,7 +917,7 @@ void fake_foreign_function_call(os_context_t *context)
 void
 undo_fake_foreign_function_call(os_context_t __attribute__((unused)) *context)
 {
-    struct thread *thread=arch_os_get_current_thread();
+    struct thread *thread=get_sb_vm_thread();
     /* Block all blockable signals. */
     block_blockable_signals(0);
 
@@ -990,7 +1010,7 @@ interrupt_internal_error(os_context_t *context, boolean continuable)
 boolean
 interrupt_handler_pending_p(void)
 {
-    struct interrupt_data *data = &thread_interrupt_data(arch_os_get_current_thread());
+    struct interrupt_data *data = &thread_interrupt_data(get_sb_vm_thread());
     return (data->pending_handler != 0);
 }
 
@@ -1018,7 +1038,7 @@ interrupt_handle_pending(os_context_t *context)
      * It gets run precisely at those places where it is safe to process
      * pending asynchronous tasks. */
 
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     struct interrupt_data *data = &thread_interrupt_data(thread);
 
     if (arch_pseudo_atomic_atomic(context)) {
@@ -1051,9 +1071,7 @@ interrupt_handle_pending(os_context_t *context)
          * the os_context for the signal we're currently in the
          * handler for. This should ensure that when we return from
          * the handler the blocked signals are unblocked. */
-#ifndef LISP_FEATURE_WIN32
         sigcopyset(os_context_sigmask_addr(context), &data->pending_mask);
-#endif
         data->gc_blocked_deferrables = 0;
     }
 #endif
@@ -1061,14 +1079,11 @@ interrupt_handle_pending(os_context_t *context)
     if (read_TLS(GC_INHIBIT,thread)==NIL) {
         void *original_pending_handler = data->pending_handler;
 
-#if defined(LISP_FEATURE_SB_SAFEPOINT) && defined(LISP_FEATURE_SB_THREAD)
+#ifdef LISP_FEATURE_SB_SAFEPOINT
         /* handles the STOP_FOR_GC_PENDING case, plus THRUPTIONS */
         if (read_TLS(STOP_FOR_GC_PENDING,thread) != NIL
-# ifdef LISP_FEATURE_SB_THRUPTION
              || (read_TLS(THRUPTION_PENDING,thread) != NIL
-                 && read_TLS(INTERRUPTS_ENABLED, thread) != NIL)
-# endif
-            ) {
+                 && read_TLS(INTERRUPTS_ENABLED, thread) != NIL)) {
             /* We ought to take this chance to do a pitstop now. */
             fake_foreign_function_call(context);
             thread_in_lisp_raised(context);
@@ -1141,8 +1156,8 @@ interrupt_handle_pending(os_context_t *context)
          * an interrupt arrived during GC (POST-GC, really) it was
          * handled. */
         if (original_pending_handler != data->pending_handler)
-            lose("pending handler changed in gc: %p -> %p.",
-                 original_pending_handler, data->pending_handler);
+            lose("pending handler changed in gc: %p -> %p, signal = %d.",
+                 original_pending_handler, data->pending_handler, data->pending_signal);
     }
 
 #ifndef LISP_FEATURE_WIN32
@@ -1160,7 +1175,7 @@ interrupt_handle_pending(os_context_t *context)
         sigcopyset(os_context_sigmask_addr(context), &data->pending_mask);
         run_deferred_handler(data, context);
     }
-#ifdef LISP_FEATURE_SB_THRUPTION
+#ifdef LISP_FEATURE_SB_SAFEPOINT
     if (read_TLS(THRUPTION_PENDING,thread)==T)
         /* Special case for the following situation: There is a
          * thruption pending, but a signal had been deferred.  The
@@ -1196,7 +1211,7 @@ interrupt_handle_now(int signal, siginfo_t *info, os_context_t *context)
     if (sigismember(&deferrable_sigset,signal))
         check_interrupts_enabled_or_lose(context);
 
-    were_in_lisp = !foreign_function_call_active_p(arch_os_get_current_thread());
+    were_in_lisp = !foreign_function_call_active_p(get_sb_vm_thread());
     if (were_in_lisp)
     {
         // Use the variant of fake_ffc that doesn't do another pthread_sigmask syscall,
@@ -1290,7 +1305,7 @@ can_handle_now(void *handler, struct interrupt_data *data,
     assert_blockables_blocked();
 #endif
 
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
 
     if (read_TLS(INTERRUPT_PENDING,thread) != NIL)
         lose("interrupt already pending");
@@ -1339,7 +1354,7 @@ static void
 maybe_now_maybe_later(int signal, siginfo_t *info, void *void_context)
 {
     SAVE_ERRNO(signal,context,void_context);
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     struct interrupt_data *data = &thread_interrupt_data(thread);
     if (can_handle_now(interrupt_handle_now, data, signal, info, context))
         interrupt_handle_now(signal, info, context);
@@ -1355,7 +1370,7 @@ sig_stop_for_gc_handler(int __attribute__((unused)) signal,
                         siginfo_t __attribute__((unused)) *info,
                         os_context_t *context)
 {
-    struct thread *thread=arch_os_get_current_thread();
+    struct thread *thread=get_sb_vm_thread();
     boolean was_in_lisp;
 
     /* Test for GC_INHIBIT _first_, else we'd trap on every single
@@ -1377,7 +1392,7 @@ sig_stop_for_gc_handler(int __attribute__((unused)) signal,
 
     /* Not PA and GC not inhibited -- we can stop now. */
 
-    was_in_lisp = !foreign_function_call_active_p(arch_os_get_current_thread());
+    was_in_lisp = !foreign_function_call_active_p(get_sb_vm_thread());
 
     if (was_in_lisp) {
         /* need the context stored so it can have registers scavenged */
@@ -1480,9 +1495,8 @@ arrange_return_to_c_function(os_context_t *context,
                              call_into_lisp_lookalike funptr,
                              lispobj function)
 {
-#if !(defined(LISP_FEATURE_WIN32) || defined(LISP_FEATURE_SB_SAFEPOINT))
-    check_gc_signals_unblocked_or_lose
-        (os_context_sigmask_addr(context));
+#ifndef LISP_FEATURE_SB_SAFEPOINT
+    check_gc_signals_unblocked_or_lose(os_context_sigmask_addr(context));
 #endif
 #if !(defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64))
     void * fun=native_pointer(function);
@@ -1619,7 +1633,7 @@ arrange_return_to_c_function(os_context_t *context,
     *os_context_register_addr(context,reg_RSI) = 0;        /* arg. array */
     *os_context_register_addr(context,reg_RDX) = 0;        /* no. args */
 #else
-    struct thread *th=arch_os_get_current_thread();
+    struct thread *th=get_sb_vm_thread();
     build_fake_control_stack_frames(th,context);
 #endif
 
@@ -1713,7 +1727,7 @@ void reset_thread_control_stack_guard_page(struct thread *th)
 boolean
 handle_guard_page_triggered(os_context_t *context,os_vm_address_t addr)
 {
-    struct thread *th=arch_os_get_current_thread();
+    struct thread *th=get_sb_vm_thread();
 
     if(addr >= CONTROL_STACK_HARD_GUARD_PAGE(th) &&
        addr < CONTROL_STACK_HARD_GUARD_PAGE(th) + os_vm_page_size) {
@@ -1853,8 +1867,8 @@ low_level_handle_now_handler(int signal, siginfo_t *info, void *void_context)
  * As well there are two asynchronous signals installed via this function:
  * - STOP_FOR_GC is low-level, but might defer the signal through
  *   an intricate bunch of decisions about the state of the world.
- * - SIGURG without sb-thruption is a high-level (Lisp) handler,
- *   but with sb-thruption is low-level handler that uses different
+ * - SIGURG without :SB-SAFEPOINT is a high-level (Lisp) handler,
+ *   but with :SB-SAFEPOINT is low-level handler that uses different
  *   criteria for when to defer. */
 void
 ll_install_handler (int signal, interrupt_handler_t handler)
@@ -1862,7 +1876,7 @@ ll_install_handler (int signal, interrupt_handler_t handler)
     struct sigaction sa;
 
     if (0 > signal || signal >= NSIG
-#ifdef LISP_FEATURE_SB_THRUPTION
+#ifdef LISP_FEATURE_SB_SAFEPOINT
     /* SIGURG is in `deferrable_sigset' so that we block&unblock it properly,
      * but we don't actually want to defer it, at least not here.
      * (It might get deferred until a safepoint). And if we put it only
@@ -2104,7 +2118,7 @@ handle_trap(os_context_t *context, int trap)
         trap = trap_Error;
     }
     switch(trap) {
-#if !(defined(LISP_FEATURE_WIN32) && defined(LISP_FEATURE_SB_THREAD))
+#ifndef LISP_FEATURE_WIN32
     case trap_PendingInterrupt:
         FSHOW((stderr, "/<trap pending interrupt>\n"));
         arch_skip_instruction(context);
@@ -2178,7 +2192,7 @@ int sb_toggle_sigprof(os_context_t* context, int block) {
         gc_assert(!block);
         // Alter the mask on return from the _outermost_ signal context, which
         // should usually be the supplied context, but not if nesting happened.
-        context = nth_interrupt_context(0, arch_os_get_current_thread());
+        context = nth_interrupt_context(0, get_sb_vm_thread());
         gc_assert(context);
         sigset_t *mask = os_context_sigmask_addr(context);
         int was_blocked = sigismember(mask, SIGPROF);
