@@ -25,6 +25,7 @@
 #include "runtime.h"
 #include "code.h"
 #include "gc-internal.h"
+#include "gc-private.h"
 #include <stdarg.h>
 #include "thread.h"              /* genesis/primitive-objects.h needs this */
 #include <errno.h>
@@ -434,15 +435,32 @@ static void brief_struct(lispobj obj)
 }
 
 #include "genesis/layout.h"
+#include "genesis/defstruct-description.h"
+#include "genesis/defstruct-slot-description.h"
+#ifdef LISP_FEATURE_METASPACE
+#include "genesis/wrapper.h"
+#endif
 static boolean tagged_slot_p(struct layout *layout, int slot_index)
 {
-  lispobj bitmap = layout->bitmap;
-  sword_t fixnum = fixnum_value(bitmap); // optimistically
-  return fixnump(bitmap)
-         ? bitmap == make_fixnum(-1) ||
-            (slot_index < N_WORD_BITS && ((fixnum >> slot_index) & 1) != 0)
-         : positive_bignum_logbitp(slot_index,
-                                   (struct bignum*)native_pointer(bitmap));
+    // Since we're doing this scan, we could return the name
+    // and exact raw type.
+#ifdef LISP_FEATURE_METASPACE
+    struct wrapper *wrapper = (void*)(layout->friend-INSTANCE_POINTER_LOWTAG);
+    if (instancep(wrapper->_info)) {
+        struct defstruct_description* dd = (void*)(wrapper->_info-INSTANCE_POINTER_LOWTAG);
+#else
+    if (instancep(layout->_info)) {
+        struct defstruct_description* dd = (void*)(layout->_info-INSTANCE_POINTER_LOWTAG);
+#endif
+        lispobj slots = dd->slots;
+        for ( ; slots != NIL ; slots = CONS(slots)->cdr ) {
+            struct defstruct_slot_description* dsd =
+                (void*)(CONS(slots)->car-INSTANCE_POINTER_LOWTAG);
+            if ((fixnum_value(dsd->bits) >> DSD_INDEX_SHIFT) == slot_index)
+                return (fixnum_value(dsd->bits) & DSD_RAW_TYPE_MASK) == 0;
+        }
+    }
+    return 0;
 }
 
 static void print_struct(lispobj obj)
@@ -629,7 +647,7 @@ static void print_fun_or_otherptr(lispobj obj)
             print_obj("fun: ", symbol_function(ptr-1));
 #ifdef LISP_FEATURE_SB_THREAD
         int tlsindex = tls_index_of((struct symbol*)(ptr-1));
-        struct thread*th = arch_os_get_current_thread();
+        struct thread*th = get_sb_vm_thread();
         if (th != 0 && tlsindex != 0) {
             lispobj v = *(lispobj*)(tlsindex + (char*)th);
             print_obj("tlsval: ", v);
@@ -917,9 +935,17 @@ struct vector * classoid_name(lispobj * classoid)
 }
 struct vector * layout_classoid_name(lispobj * layout)
 {
+#ifdef LISP_FEATURE_METASPACE
+  // layout can't be forwarded, but wrapper could be
+  lispobj* wrapper = native_pointer(((struct layout*)layout)->friend);
+  if (forwarding_pointer_p(wrapper))
+      wrapper = native_pointer(forwarding_pointer_value(wrapper));
+  lispobj classoid = ((struct wrapper*)wrapper)->classoid;
+#else
   if (forwarding_pointer_p(layout))
       layout = native_pointer(forwarding_pointer_value(layout));
   lispobj classoid = ((struct layout*)layout)->classoid;
+#endif
   return instancep(classoid) ? classoid_name(native_pointer(classoid)) : NULL;
 }
 struct vector * instance_classoid_name(lispobj * instance)
