@@ -375,8 +375,7 @@
 ;;; Skip for MSAN. Instead of returning 0, the intercepted malloc is configured
 ;;; to cause process termination by default on failure to allocate memory.
 (with-test (:name :malloc-failure
-                  :skipped-on :msan
-                  :fails-on :alpha) ;; Alpha has address space to burn
+                  :skipped-on :msan)
   (assert (eq :enomem
               (handler-case
                   (loop repeat 128
@@ -532,3 +531,11 @@
     (t (c)
       (assert (typep c 'sb-kernel::undefined-alien-function-error))
       (assert (equal (cell-error-name c) "bar")))))
+
+(defconstant fleem 3)
+;; We used to expand into
+;; (SYMBOL-MACROLET ((FLEEM (SB-ALIEN-INTERNALS:%ALIEN-VALUE
+;; which conflicted with the symbol as global variable.
+(with-test (:name :def-alien-rtn-use-gensym)
+  (checked-compile '(lambda () (define-alien-routine "fleem" int (x int)))
+                   :allow-style-warnings (or #-(or :x86-64 :arm :arm64) t)))
