@@ -3099,3 +3099,38 @@
      `(lambda ()
         (,name :k 0f0))
      (() 0f0))))
+
+(with-test (:name :no-*-as-type)
+  (multiple-value-bind (fun warn err) (compile nil '(lambda (x) (the * x)))
+    (assert (and warn err))
+    (assert-error (funcall fun 1)))
+  ;; (values t) parses into *wild-type* and has to be allowed
+  ;; even though * which parses into *wild-type* isn't.
+  (checked-compile '(lambda () (the (values t) t))))
+
+(with-test (:name :hairy-data-vector-set-t-upgrade)
+  (checked-compile
+   '(lambda (x) (sb-kernel:hairy-data-vector-set
+                 (the (simple-array symbol) x) 1 'hey))))
+
+(with-test (:name :ir2-convert-reffer-no-lvar)
+  (checked-compile-and-assert
+   (:allow-style-warnings t)
+   `(lambda (a)
+      (/ (unwind-protect (if a
+                             (values nil (cdr a))
+                             (values 1 0))
+           a)
+         1))
+   ((nil) 1)))
+
+(with-test (:name :%eql-integer-fold)
+  (checked-compile-and-assert
+   ()
+   `(lambda (d)
+      (declare (type fixnum d))
+      (or (find d '(-98 27749116333474161060))
+          t))
+   ((-98) -98)
+   ((95) t)))
+
