@@ -53,7 +53,11 @@
                 (expect
                  (map 'list 'sb-kernel:layout-id
                       (sb-kernel:layout-inherits layout))))
-            (unless (equal (append '(1 2) ids)
+            (unless (equal (list* (sb-kernel:layout-id
+                                   (sb-kernel:find-layout 't))
+                                  (sb-kernel:layout-id
+                                   (sb-kernel:find-layout 'structure-object))
+                                  ids)
                            (append expect (list (sb-kernel:layout-id layout))))
               (error "Wrong IDs for ~A: expect ~D actual ~D~%"
                      layout expect ids))))))))
@@ -808,33 +812,19 @@
                              (aref (vector x) (incf i)))
                     (bug-348-x x)))))
 
-;;; obsolete instance trapping
-;;;
-;;; FIXME: Both error conditions below should possibly be instances
-;;; of the same class. (Putting this FIXME here, since this is the only
-;;; place where they appear together.)
-
 (with-test (:name :obsolete-defstruct/print-object)
   (eval '(defstruct born-to-change))
   (let ((x (make-born-to-change)))
     (handler-bind ((error 'continue))
       (eval '(defstruct born-to-change slot)))
-    (assert (eq :error
-                (handler-case
-                    (princ-to-string x)
-                  (sb-pcl::obsolete-structure ()
-                    :error))))))
+    (assert (search "UNPRINTABLE instance" (princ-to-string x)))))
 
 (with-test (:name :obsolete-defstruct/typep)
   (eval '(defstruct born-to-change-2))
   (let ((x (make-born-to-change-2)))
     (handler-bind ((error 'continue))
       (eval '(defstruct born-to-change-2 slot)))
-      (assert (eq :error2
-                  (handler-case
-                      (typep x (find-class 'standard-class))
-                    (sb-kernel:layout-invalid ()
-                      :error2))))))
+      (assert (not (typep x (find-class 'standard-class))))))
 
 ;; EQUALP didn't work for structures with float slots (reported by
 ;; Vjacheslav Fyodorov).
@@ -955,13 +945,8 @@ redefinition."
 (defun assert-is (predicate instance)
   (assert (funcall predicate instance)))
 
-;;; It used to call the predicate function, but out-of-line predicate
-;;; functions now don't signal layout-invalid, just like inlined
-;;; predicates didn't signal it.
 (defun assert-invalid (instance)
-  (declare (notinline typep))
-  (assert (typep (nth-value 1 (ignore-errors (typep instance 'structure-object)))
-                 'sb-kernel::layout-invalid)))
+  (assert (sb-kernel:layout-invalid (sb-kernel:%instance-layout instance))))
 
 ;; Don't try to understand this macro; just look at its expansion.
 (defmacro with-defstruct-redefinition-test (name
