@@ -51,6 +51,43 @@ TEST_DIRECTORY=$junkdir SBCL_HOME=../obj/sbcl-home exec ../src/runtime/sbcl \
             (cond ((< posn-a posn-b) t)
                   ((> posn-a posn-b) nil)
                   (t (string< a b)))))))
+(defun summarize-gc-times ()
+  (let (observations)
+    (flet ((parse-triple (string pos)
+             (sb-int:binding* (((int1 end) (parse-integer string :start (1+ pos)
+                                                                 :junk-allowed t))
+                               ((int2 end) (parse-integer string :start (1+ end)
+                                                                 :junk-allowed t))
+                               ((int3) (parse-integer string :start (1+ end)
+                                                             :junk-allowed t)))
+               (list int1 int2 int3))))
+      (dolist (pn (directory "$logdir/*.*"))
+        (with-open-file (f pn)
+          (let ((legend "GC: time-to-stw"))
+            (loop
+             (let ((line (read-line f nil)))
+               (unless line (return))
+               (when (and (> (length line) (length legend))
+                          (string= line legend :end1 (length legend)))
+                 (let* ((p1 (position #\= line))
+                        (p2 (position #\= line :start (1+ p1)))
+                        (stw-dur (parse-triple line p1))
+                        (gc-dur (parse-triple line p2))
+                        (count (parse-integer line :start (+ (search "over " line) 5)
+                                              :junk-allowed t)))
+                   (let ((name (concatenate 'string (pathname-name pn) "."
+                                            (pathname-type pn))))
+                     (push (list count stw-dur gc-dur name) observations))))))))))
+    (let ((fmt "  ~5d (~{~10d~^ ~}) (~{~7d~^ ~}) ~a~%"))
+      (format t "~&Top 15 worst by max time to stop-the-world:~%")
+      (let ((list (sort (copy-list observations) #'> :key (lambda (x) (third (second x))))))
+        (dotimes (i 15) (apply #'format t fmt (pop list))))
+      (format t "~&Top 15 worst by avg GC duration (excluding STW delay):~%")
+      (let ((list (sort (copy-list observations) #'> :key (lambda (x) (second (third x))))))
+        (dotimes (i 15) (apply #'format t fmt (pop list))))
+      (format t "~&Top 15 worst by max GC duration (excluding STW delay):~%")
+      (let ((list (sort (copy-list observations) #'> :key (lambda (x) (third (third x))))))
+        (dotimes (i 15) (apply #'format t fmt (pop list)))))))
 (defun parallel-execute-tests (files max-jobs vop-summary-stats-p)
   (format t "Using ~D processes~%" max-jobs)
   ;; Interleave the order in which all tests are launched rather than
@@ -106,6 +143,9 @@ TEST_DIRECTORY=$junkdir SBCL_HOME=../obj/sbcl-home exec ../src/runtime/sbcl \
           (wait))
         (let ((pid (sb-posix:fork)))
           (when (zerop pid)
+            #+(and linux sb-thread 64-bit)
+            (sb-alien:alien-funcall (sb-alien:extern-alien "reset_gc_stats"
+                                    (function sb-alien:void)))
             ;; FILE is (filename . test-iteration)
             (with-open-file (stream (format nil "$logdir/~a~@[-~d~]" (car file) (cdr file))
                                     :direction :output :if-exists :supersede)
@@ -120,7 +160,7 @@ TEST_DIRECTORY=$junkdir SBCL_HOME=../obj/sbcl-home exec ../src/runtime/sbcl \
                    (let ((shell (or #+sunos (posix-getenv "SHELL") "/bin/sh")))
                      ;; exec the shell with the test and we'll pick up its exit code
                      (alien-funcall (extern-alien "execl" (function int c-string c-string
-                                                                    c-string unsigned))
+                                                                    &optional c-string unsigned))
                                     shell shell
                                     (concatenate 'string file ".sh") 0))
                    ;; if exec fails, just exit with a wrong (not 104) status
@@ -184,7 +224,8 @@ TEST_DIRECTORY=$junkdir SBCL_HOME=../obj/sbcl-home exec ../src/runtime/sbcl \
                                 (impure-cload-files)
                                 (sh-files)))))
        jobs
-       t))
+       t)
+      #+(and linux sb-thread 64-bit) (summarize-gc-times))
     ;; long form
     (let ((jobs 4)
           (runs-per-test 1)
