@@ -125,13 +125,7 @@
   ;; needs to wait, to have a facsimile of the situation prior to implementation
   ;; of the so-called pauseless thread start feature.
   ;; Wouldn't you know, it's just reintroducing a startup semaphore.
-  ;; And interruption tests are even more likely to fail with sb-thruption
-  ;; because sb-thruption is flawed: it presumes that there is enough synchronization
-  ;; between sender/receiver that checking the INVOKED variable (shared via a closure)
-  ;; makes any sense at all, which it doesn't. (In addition, it supposes that merely
-  ;; by polling at safepoints, interrupts somehow become safe, which is not true
-  ;; in general - it is only true of the GC "interrupt" delivered by the kernel
-  ;; when the safepoint page trap is hit.)
+  ;; And interruption tests are even more likely to fail with :sb-safepoint.
   ;; Noneless, this tries to be robust enough to pass.
   (let* ((sem (sb-thread:make-semaphore))
          (child  (make-kill-thread
@@ -185,10 +179,6 @@
         (setf threads (union (union *threads-to-kill*
                                     *threads-to-join*)
                              threads))
-        #+(and sb-safepoint-strictly (not win32))
-        (dolist (thread (sb-thread:list-all-threads))
-          (when (typep thread 'sb-thread:signal-handling-thread)
-            (ignore-errors (sb-thread:join-thread thread))))
         (dolist (thread (sb-thread:list-all-threads))
           (unless (or (not (sb-thread:thread-alive-p thread))
                       (eql (the sb-thread:thread thread)
@@ -833,7 +823,8 @@
         finally (return (/ min-internal-time-units-per-call
                            (float internal-time-units-per-second)))))
 
-(defmacro runtime (form &key (repetitions 5) (precision 30))
+(defmacro runtime (form &key (repetitions 5) (precision (* 30
+                                                           (/ internal-time-units-per-second 1000))))
   `(runtime* (lambda () ,form) ,repetitions ,precision))
 
 (declaim (notinline opaque-identity))
@@ -925,3 +916,21 @@
             (when (>= (sb-disassem:dstate-cur-offs dstate) (sb-disassem:seg-length segment))
               (return)))
       (result))))
+
+;;; If a test file is not very tolerant of the statistical profiler, then
+;;; it should call this. There seem to be at least 2 common categories of failure:
+;;;  * the stack exhaustion test can get a profiling signal when you're already
+;;;    near exhaustion, and then the profiler exhausts the stack, which isn't
+;;;    handled very well. Example:
+;;;    Control stack exhausted, fault: 0xd76b9ff8, PC: 0x806cad2
+;;;       0: fp=0xd76ba008 pc=0x806cad2 Foreign function search_dynamic_space
+;;;       1: fp=0xd76ba028 pc=0x80633d1 Foreign function search_all_gc_spaces
+;;;       2: fp=0xd76ba048 pc=0x805647e Foreign function component_ptr_from_pc
+;;;       3: fp=0xd76baa18 pc=0x8065dfd Foreign function (null)
+;;;       4: fp=0xd76baa98 pc=0x806615a Foreign function record_backtrace_from_context
+;;;       5: fp=0xd76baab8 pc=0x806625e Foreign function sigprof_handler
+;;;  * debugger-related tests. I'm not sure why.
+(defun disable-profiling ()
+  (when (find-package "SB-SPROF")
+    (format t "INFO: disabling SB-SPROF~%")
+    (funcall (intern "STOP-PROFILING" "SB-SPROF"))))
