@@ -148,7 +148,8 @@
   (let* ((lines
           (split-string
            (with-output-to-string (s)
-            (let ((sb-disassem:*disassem-location-column-width* 0))
+            (let ((sb-disassem:*disassem-location-column-width* 0)
+                  (sb-kernel::*print-layout-id* nil))
               (disassemble '(lambda (x) (the sb-assem:label x))
                            :stream s)))
            #\newline))
@@ -346,6 +347,14 @@
                   (and (search "CMP QWORD PTR [" line)
                        (search ":YUP" line))))))
 
+(defun thing-ref-thing-ref (arg1 arg2)
+  (declare (optimize (safety 0)))
+  (let ((answer (typep (thing-x (thing-x arg1)) 'fixnum)))
+    (frobify arg1 arg2)
+    answer))
+(defun frobify (a b) (values a b))
+(compile 'thing-ref-thing-ref)
+
 (with-test (:name :fixnump-thing-ref)
   (flet ((try (access-form true false)
            (let* ((f (compile nil `(lambda (obj) (typep ,access-form 'fixnum))))
@@ -364,7 +373,9 @@
     (try '(thing-x (truly-the thing obj))
          (make-thing :x 1) (make-thing :x "hi"))
     (try '(car obj) '(1) '("hi"))
-    (try '(cdr obj) '("hi" . 1) '("hi"))))
+    (try '(cdr obj) '("hi" . 1) '("hi")))
+  ;; fixnump of memref of memref was eliding one memref by accident
+  (assert (thing-ref-thing-ref (make-thing :x (make-thing :x 3)) 'foo)))
 
 (with-test (:name :huge-code :skipped-on (not :immobile-code))
   (sb-vm::allocate-code-object :immobile 0 4 (* 2 1024 1024)))
@@ -616,7 +627,8 @@ sb-vm::(define-vop (cl-user::test)
    ;; to :dword because zeroing the upper 32 bits is a visible effect.
    (inst mov (reg-in-size rax-tn :qword) (reg-in-size rcx-tn :qword))
    (inst mov rax-tn rcx-tn)))
-(with-test (:name :mov-mov-elim-ignore-resized-reg) ; just don't crash
+(with-test (:name :mov-mov-elim-ignore-resized-reg
+                  :fails-on :sbcl) ; just don't crash
   (checked-compile '(lambda () (sb-sys:%primitive test) 0)))
 
 (defstruct a)
@@ -743,14 +755,17 @@ sb-vm::(define-vop (cl-user::test)
   ;; This test verifies that debug-fun-from-pc does not croak when the PC points
   ;; within a trampoline allocated to wrap a closure in a simple-funifying wrapper
   ;; for installation into a global symbol.
-  (assert (sb-kernel:closurep #'sb-int:constantly-0))
-  (let ((trampoline
-          (sb-di::code-header-from-pc
-           (sb-sys:int-sap (sb-vm::fdefn-raw-addr
-                            (sb-kernel::find-fdefn 'sb-int:constantly-0))))))
-    (assert (zerop (sb-kernel:code-n-entries trampoline)))
-    (assert (typep (sb-di::debug-fun-from-pc trampoline 8)
-                   'sb-di::bogus-debug-fun))))
+  (let ((closure (funcall (compile nil '(lambda (x)  (lambda () x))) 0))
+        (symbol (gensym)))
+    (assert (sb-kernel:closurep closure))
+    (setf (fdefinition symbol) closure)
+    (let ((trampoline
+            (sb-di::code-header-from-pc
+             (sb-sys:int-sap (sb-vm::fdefn-raw-addr
+                              (sb-kernel::find-fdefn symbol))))))
+      (assert (zerop (sb-kernel:code-n-entries trampoline)))
+      (assert (typep (sb-di::debug-fun-from-pc trampoline 8)
+                     'sb-di::bogus-debug-fun)))))
 
 (defstruct foo (s 0 :type (or null string)))
 (with-test (:name :reduce-stringp-to-not-null)
@@ -788,3 +803,148 @@ sb-vm::(define-vop (cl-user::test)
 (with-test (:name :gf-self-contained-trampoline)
   (let ((l (sb-kernel:find-layout 'standard-generic-function)))
     (assert (/= (sb-kernel:layout-bitmap l) sb-kernel:+layout-all-tagged+))))
+
+(with-test (:name :known-array-rank)
+  (flet ((try (type)
+           (let ((lines
+                  (disassembly-lines
+                   `(lambda (x)
+                     #+sb-safepoint (declare (optimize (sb-c::insert-safepoints 0)))
+                     (array-rank (truly-the ,type x))))))
+             ;; (format t "~{~&~A~}" lines)
+             ;; Naturally this is brittle as heck. I wish we had a better way.
+             (assert (<= (length lines) 9)))))
+    (try 'string)
+    (try '(and vector (not simple-array)))
+    (try '(or string bit-vector)) ; not an array type, but known rank
+    (try '(array * (4 5 *)))))
+
+;;; Match the same set of objects that %OTHER-POINTER-P does.
+(deftype other-pointer-object ()
+  '(not (or fixnum single-float function list sb-kernel:instance character)))
+
+;;; Helper to assert something about how many comparisons it takes to test
+;;; for various sets of widetags.
+(defun count-cmp-opcodes (type expect function)
+  (let ((lines (disassembly-lines function)))
+    (let ((actual
+           (loop for line in lines
+                 count (or (search "CMP" line)
+                           (search "TEST" line)))))
+      (unless (eql actual expect)
+        (format t "~{~&~a~}~%" lines)
+        (error "typep: needed ~d test ops but expected ~d for ~a"
+               actual expect type)))))
+
+(defun check-arrayp-cmp-opcodes (expect type)
+  ;; Assume the lowtag test passed already.
+  (count-cmp-opcodes type expect
+          `(lambda (x)
+             (declare (optimize (sb-c::verify-arg-count 0)
+                                #+sb-safepoint (sb-c::insert-safepoints 0)))
+             (typep (truly-the other-pointer-object x) ',type))))
+
+(with-test (:name :arrayp-exactly-one-comparison-etc)
+  (check-arrayp-cmp-opcodes 1 'array)
+
+  (check-arrayp-cmp-opcodes 1 'string)
+  (check-arrayp-cmp-opcodes 1 'base-string) ; widetags differing in one bit
+  #+sb-unicode
+  (check-arrayp-cmp-opcodes 1 'sb-kernel::character-string) ; ditto
+  (check-arrayp-cmp-opcodes 1 'simple-string) ; 2 adjacent widetags
+  (check-arrayp-cmp-opcodes 1 '(and string (not simple-array)))
+  (check-arrayp-cmp-opcodes 1 'simple-base-string)
+  #+sb-unicode
+  (check-arrayp-cmp-opcodes 1 'sb-kernel::simple-character-string)
+  ;; FIXME: (AND STRING (NOT SIMPLE-STRING)) executs 4 tests
+  ;; but it denotes the same set of objects as (AND STRING (NOT SIMPLE-ARRAY)).
+  ;; The problem is with type algebra, not in the backend.
+
+  ;; FIXME: this should be 1 comparison: widetag >= start-of-complex-widetags
+  (check-arrayp-cmp-opcodes 2 '(and array (not simple-array)))
+
+  ;; some other interesting pairs
+  ;; This was passing just by random coincidence.
+  ;; The widetag patterns no longer differ in exactly 1 bit.
+  ;; Is there any real relevance to this test?
+  #+nil
+  (check-arrayp-cmp-opcodes 1 '(or (simple-array (unsigned-byte 8) (*))
+                                    (simple-array (unsigned-byte 16) (*))))
+  ;; FIXME: what's up with SIMPLE-UNBOXED-ARRAY requiring 1 SUB,
+  ;; 3 CMPs, and a MOVZX. Something doesn't feel right.
+  ;; In general, a lot of the array types are source-transforming to
+  ;; (AND (SB-KERNEL:%OTHER-POINTER-P #:OBJECT0)
+  ;;      (EQ (SB-KERNEL:%OTHER-POINTER-WIDETAG #:OBJECT0) something))
+  )
+
+(defun check-integerp-cmp-opcodes (expect type)
+  (count-cmp-opcodes type expect
+          `(lambda (x)
+             (declare (optimize (sb-c::verify-arg-count 0)
+                                #+sb-safepoint (sb-c::insert-safepoints 0)))
+             (typep x ',type))))
+
+(with-test (:name :typep-integer-doubleton)
+  ;; This was taking 3 comparisons because it was a FIXNUMP test
+  ;; and some range-based testing rather than just 2 EQ tests.
+  (check-integerp-cmp-opcodes 2 '(integer 1 2)))
+
+(defun typep-asm-code-length (type)
+  (let* ((lines
+          (disassembly-lines
+           (compile nil
+                    `(lambda (x)
+                       (declare (optimize (sb-c::verify-arg-count 0) (debug 0)))
+                       (typep x ',type)))))
+         (callp
+          (some (lambda (x) (or (search "#<FUNCTION" x)
+                                (search "#<FDEFN" x)))
+                lines)))
+    (values (length lines) callp)))
+
+;;; Counting instructions of assembly is sort of a very rough guess
+;;; as to whether widetags are being tested as efficiently as possible.
+;;; This file is for both human and machine consumption.
+;;; #\! precedes any line where the type test involves a function call.
+;;; It might be just a call. We might consider inlining some of those.
+(defun write-golden-typep-data (input-name output-name)
+  (with-open-file (input input-name)
+    (with-open-file (output output-name :direction :output
+                            :if-exists :supersede
+                            :if-does-not-exist :create)
+      (let ((*package* (find-package "SB-KERNEL"))
+            (*print-pretty* nil))
+        (loop (let ((type (read input nil input)))
+                (when (eq type input) (return))
+                (multiple-value-bind (linecount callp) (typep-asm-code-length type)
+                  (format output "~a ~3d ~s~%"
+                          (if callp "!" " ")
+                          linecount type))))))))
+
+#+nil
+(write-golden-typep-data "../interesting-types.lisp-expr"
+                         "typep-golden-data.txt")
+
+(defun compare-to-golden-typep-data (pathname)
+  (with-open-file (input pathname)
+    (let ((*package* (find-package "SB-KERNEL")))
+      (loop (let ((line (read-line input nil input)))
+              (when (eq line input) (return))
+              (with-input-from-string (stream line :start 2)
+                (let ((expect-n (read stream))
+                      (type (read stream)))
+                  (multiple-value-bind (linecount callp) (typep-asm-code-length type)
+                    (declare (ignore callp))
+                    (when (/= linecount expect-n)
+                      (warn "~S was ~d is ~d" type expect-n linecount))))))))))
+
+(with-test (:name :many-interesting-array-types :skipped-on (:not :sb-unicode))
+  (compare-to-golden-typep-data "typep-golden-data.txt"))
+
+(with-test (:name :integerp->bignump-strength-reduction)
+  (let ((f1 (compile nil '(lambda (x)
+                           (typecase x (fixnum 'a) (integer 'b) (t 'c)))))
+        (f2 (compile nil '(lambda (x)
+                           (typecase x (fixnum 'a) (bignum 'b) (t 'c))))))
+    (assert (= (length (disassembly-lines f1))
+               (length (disassembly-lines f2))))))
